#if !defined(__branch_h)
#define __branch_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

#include "sink.h"
#include "lqueue.h"

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
class CBranch : public CSink<CInItem>
{
public:
	typedef CQueue<COutItem> COutQueue;
	typedef CQueue<COutItem1> COutQueue1;
	typedef CQueue<COutItem2> COutQueue2;

private:
	COutQueue &OutQueue;
	COutQueue1 &OutQueue1;
	COutQueue2 &OutQueue2;

public:
	CBranch(CReporter *aReporter,
		CQueue<CInItem> &aInQueue,
		CQueue<COutItem> &aOutQueue,
		CQueue<COutItem1> &aOutQueue1,
		CQueue<COutItem2> &aOutQueue2);
	virtual ~CBranch() { }

	virtual short Shuffle();

protected:
	virtual void SetIn(CInItem *InItem) = 0;
	virtual void NewOut(COutItem **OutItem,
			COutItem1 **OutItem1, COutItem2 **OutItem2) = 0;
	virtual void InformNext();
	virtual void CleanUp(); // empty

private:
	virtual void Step(); // cannot be called
};

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
CBranch<CInItem, COutItem, COutItem1, COutItem2>::
		CBranch(CReporter *aReporter,
	CQueue<CInItem> &aInQueue,
	CQueue<COutItem> &aOutQueue,
	CQueue<COutItem1> &aOutQueue1,
	CQueue<COutItem2> &aOutQueue2):
	CSink<CInItem>(aReporter, aInQueue),
	OutQueue(aOutQueue),
	OutQueue1(aOutQueue1),
	OutQueue2(aOutQueue2)
{
}

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
short CBranch<CInItem, COutItem, COutItem1, COutItem2>::Shuffle()
{	COutItem *OutItem;
	COutItem1 *OutItem1;
	COutItem2 *OutItem2;

	if (!IsFilled())
	{	ReportIn();
		return ShuffleIn();
	}
	else
	{
		if (OutQueue.IsFull())
		{	ReportProgress(0, GetPerc());
			OutQueue.SetBlockedOnPut(this);
			return 1;
		}
		if (OutQueue1.IsFull())
		{	ReportProgress(0, GetPerc());
			OutQueue1.SetBlockedOnPut(this);
			return 1;
		}
		if (OutQueue2.IsFull())
		{	ReportProgress(0, GetPerc());
			OutQueue2.SetBlockedOnPut(this);
			return 1;
		}

		NewOut(&OutItem, &OutItem1, &OutItem2);
		Assert(OutItem || OutItem1 || OutItem2);
		if (OutItem)
			OutQueue.Put(OutItem);
		if (OutItem1)
			OutQueue1.Put(OutItem1);
		if (OutItem2)
			OutQueue2.Put(OutItem2);
		ReportProgress(1, GetPerc());
		return 0;
	}
}

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
void CBranch<CInItem, COutItem, COutItem1, COutItem2>::InformNext()
{
	CSink<CInItem>::InformNext();
	OutQueue.SetSourceDone();
	OutQueue1.SetSourceDone();
	OutQueue2.SetSourceDone();
}

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
void CBranch<CInItem, COutItem, COutItem1, COutItem2>::CleanUp()
{
	CSink<CInItem>::CleanUp();
	OutQueue.Wake();
	OutQueue1.Wake();
	OutQueue2.Wake();
}

template<class CInItem, class COutItem, class COutItem1,
		class COutItem2>
void CBranch<CInItem, COutItem, COutItem1, COutItem2>::Step()
{
	throw CGeneralException("Step called for CBranch");
}

#endif

