#include "myfndt.h"
#include "cluwrit.h"

CClusterWriter::CClusterWriter(CInQueue &aInQueue,
	const DString &ClusterFile):
	CSink<CCluster>(NULL, // CClusterWriter does not report progress
		aInQueue),
	ClusterBook(new CULongHandler(), ClusterFile, IT_CREATE | IT_UPDATE),
			Cluster(NULL)
{
}

CClusterWriter::~CClusterWriter()
{
	delete Cluster;
}

short CClusterWriter::IsFilled()
{
	return Cluster != NULL;
}

void CClusterWriter::SetIn(CCluster *InCluster)
{
	Assert(!Cluster);
	Cluster = InCluster;
}

void CClusterWriter::Step()
{
	Assert(Cluster != NULL);
	ulong FirstWordNum = Cluster->GetFirstWordNum();
	CNewKey Key(CULongHandler::NewKey(FirstWordNum));
	ClusterBook.Add(Key, Cluster->GetOffs());

	delete Cluster;
	Cluster = NULL;
}

void CClusterWriter::CleanUp()
{
	CSink<CCluster>::CleanUp();
	ClusterBook.StoreAll();
}

#if defined(DEBUG_CClusterWriter)
short CClusterWriter::Shuffle()
{
	if (!IsFilled())
	{	ReportIn();
		return ShuffleInX();
	}
	else
	{  Step();
		ReportProgress(1, GetPerc());
		return 0;
	}
}

short CClusterWriter::ShuffleInX()
{
	if (InQueue.IsEmpty())
	{	if (IsSourceDone())
		{  CleanUp();
			InformNext();
			return 2;
		}
		else
		{	InQueue.SetBlockedOnGet(this);
			return 1;
		}
	}
	CCluster *InItem = InQueue.Get();
	SetIn(InItem);
	return 0;
}
#endif

