#include "myfndt.h"
#include "cocc.h"
#include <stdlib.h>

COccCol::COccCol():
    Data(new COccColData),
    Master(true)
{
}

COccCol::COccCol(const COccCol &Other):
    Data(Other.Data),
    Master(Other.Master)
{
    ++(Data->RefCount);
}

COccCol::COccCol(bool aMaster):
    Data(new COccColData),
    Master(aMaster)
{
}

COccCol::COccCol(const COccCol &Other, bool aMaster):
    Data(Other.Data),
    Master(aMaster)
{
    ++(Data->RefCount);
}

COccCol::~COccCol()
{
    if (!(--(Data->RefCount)))
	   delete Data;
}

COccCol &COccCol::operator = (const COccCol &Other)
{
    ++(Other.Data->RefCount);
    if (!(--(Data->RefCount)))
	   delete Data;
    Data = Other.Data;
    Master = Other.Master;
    return *this;
}

void COccCol::Add(ulong Occ)
{
    Data->push_back(Occ);
}

CComposedOcc::CComposedOcc(DString &aOccWord):
	OccWord(aOccWord)
{
}

CComposedOcc::CComposedOcc(const DString &aOccWord, const COccCol &aOccCol):
     OccCol(aOccCol),
	OccWord(aOccWord)
{
}

void CComposedOcc::Add(ulong Occ)
{
	OccCol.Add(Occ);
}

// default DString comparisons do not work in gcc!
int CComposedOcc::operator == (const CComposedOcc &Other) const
{ 
    const char *p = GetWord().c_str();
    const char *q = Other.GetWord().c_str(); 
    Assert(p != NULL && q != NULL);
    return strcmp(p, q) == 0;
}

int CComposedOcc::operator < (const CComposedOcc &Other) const
{ 
    const char *p = GetWord().c_str();
    const char *q = Other.GetWord().c_str(); 
    Assert(p != NULL && q != NULL);
    return strcmp(p, q) < 0;
}

