#if !defined(__cocc_h)
#define __cocc_h

#include "vector.h"

class COccColData : public vector<ulong>
{
public:
    int RefCount;

    COccColData():RefCount(1) { vector<ulong>::reserve(5); }
    COccColData(const COccColData &Other):vector<ulong>(Other), RefCount(1) { }
};

class COccCol
{
public:
    COccCol();
    COccCol(const COccCol &Other);
    COccCol(bool aMaster);
    COccCol(const COccCol &Other, bool aMaster);
    virtual ~COccCol();
    COccCol &operator = (const COccCol &Other);

    inline const vector<ulong> &GetOccs() const { return *Data; }

    // N. B.: does _not_ copy shared data
    void Add(ulong Occ);

    ulong operator[](int Index) const { return Data->operator[](Index); }

    bool IsMaster() const { return Master; }

private:
    int operator == (const COccCol &Other) const;
    int operator != (const COccCol &Other) const;
    int operator < (const COccCol &Other) const;

private:
    COccColData *Data;
    bool Master;
};

class CComposedOcc
{
private:
    DString OccWord;
    COccCol OccCol;

    CComposedOcc(CComposedOcc &Src); // not implemented
    CComposedOcc &operator = (CComposedOcc &Src);
        // not implemented

public:
    CComposedOcc(DString &aOccWord);
    CComposedOcc(const DString &aOccWord, const COccCol &aOccCol);

    ~CComposedOcc() { }
        // not virtual to save space;
        // do not inherit from CComposedOcc!

    void Add(ulong Occ);

    const DString &GetWord() const { return OccWord; }

    inline const vector<ulong> &GetOccs() const { return OccCol.GetOccs(); }

    int operator == (const CComposedOcc &Other) const;
    int operator < (const CComposedOcc &Other) const;
};

#endif

