#include "myfndt.h"

#include "crunopt.h"
#include "fillopt.h"
#include "obase.h"
#include "offind.h"
#include "oindex.h"
#include "otxtrpt.h"
#include "opunct.h"
#include "ostoplst.h"
#include "ocross.h"
#include "ocase.h"
#include "ocache.h"
#include "osubdoc.h"
#include "ocrunout.h"

#include "conveyor.h"
#include "lqueue.h"
#include "indexer.h"
#include "occsort.h"
#include "cluster.h"
#include "cluwrit.h"
#include "rdisp.h"
#include "ffinder.h"
#include "globfile.h"
#include "procfile.h"
#include "filewrit.h"
#include "filesiv.h"
#include "wordsiv.h"
#include "txtrpt.h"
#include "stoplist.h"
#include "growsl.h"
#include "idxstor.h"
#include "filebase.h"
#include "newcase.h"
#include "decompr.h"
#include "exedrive.h"
#include "newstem.h"
#include "infra.h"

#include <stdio.h>
#if defined(__BORLANDC__)
#include <io.h>
#else
#if defined(__GNUG__)
#include <unistd.h>
#else
#if defined(_MSC_VER)
#include <io.h>
#define access _access
#endif
#endif
#endif

const int RawFileQueueSize = 10;
const int GlobalFileQueueSize = 10;
const int FinalFileQueueSize = 5;
const int OccQueueSize = 10;
const int ClusterQueueSize = 5;
const int OccBlockQueueSize = 1;
	// must be short - COccBlock is big

void RunScan(CStopList *StopList,
	CIndexStorage *IndexStorage,
	CCruncherOptions &Options);

CInfrastructure *NewInfrastructureOrNull(CCruncherOptions &Options);

void Run(int Count, char *Args[])
{	
#if !defined(DEBUG_hollow)
    CObjPtr<CStopList> StopList = NULL;

#if defined(_MSC_VER)
	SetThreadPriority(GetCurrentThread(), 
		THREAD_PRIORITY_HIGHEST);
#endif
#endif

	if (Count != 1)
	{	printf("all parameters must be specified in locus.opt");
		return;
	}

#if !defined(DEBUG_hollow)
	CCruncherOptions Options(GetExeDrive(Args[0]));
	if (FillOptions(Options))
		return;

	CIndexStorage IndexStorage(
		new CCompressedStrHandler((ushort)Options.COptions<int>::Get(
      		O_AVEWORDLEN)), Options);

	DString StopListSource = Options.Get(O_STOPLISTSRC);
	DString StopListFile = Options.Get(O_STOPLIST);
	if (!StopListSource.IsEmpty())
	{	printf("adding words from %s into the stop list... ",
			StopListSource.c_str());
		StopList = new CStopList(StopListFile, IT_CREATE | IT_UPDATE);
		GrowStopList(StopList, StopListSource, &IndexStorage);
		printf("Done.\n");
	}
	else
	{  if (!access(StopListFile.c_str(), 0))
			StopList = new CStopList(StopListFile, 0);
	}

	RunScan(StopList, &IndexStorage, Options);
#endif
}

void RunScan(CStopList *StopList,
	CIndexStorage *IndexStorage,
	CCruncherOptions &Options)
{  CReporter *Reporter;
	CManipulator *Manipulator;
	short ret;

	CConveyor Conveyor(8);

	CQueue<DString> RawFileQueue(RawFileQueueSize, Conveyor);
	// here sits FileSieve
	CQueue<CGlobalFile> GlobalFileQueue(GlobalFileQueueSize, Conveyor);
	// here sits ReaderDispatch
	CQueue<CProcessedFile> FinalFileQueue(FinalFileQueueSize, Conveyor);
	CQueue<CCluster> ClusterQueue(ClusterQueueSize, Conveyor);
	CQueue<CSimpleOcc> RawOccQueue(OccQueueSize, Conveyor);
	// here sits WordSieve
	CQueue<CSimpleOcc> StemmedOccQueue(OccQueueSize, Conveyor);
	// here sits sorter
	CQueue<COccBlock> OccBlockQueue(OccBlockQueueSize, Conveyor);

	// used by FileSieve & FileWriter
	CFileBase FileBase(IT_CREATE | IT_UPDATE,
		Options);

	Reporter = new CTextReporter(Options.Get(T_FINDERFORMAT_DDN),
		Options.COptions<int>::Get(O_FINDCOUNTER));
	Manipulator = new CFileFinder(Reporter,
		RawFileQueue,
		Options.COptions<CStringListValue>::Get(O_DOCDIR),
		Options.COptions<CStringListValue>::Get(O_MASK));
	Conveyor.AddNew(Manipulator);

	Conveyor.AddNew(new CFileSieve(RawFileQueue,
		GlobalFileQueue,
		FileBase));

	Reporter = new CTextReporter(Options.Get(T_READERFORMAT_DDN),
		Options.COptions<int>::Get(O_READCOUNTER));
	Manipulator = new CReaderDispatch(Reporter,
		GlobalFileQueue,
		RawOccQueue,
		ClusterQueue,
		FinalFileQueue,
		FileBase.GetGlobalDelta(),
		new CDecompressor(Options),
		Options);
	Conveyor.AddNew(Manipulator);

	Conveyor.AddNew(new CFileWriter(FinalFileQueue,
		FileBase,
        NewInfrastructureOrNull(Options)));
	Conveyor.AddNew(new CClusterWriter(ClusterQueue,
		Options.Get(O_CLUSTERFILEEXT)));

	Conveyor.AddNew(new CWordSieve(RawOccQueue,
		StemmedOccQueue,
		NewCaseStandard(Options.COptions<int>::Get(O_EARLYCASECONV)),
		Options));

	Manipulator = new COccSorter(NULL,
		StemmedOccQueue,
		OccBlockQueue,
		Options.COptions<int>::Get(O_OCCBLOCKSIZE),
		Options.COptions<int>::Get(O_OCCCOLSIZE),
		NewStemmer(Options));
	Conveyor.AddNew(Manipulator);

	Reporter = new CTextReporter(Options.Get(T_INDEXERFORMAT_DDN),
		Options.COptions<int>::Get(O_INDEXCOUNTER));
	Manipulator = new CIndexer(Reporter,
		OccBlockQueue,
		StopList,
		*IndexStorage);
	Conveyor.AddNew(Manipulator);

	do
	{	ret = Conveyor.Spin(0);
	}	while (!ret);
}

CInfrastructure *NewInfrastructureOrNull(CCruncherOptions &Options)
{
    CStringListValue List = Options.COptions<CStringListValue>::Get(O_DOCPARTS);
    if (!List.GetCount())
        return NULL;
    return new CInfrastructure(IT_CREATE | IT_UPDATE, Options);
}

