#include "myfndt.h"
#include "filewrit.h"
#include "filebase.h"
#include "subdesc.h"
#include "infra.h"

CFileWriter::CFileWriter(CInQueue &aInQueue,
	CFileBase &aFileBase,
    CInfrastructure *InInfrastructure):
	CSink<CProcessedFile>(NULL,
		aInQueue),
	FileBase(aFileBase),
    Infrastructure(InInfrastructure),
	ProcessedFile(NULL)
{
}

CFileWriter::~CFileWriter()
{
	delete ProcessedFile;
    delete Infrastructure;
}

short CFileWriter::IsFilled()
{
	return ProcessedFile != NULL;
}

void CFileWriter::SetIn(CProcessedFile *InFile)
{
	Assert(!ProcessedFile);
	ProcessedFile = InFile;
}

void CFileWriter::Step()
{	CFileBase::SFileRecord Rec;
	ulong Num;

	Assert(ProcessedFile != NULL);

	Rec.WordCount = ProcessedFile->GetWordNum();
//    printf("CFileWriter::Step: Rec.WordCount == %d\n", Rec.WordCount);
	if (Rec.WordCount > 0)
	{    Rec.FirstWord = ProcessedFile->GetGlobalDelta();   
		short KnownRet = FileBase.Add(*ProcessedFile, Rec, Num);
		Assert(!KnownRet);
	}

    CSubDesc *Sub = ProcessedFile->YankSubDesc();
    if (Sub)
    {   Assert(Infrastructure != NULL);
        Infrastructure->AddSubdocDesc(Num, Sub);
    }

	delete ProcessedFile;
	ProcessedFile = NULL;
}

void CFileWriter::CleanUp()
{
	CSink<CProcessedFile>::CleanUp();
	FileBase.StoreAll();
    if (Infrastructure)
        Infrastructure->StoreAll();
}


