#include "myfndt.h"
#include "occbuild.h"

const int COccBuilder::MarginThreshold = 4;
const int COccBuilder::AQuotient = 2;
const double COccBuilder::BQuotient = 0.5;
	/* profile 13. 9.: introducing margin cut time to process
		1 file, 780 884b from 100 min. to 45 min. while indexes
		_shrank_ from 1 658 049b to 1379 349b */
const int COccBuilder::SettledThreshold = 500;
	/* profile 20. 9.: 1 file, 780 884b:
			1000 ... 40 min
			 100 ... 45 min
			  10 ... 55 min
		with optimizations elsewhere:
			 500 ... 35 min
				0 ... 50 min */

COccBuilder::COccBuilder(CStorage *aStorage, long aTicket):
	COccDatum(aStorage, aTicket)
{
}

COccBuilder::~COccBuilder()
{
}

long COccBuilder::Finish()
{
	return GetTicket();
		// always ready
}

void COccBuilder::Add(ulong *OccArray, int Count)
{	long NewTicket;

	if (GetTicket() == 0)
		NewTicket = AddNew(OccArray, Count);
	else
		NewTicket = AddEx(OccArray, Count);
	SetTicket(NewTicket);
}

long COccBuilder::AddNew(ulong *OccArray, int Count)
{
	Assert(Count > 0);
	if (Count == 1)
		return OccArray[0] | 0x80000000l;

	CConductorType &Conductor = GetConductor();
	long Ticket = Conductor.Add(sizeof(long) + Count * sizeof(ulong));
	long Total = Count;
	Conductor.Put(Ticket, 0, &Total, sizeof(long));
	Conductor.Put(Ticket,
		sizeof(long),
		OccArray,
		(short)(Count * sizeof(ulong)));
	return Ticket;
}

long COccBuilder::AddEx(ulong *OccArray, int Count)
{	long Head;

	long OldTicket = GetTicket();
	if (OldTicket & 0x80000000l)
	{	OldTicket &= ~0x80000000l;
		ulong OldTicketRepre = OldTicket;
		return AddNewTicket(&OldTicketRepre,
			1,
			OccArray,
			Count);
	}

	CConductorType &Conductor = GetConductor();
	Conductor.Get(OldTicket, 0, &Head, sizeof(long));
	if (!(Head & 0x80000000l))
	{	if (!SettledThreshold || Head < SettledThreshold)
			return AddOnTip(OldTicket,
				Head,
				OccArray,
				Count);
		else
		{	BuildSecond(OldTicket,
				Head,
				OccArray,
				Count);
			return OldTicket;
		}
	}

	BuildSegmented(OldTicket,
		Head & ~0x80000000l,
		OccArray,
		Count);
	return OldTicket;
}

long COccBuilder::AddNewTicket(ulong *Occ1,
	int Count1,
	ulong *Occ2,
	int Count2)
{
	long Count = Count1 + Count2;
	Assert(Count > 0);
	long NewLen = (1 + Count) * sizeof(ulong);
	CConductorType &Conductor = GetConductor();
	long NewTicket = Conductor.Add(NewLen);
	Conductor.Put(NewTicket, 0, &Count, sizeof(ulong));
	if (Occ1)
		Conductor.Put(NewTicket,
			sizeof(ulong),
			Occ1,
			Count1 * sizeof(ulong));
	if (Occ2)
		Conductor.Put(NewTicket,
			(1 + Count1) * sizeof(ulong),
			Occ2,
			Count2 * sizeof(ulong));
	return NewTicket;
}

long COccBuilder::AddOnTip(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	long Margin;

	CConductorType &Conductor = GetConductor();
	ulong NewCount = OldCount + Count;
	long NewSize = (1 + NewCount) * sizeof(ulong);
	long NewTicket = OldTicket;

	if (OldCount >= MarginThreshold)
		Margin = static_cast<long>((AQuotient + BQuotient * 
		        OldCount) * sizeof(ulong));
	else
		Margin = 0;
	if (Margin > 0)
		Margin = RoundMargin(NewSize, Margin);

	long RealSize = Conductor.SoftGrow(NewTicket, NewSize + Margin);
	if (RealSize < NewSize)
		NewTicket = Conductor.Resize(OldTicket, NewSize + Margin);
	Conductor.Put(NewTicket, 0, &NewCount, sizeof(ulong));
	Conductor.Put(NewTicket,
		(1 + OldCount) * sizeof(ulong),
		OccArray,
		Count * sizeof(ulong));

	return NewTicket;
}

void COccBuilder::BuildSecond(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	STail MovedOcc;
	STail NewTail;

	//	we need sizeof(STail) % sizeof(ulong) == 0
	int MovedCount = sizeof(STail) / sizeof(ulong);
	Assert(OldCount >= MovedCount);
	CConductorType &Conductor = GetConductor();
	ulong *MovedOccAlias = reinterpret_cast<ulong *>(&MovedOcc);
	MakeRoom(OldTicket,
		OldCount,
		MovedOccAlias,
		MovedCount,
		OccArray,
		Count);
	if (!MovedCount && !Count)
		return;

	long TipTicket = AddNewTicket(MovedCount ? MovedOccAlias : 0,
		MovedCount,
		OccArray,
		Count);
	NewTail.InnerFirst = NewTail.InnerLast = NewTail.Tip = TipTicket;
	long NewCount = OldCount - MovedCount;
	NewCount |= 0x80000000l;
	Conductor.Put(OldTicket,
		0,
		&NewCount,
		sizeof(long));
	Conductor.Put(OldTicket,
		(1 + OldCount - MovedCount) * sizeof(ulong),
		&NewTail,
		sizeof(STail));
}

void COccBuilder::BuildSegmented(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	STail Tail;
	STail OldTail;
	long TipCount;

	CConductorType &Conductor = GetConductor();
	Conductor.Get(OldTicket,
		(1 + OldCount) * sizeof(ulong),
		&Tail,
		sizeof(STail));
	memcpy(&OldTail, &Tail, sizeof(STail));
	Conductor.Get(Tail.Tip, 0, &TipCount, sizeof(long));
	if (!SettledThreshold || TipCount < SettledThreshold)
		BuildExSegment(Tail, TipCount, OccArray, Count);
	else
		BuildNewSegment(Tail, OccArray, Count);
	if (Tail != OldTail)
		Conductor.Put(OldTicket,
			(1 + OldCount) * sizeof(ulong),
			&Tail,
			sizeof(STail));
}

void COccBuilder::BuildExSegment(STail &Tail,
	long TipCount,
	ulong *OccArray,
	int Count)
{
	long NewTip = AddOnTip(Tail.Tip, TipCount, OccArray, Count);
	if (Tail.InnerFirst == Tail.Tip)
		Tail.InnerFirst = NewTip;
	if (Tail.InnerLast == Tail.Tip)
		Tail.InnerLast = NewTip;
	Tail.Tip =NewTip;
}

void COccBuilder::BuildNewSegment(STail &Tail,
	ulong *OccArray,
	int Count)
{
	ulong OldTipMoved = GetOldTipMoved(Tail.Tip);
	long NewTip = AddNewTicket(&OldTipMoved, 1, OccArray, Count);
	SetOldInnerLastNext(Tail.InnerLast,	Tail.Tip);
	Tail.InnerLast = Tail.Tip;
	Tail.Tip = NewTip;
}

ulong COccBuilder::GetOldTipMoved(long OldTipTicket)
{	long OldTipCount;
	ulong Moved;

	CConductorType &Conductor = GetConductor();
	Conductor.Get(OldTipTicket, 0, &OldTipCount, sizeof(long));
	Assert(OldTipCount > 1);
	Conductor.Get(OldTipTicket,
		OldTipCount * sizeof(ulong),
			// (1 + OldTipCount - 1) * sizeof(ulong)
		&Moved,
		sizeof(ulong));
	OldTipCount--;
	Conductor.Put(OldTipTicket, 0, &OldTipCount, sizeof(long));
	// now there is free place at the end of OldTipTicket;
	// it will be filled by SetOldInnerLastNext
	return Moved;
}

void COccBuilder::SetOldInnerLastNext(long InnerLastTicket,
		long NextTicket)
{	long InnerLastCount;

	CConductorType &Conductor = GetConductor();
	Conductor.Get(InnerLastTicket, 0, &InnerLastCount, sizeof(long));
	Conductor.Put(InnerLastTicket,
		(1 + InnerLastCount) * sizeof(ulong),
		&NextTicket,
		sizeof(long));
}

long COccBuilder::RoundMargin(long Size, long OldMargin)
{
	long Num = 128;
	while (Num > 1)
	{	if (OldMargin > Num)
			return OldMargin + (Size + OldMargin) % Num;
		Num /= 2;
	}
	return OldMargin;
}

void COccBuilder::MakeRoom(long CurTicket,
	long CurCount,
	ulong *DestBuf,
	int &DestBufCount,
	ulong *&SrcPadding,
	int &SrcPaddingCount)
{
	CConductorType &Conductor = GetConductor();
	long MaxSize = 1 + CurCount;
	if (SrcPaddingCount > DestBufCount)
		MaxSize += SrcPaddingCount;
	else
		MaxSize += DestBufCount;
	MaxSize *= sizeof(long);
	long RealSize = Conductor.SoftGrow(CurTicket, MaxSize);

	if (RealSize >= (1 + CurCount + SrcPaddingCount) *
			sizeof(long))
	{	// put in whole padding
		long NewCount = CurCount + SrcPaddingCount;
		Conductor.Put(CurTicket,
			0,
			&NewCount,
			sizeof(long));
		Conductor.Put(CurTicket,
			(1 + CurCount) * sizeof(ulong),
			SrcPadding,
			SrcPaddingCount * sizeof(ulong));
		DestBufCount = SrcPaddingCount = 0;
		SrcPadding = NULL;
	}
	else if (RealSize >= (1 + CurCount + DestBufCount) *
			sizeof(long))
	{	// put in some padding and leave room
		long InCount = (RealSize / sizeof(long)) -
				(1 + CurCount + DestBufCount);
		Assert(InCount < SrcPaddingCount);
		if (InCount > 0)
		{	Conductor.Put(CurTicket,
				(1 + CurCount) * sizeof(ulong),
				SrcPadding,
				static_cast<int>(InCount) * sizeof(ulong));
			SrcPaddingCount -= static_cast<int>(InCount);
			SrcPadding += static_cast<int>(InCount);
		}
		DestBufCount = 0;
	}
	else
	{	// make room
		long OutCount = (1 + CurCount + DestBufCount) -
				(RealSize / sizeof(long));
		Assert(0 < OutCount && OutCount <= DestBufCount);
		Conductor.Get(CurTicket,
			(1 + CurCount - OutCount) * sizeof(ulong),
			DestBuf,
			static_cast<int>(OutCount) * sizeof(ulong));
		DestBufCount = static_cast<int>(OutCount);
	}
}

