#if !defined(__occmap_h)
#define __occmap_h

#include "arrays.h"
#include "mape.h"

#include "cocc.h"
#include "socc.h"

class CAbstractStemmer;

//#define TWOWAYBLOCK
/* profile 19Jul98: flipping the direction of iteration does *not*
speed the indexing up; 4 cycles adding to 67 331 534 bytes:
oneway		twoway
33 min		34 min */

class COccBlock : public DMap<DString,  COccCol>
{
public:
    COccBlock(int aOccColMax);
    virtual ~COccBlock();

    void SetIn(CSimpleOcc *InSimpleOcc,
        CAbstractStemmer *Stemmer);
	// responsible for deleting SimpleOcc

    int IsEmpty() const;
    int GetCount() const;
    int HasBigCol() const { return BigCol; }

#if !defined(TWOWAYBLOCK)
    CComposedOcc *NewComposedOcc();
#else
    virtual CComposedOcc *NewComposedOcc() = 0;
#endif

#if !defined(TWOWAYBLOCK)
private:
#else
protected:
#endif
    const int OccColMax;
    int Emptying;
    int PhaseCount;
#if !defined(TWOWAYBLOCK)
    DMap<DString, COccCol>::iterator Itr;
#endif
    int BigCol;
    int AliasCount;

    void AddToOld(COccCol &OldCol, ulong Data);
};

#if defined(TWOWAYBLOCK)

class COccBlockForward : public COccBlock
{
public:
    COccBlockForward(int aSize):COccBlock(aSize), Itr(begin()) { }

    virtual CComposedOcc *NewComposedOcc();

private:
    DMap<DString, COccCol>::iterator Itr;
};

class COccBlockBackward : public COccBlock
{
public:
    COccBlockBackward(int aSize):COccBlock(aSize), Itr(rbegin()) { }

    virtual CComposedOcc *NewComposedOcc();

private:
    DMap<DString, COccCol>::reverse_iterator Itr;
};

#endif // TWOWAYBLOCK

#endif // __occmap_h
