#include "myfndt.h"
#include "occsort.h"
#include "abststem.h"

COccSorter::COccSorter(CReporter *aReporter,
    CInQueue &aInQueue,
    COutQueue &aOutQueue,
    int aBlockSize,
	int aColSize,
    CAbstractStemmer *InStem):
    CTransmission<CSimpleOcc, COccBlock>(aReporter, aInQueue, aOutQueue),
    OccBlock(NULL), 
    BlockSize(aBlockSize),
    ColSize(aColSize),
#if defined(TWOWAYBLOCK)
    Forward(true),
#endif
    Stem(InStem)
{
}

COccSorter::~COccSorter()
{
    delete Stem;
    delete OccBlock;
}

short COccSorter::IsFilled()
{
	if (!OccBlock)
		return 0;

     if (OccBlock->HasBigCol())
	     return 1;

	int Count = OccBlock->GetCount();
	return IsSourceDone() ? (Count > 0) : (Count >= BlockSize);
}

short COccSorter::GetPerc()
{
	if (!OccBlock)
		return 0;
	return (short)((100 * OccBlock->GetCount()) / BlockSize);
}

void COccSorter::SetIn(CSimpleOcc *InOcc)
{
	CObjPtr<CSimpleOcc> Occ = InOcc;
	if (!OccBlock)
    {
#if !defined(TWOWAYBLOCK)
        OccBlock = new COccBlock(ColSize);
#else
        if (Forward)
		{   OccBlock = new COccBlockForward(ColSize);
            Forward = false;
        }
        else
        {   OccBlock = new COccBlockBackward(ColSize);
            Forward = true;
        }
#endif
    }
	OccBlock->SetIn(Occ.Drop(), Stem);
}

COccBlock *COccSorter::NewOut()
{
	Assert(OccBlock != NULL);
	COccBlock *NewOccBlock = OccBlock;
	OccBlock = NULL;
	return NewOccBlock;
}

