#include "myfndt.h"
#include "rdisp.h"
#include "crunopt.h"
#include "decompr.h"
#include "preader.h"
#include "watch.h"
#include "subdesc.h"

#include "ocross.h"
#include "oreader.h"

//#define DEBUG_wrap

#if defined(DEBUG_wrap)
#include <stdio.h>
#endif

CReaderDispatch::CReaderDispatch(CReporter *aReporter,
	CInQueue &aInQueue,
	COutQueue &aOutQueue,
	COutQueue1 &aOutQueue1,
	COutQueue2 &aOutQueue2,
	ulong StartWordNum,
	CDecompressor *aDecompressor,
	CCruncherOptions &Options):
	CBranch<CGlobalFile, CSimpleOcc, CCluster, CProcessedFile>(aReporter,
		aInQueue,
		aOutQueue,
		aOutQueue1,
		aOutQueue2),
	Reader(new CReader(Options.Get(O_SEPARATORS), new CWatch(Options))),
	CurFile(NULL),
	PreparedOcc(NULL),
	PreparedCluster(NULL),
	PreparedFile(NULL),
	NextLine(0),
	WordNum(StartWordNum),
	GlobalDelta(0),
	DivisionOffs(0),
	ClusterCounter(0),
	ClusterCounterMax(Options.COptions<int>::Get(O_NRMCLUSTERCOUNTER)),
	Decompressor(aDecompressor),
	ReserveCounter(0),
	ReserveCounterMax(Options.COptions<int>::Get(O_RESERVECOUNTER)),
    SubDesc(NULL)
{
}

CReaderDispatch::~CReaderDispatch()
{
    delete SubDesc;
	delete Decompressor;
	delete PreparedFile;
	delete PreparedCluster;
	delete PreparedOcc;
	delete CurFile;
	delete Reader;
}

short CReaderDispatch::IsFilled()
{
	return PreparedOcc || PreparedCluster || PreparedFile;
}

void CReaderDispatch::SetIn(CGlobalFile *GlobalFile)
{
	Assert(!CurFile);
	CurFile = GlobalFile;
	DString ContentFile = Decompressor->GrabContentFile(CurFile->GetName());
	Reader->SetIn(ContentFile, 0);
	DivisionOffs = 0;
	DocBegin();
	FillPrepOcc();
}

void CReaderDispatch::NewOut(CSimpleOcc **SimpleOcc,
	CCluster **Cluster,
	CProcessedFile **ProcessedFile)
{
	Assert(PreparedOcc || PreparedFile);
	*SimpleOcc = PreparedOcc;
	PreparedOcc = NULL;
	*Cluster = PreparedCluster;
	PreparedCluster = NULL;
	*ProcessedFile = PreparedFile;
	PreparedFile = NULL;
	if (CurFile)
		FillPrepOcc(); // else reader cannot be filled again
}

short CReaderDispatch::GetPerc()
{
	return Reader->GetPerc();
}

void CReaderDispatch::FillPrepOcc()
{	long Offs;
	CReader::EShiftRes ShiftRes;

	Assert(!PreparedOcc && !PreparedCluster && !PreparedFile);

	short PreparedLine = NextLine; // for PreparedCluster
	ShiftRes = Reader->SkipWhite(Offs);
	switch (ShiftRes)
    {   case CReader::over_newline:
			PreparedLine = 1;
            break;
        case CReader::at_docend:
            FillPreparedDoc();
            DivisionOffs = Offs;
            DocBegin();
            return;
        case CReader::at_eof:
            FillPreparedDoc();
            Decompressor->DropContentFile(CurFile->GetName());
            delete CurFile;
            CurFile = NULL;
            return; // PreparedOcc remains NULL
        case CReader::at_sign:
            AddSubdoc();
            break;
        default:
            break;
	}

	if (HandleWordEndsDoc(Offs))
	     return;

	// set PreparedCluster
	if (ReserveCounter >= ReserveCounterMax ||
	    PreparedLine && !(--ClusterCounter))
	{   PreparedCluster = new CCluster(PreparedOcc->GetData(),
		   Offs);
		   ClusterCounter = ClusterCounterMax;
		   ReserveCounter = 0;
	}
}

void CReaderDispatch::DocBegin()
{
	ClusterCounter = 1;
	GlobalDelta = WordNum;
	NextLine = 1;
}

int CReaderDispatch::HandleWordEndsDoc(long Offs)
{   DString WordStr;

	CReader::EShiftRes ShiftRes = Reader->GetWord(WordStr);
	NextLine = ShiftRes != CReader::ok;
	if (!WordStr.IsEmpty())
	{	PreparedOcc = new CSimpleOcc(WordStr,
			WordNum,
			WordNum - GlobalDelta);
#if defined(DEBUG_wrap)
        printf("%s ", WordStr.c_str());
#endif
		++WordNum;
		++ReserveCounter;
		return 0;
	}
	else
	{   Assert(ShiftRes == CReader::at_eof || ShiftRes == CReader::at_docend);
        FillPreparedDoc();
		DivisionOffs = Offs;
		++WordNum;
		++ReserveCounter;
		DocBegin();
		return 1;
	}
}

void CReaderDispatch::FillPreparedDoc()
{  
    Assert(CurFile != NULL && PreparedFile == NULL);
	PreparedFile = new CProcessedFile(CurFile->GetName(),
	    DivisionOffs,
		WordNum - GlobalDelta,
		GlobalDelta,
        SubDesc);
    SubDesc = NULL;
}

void CReaderDispatch::AddSubdoc()
{
#if defined(DEBUG_wrap)
    printf("\nafter %d words starts subdoc %d\n", 
        WordNum - GlobalDelta, 
        Reader->GetCurrentSubdoc());
#endif
    if (!SubDesc)
        SubDesc = new CSubDesc();
    SubDesc->Add(WordNum - GlobalDelta, Reader->GetCurrentSubdoc());
}
