#include "myfndt.h"
#include "sortdir.h"

const int CSortedDirectory::StackLimit = 10;
const int CSortedDirectory::StackDelta = 50;

CSortedDirectory::CSortedDirectory(const DString &aDirName, const DString &aMask):
    Mask(aMask),
    Directory(NULL),
    Files(20, 100),
    SubdirStack(new DArray<DString>(StackLimit, StackDelta)),
    CurFile(0)
{
    SubdirStack->Add(aDirName);
}

CSortedDirectory::~CSortedDirectory()
{
    delete SubdirStack;
    delete Directory;
}

DString CSortedDirectory::ReadNext()
{
    if (CurFile >= Files.GetCount())
    {   int Finished = SubInit();
	while (!Finished && Files.IsEmpty())
	    Finished = SubInit();
	if (Finished || Files.IsEmpty())
	    return DString("");
    }

    DString File = Files[CurFile].Convert();
    CurFile++;
    return File;
}

int CSortedDirectory::SubInit()
{
    if (SubdirStack->IsEmpty())
	   return 1;

    delete Directory;
    Directory = new CDirectory(SubdirStack->operator[](0), Mask);
    Files.erase(Files.begin(), Files.end());
    CurFile = 0;

    DSArray<CFileString> ImmedSubdirs(10, 50);

    DString Name;
    CDirectory::EType Type = Directory->NextFileOrDir(Name);
    while (Type != CDirectory::nomore)
    {   if (Type == CDirectory::file)
	        Files.Add(CFileString(Name));
	    else
	        ImmedSubdirs.Add(CFileString(Name));
	    Type = Directory->NextFileOrDir(Name);
    }

    CObjPtr<DArray<DString> > Stack = new DArray<DString>(SubdirStack->GetCount(),
	    StackDelta);
    DSArray<CFileString>::DIterator Itr = ImmedSubdirs.InitIterator();
    while (ImmedSubdirs.IsDereferenceable(Itr))
    {   Stack->Add((*Itr).Convert());
	    ++Itr;
    }

    if (!Stack->IsEmpty())
    {   DArray<DString>::DIterator Itr2 = SubdirStack->InitIterator();
	++Itr2; // skip used dir
	while (SubdirStack->IsDereferenceable(Itr2))
	{   Stack->Add(*Itr2);
	    ++Itr2;
	}

	delete SubdirStack;
	SubdirStack = Stack.Drop();
    }
    else
	SubdirStack->DetachByIndex(0);

    return 0;
}
