#if !defined(__transmt_h)
#define __transmt_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

#include "sink.h"
#include "lqueue.h"

template<class CInItem, class COutItem>
class CTransmission : public CSink<CInItem>
{
public:
	typedef CQueue<COutItem> COutQueue;

private:
	COutQueue &OutQueue;

public:
	CTransmission(CReporter *aReporter,
		CQueue<CInItem> &aInQueue,
		CQueue<COutItem> &aOutQueue);

	virtual ~CTransmission() { }

	virtual short Shuffle();

protected:
	virtual void SetIn(CInItem *InItem) = 0;
	virtual COutItem *NewOut() = 0;
	virtual void InformNext();
	virtual void CleanUp();

private:
	virtual void Step(); // cannot be called
};

template<class CInItem, class COutItem>
CTransmission<CInItem, COutItem>::CTransmission(CReporter *aReporter,
	CQueue<CInItem> &aInQueue,
	CQueue<COutItem> &aOutQueue):
	CSink<CInItem>(aReporter, aInQueue),
	OutQueue(aOutQueue)
{
}

template<class CInItem, class COutItem>
short CTransmission<CInItem, COutItem>::Shuffle()
{
	if (!IsFilled())
	{	ReportIn();
		return ShuffleIn();
	}
	else
	// this code should be inherited from CSource
	{	if (OutQueue.IsFull())
		{	ReportProgress(0, GetPerc());
			OutQueue.SetBlockedOnPut(this);
			return 1;
		}
		COutItem *OutItem = NewOut();
		OutQueue.Put(OutItem);
		ReportProgress(1, GetPerc());
		return 0;
	}
}

template<class CInItem, class COutItem>
void CTransmission<CInItem, COutItem>::InformNext()
{
	OutQueue.SetSourceDone();
}

template<class CInItem, class COutItem>
void CTransmission<CInItem, COutItem>::CleanUp()
{
	CSink<CInItem>::CleanUp();
	OutQueue.Wake();
}

template<class CInItem, class COutItem>
void CTransmission<CInItem, COutItem>::Step()
{
	throw CGeneralException("Step() called for CTransmission");
}

#endif

