#include "myfndt.h"
#include "wordsiv.h"
#include "punctaxe.h"
#include "abstcase.h"
#include "crunopt.h"

#include "owordsiv.h"
#include "opunct.h"

CWordSieve::CWordSieve(CInQueue &aInQueue,
	COutQueue &aOutQueue,
	CAbstractCase *aCase,
	CCruncherOptions &Options):
	CTransmission<CSimpleOcc, CSimpleOcc>(NULL,
		aInQueue,
		aOutQueue),
	Occ(NULL),
	PunctAxe(new CPunctAxe(Options.Get(O_PUNCT))),
	Case(aCase),
	IndexedLength(Options.COptions<int>::Get(O_DOCMAXLEN)),
	MaxWordLength(Options.COptions<int>::Get(O_MAXWORDLEN))
{
	Assert(Case != NULL);
}

CWordSieve::~CWordSieve()
{
	delete Case;
	delete Occ;
	delete PunctAxe;
}

short CWordSieve::IsFilled()
{
	return (short)(Occ != NULL ? 1 : 0);
}

void CWordSieve::SetIn(CSimpleOcc *InOcc)
{
	Assert(!Occ);
	Occ = InOcc;

	if (IndexedLength > 0 &&
		InOcc->GetPosInDoc() > IndexedLength)
		goto DELE;
	if (InOcc->OccWord.GetLength() > MaxWordLength)
		goto DELE;
	if (PunctAxe->AxePunct(InOcc->OccWord))
		goto DELE;

	Case->Confine(InOcc->OccWord);

	return;
DELE:
	delete Occ;
	Occ = NULL;
}

CSimpleOcc *CWordSieve::NewOut()
{
	Assert(Occ != NULL);
	CSimpleOcc *OutOcc = Occ;
	Occ = NULL;
	return OutOcc;
}


