#include "myfndt.h"
#include "answer.h"

#include <string.h>
#include <stdlib.h>

const int CAnswer::OccLimit = 15;
const int CAnswer::OccDelta = 15;
const int CAnswer::MemLimit = 300;
const int CAnswer::OutLimit = 300;

CAnswer::COcc::COcc(ulong aOcc, ushort aWordIndex):
	Occ(aOcc), WordIndex(aWordIndex)
{
}

CAnswer::COcc::COcc(const CAnswer::COcc &Src):
	Occ(Src.GetOccNum()), WordIndex(Src.GetWordIndex())
{
}

CAnswer::COcc::~COcc()
{
}

int CAnswer::COcc::operator == (const COcc &Other) const
{
	return Occ == Other.Occ && WordIndex == Other.WordIndex;
}

int CAnswer::COcc::operator < (const COcc &Other) const
{
	if (Occ != Other.Occ)
		return Occ < Other.Occ;
	return WordIndex < Other.WordIndex;
}

CAnswer::COccList::COccList(CConductor &Conductor,
	long Ticket,
	int From,
	long aDelta,
	int Count):
	OccList(NULL),
	ListCount(0),
	Delta(aDelta)
{	long TotalCount;

	Assert(Ticket != 0);
	Conductor.Get(Ticket, From, &TotalCount, sizeof(long));
	Assert(TotalCount > 0 &&
		Delta >= 0 &&
		Delta <= TotalCount);
	if (Delta == TotalCount)
		return;

	if (TotalCount - Delta < Count)
		ListCount = static_cast<int>(TotalCount - Delta);
	else
		ListCount = Count;

	OccList = (COcc *)malloc(ListCount * sizeof(COcc));
	if (!OccList)
		throw DMemoryException(ListCount * sizeof(COcc));
	Conductor.Get(Ticket,
		From + sizeof(long) + Delta * sizeof(COcc),
		OccList,
		ListCount * sizeof(COcc));
}

CAnswer::COccList::COccList(
	DISArray<COcc> &MemOcc):
	OccList(NULL),
	ListCount(0),
	Delta(0)
{
	ListCount = MemOcc.GetCount();
	Assert(ListCount > 0);
	OccList = (COcc *)malloc(ListCount * sizeof(COcc));
	if (!OccList)
		throw DMemoryException(ListCount * sizeof(COcc));
	for (int i = 0; i < ListCount; i++)
	{	COcc *Occ = MemOcc[i];
		memcpy(OccList + i, Occ, sizeof(COcc));
	}
}

CAnswer::COccList::~COccList()
{
	if (OccList)
		free(OccList);
}

#if defined(DEBUG_answer)
void CAnswer::COccList::Check()
{
    Assert(ListCount >= 0);
    for (int i = 0; i < ListCount; ++i)
	   Assert((OccList + i)->GetWordIndex() < 100);
}
#endif

void CAnswer::COccList::Subtract(int Count)
{
	Assert(Count <= ListCount);
	int NewCount = ListCount - Count;
	if (!NewCount)
	{	if (OccList)
		{	free(OccList);
			OccList = NULL;
		}
		ListCount = 0;
		return;
	}

	COcc *NewList = (COcc *)malloc(NewCount * sizeof(COcc));
	if (!NewList)
		throw DMemoryException(NewCount * sizeof(COcc));
	memcpy(NewList, OccList + Count, NewCount * sizeof(COcc));
	free(OccList);
	OccList = NewList;
	ListCount = NewCount;
	Delta += Count;
}

CAnswer::CAnswer(CStorage *aStorage, long aTicket):
	CDatum(aStorage, aTicket),
	MemOcc(OccLimit, OccDelta),
	OutOcc(NULL),
	HeadLen(0)
{
}

CAnswer::CAnswer(CStorage *aStorage, long aTicket, int aHeadLen):
	CDatum(aStorage, aTicket),
	MemOcc(OccLimit, OccDelta),
	OutOcc(NULL),
	HeadLen(aHeadLen)
{
}

CAnswer::~CAnswer()
{
	delete OutOcc;
}

long CAnswer::Finish()
{
	if (!IsDirty())
		return 0;

	Store();
	return GetTicket();
}

void CAnswer::Add(ushort WordIndex, ulong WordNum)
{
	CObjPtr<COcc> Occ = new COcc(WordNum, WordIndex);
	if (MemOcc.Find(Occ) == INT_MAX)
	{
		MemOcc.Add(Occ.Drop());
	}
	// else nothing; every occurence is stored at most once
	if (MemOcc.GetCount() > MemLimit)
		Store();
}

long CAnswer::InitIterator()
{
	if (IsDirty())
		Store();
	long Ticket = GetTicket();
	if (Ticket == 0)
		return 0;

	delete OutOcc;
	OutOcc = NULL; // necessary to handle exceptions
	OutOcc = new COccList(GetConductor(),
		Ticket,
      HeadLen,
		0,
		OutLimit);
	return 1;
}

void CAnswer::ResetIterator(long Iterator)
{
	Assert(Iterator > 0);
	Assert(MemOcc.GetCount() == 0);
	long Ticket = GetTicket();
	Assert(Ticket > 0);
	delete OutOcc;
	OutOcc = NULL; // necessary to handle exceptions
	OutOcc = new COccList(GetConductor(),
		Ticket,
      HeadLen,
		Iterator - 1,
		OutLimit);
}

int CAnswer::GetIndex(long Iterator)
{
    Assert(OutOcc != NULL &&
	   Iterator > OutOcc->Delta);
    int Index = static_cast<int>(Iterator - OutOcc->Delta - 1);
    Assert(Index < OutOcc->ListCount);
    return Index;
}

CAnswer::COcc CAnswer::Peek(long Iterator)
{
    return COcc(OutOcc->OccList[GetIndex(Iterator)]);
}

void CAnswer::Bump(long &Iterator)
{
    Iterate(GetIndex(Iterator), Iterator);
}

CAnswer::COcc CAnswer::GetHit(long &Iterator)
{
    int Index = GetIndex(Iterator);
    COcc Occ(OutOcc->OccList[Index]);
#if defined(DEBUG_answer)
    Assert(Occ.GetWordIndex() < 100);
#endif
    Iterate(Index, Iterator);
    return Occ;
}

long CAnswer::GetHitCount()
{	long HitCount;

	long Ticket = GetTicket();
	if (Ticket == 0)
		return 0;

	CConductor &Conductor = GetConductor();
	Conductor.Get(Ticket, HeadLen, &HitCount, sizeof(long));
	return HitCount;
}

void CAnswer::Fill(CAnswer *SrcAnswer)
{
	long Itr = SrcAnswer->InitIterator();
	while (Itr)
	{	COcc Occ = SrcAnswer->GetHit(Itr);
		Add(Occ.GetWordIndex(), Occ.GetOccNum());
	}
}

CRelevance *CAnswer::GetRelevance()
{
	return NULL;
}

int CAnswer::IsDirty()
{
	return MemOcc.GetCount();
}

void CAnswer::Store()
{	long NewTicket;

	delete OutOcc;
	OutOcc = NULL;
	long OldTicket = GetTicket();

	if (OldTicket == 0)
		NewTicket = SaveMem();
	else
		NewTicket = MergeMem(OldTicket);

	int Count = MemOcc.GetCount();
	for (int i = Count - 1; i >= 0; i--)
	{	COcc *Doomed = MemOcc[i];
		MemOcc.Detach(Doomed);
		delete Doomed;
	}

	SetTicket(NewTicket);
}

long CAnswer::SaveMem()
{
	CConductor &Conductor = GetConductor();
	int Count = MemOcc.GetCount();
	Assert(Count > 0);
	long NewTicket = Conductor.Add(HeadLen + sizeof(long) +
			Count * sizeof(COcc));

	long Total = Count;
	Conductor.Put(NewTicket, HeadLen, &Total, sizeof(long));

	COccList MemList(MemOcc);
#if defined(DEBUG_answer)
     MemList.Check();
#endif
	Conductor.Put(NewTicket,
		HeadLen + sizeof(long),
		MemList.OccList,
		Count * sizeof(COcc));

	return NewTicket;
}

long CAnswer::MergeMem(long OldTicket)
{
    if (MemOcc.GetCount() == 0)
	   // can happen in CRelAnswer
	   return OldTicket;

    CConductor &Conductor = GetConductor();
    long NewTicket = BeginMerge(OldTicket);

    long OldDelta = 0;
    long From = HeadLen + sizeof(long); // current write position in new ticket
    COccList NewOcc(MemOcc);
    Assert(NewOcc.ListCount > 0);
    do
    {   MergeCycle(NewTicket,
		  From,
		  &NewOcc,
		  OldTicket,
		  OldDelta);
    }   while (NewOcc.ListCount > 0);

    EndMerge(NewTicket, From, OldTicket, OldDelta); // deletes old ticket
    return NewTicket;
/*    while (1)
    {   COccList OldOccTail(Conductor,
		  OldTicket,
		  HeadLen,
		  OldDelta,
		  OutLimit);
	   if (!OldOccTail.ListCount)
	   {   Conductor.Del(OldTicket);
		  return NewTicket;
	   }

	   int Len = OldOccTail.ListCount * sizeof(COcc);
	   Conductor.Put(NewTicket,
		  From,
		  LastOldOcc.OccList,
		  Len);

	   From += Len;
	   OldDelta += OldOccTail.ListCount;
    }*/
}

void CAnswer::EndMerge(long NewTicket, long From, long OldTicket, long OldDelta)
{   long OldTotal;

    CConductor &Conductor = GetConductor();
    Conductor.Get(OldTicket, HeadLen, &OldTotal, sizeof(long));
        // maybe it's more effective to pass total as a parameter from
        // BeginMerge than to read it from disk

    Assert(OldTotal > 0 &&
	   OldDelta >= 0 &&
	   OldDelta <= OldTotal);
    if (OldDelta < OldTotal)
	   Conductor.Copy(NewTicket,
		From,
		OldTicket,
		HeadLen + sizeof(long) + OldDelta * sizeof(COcc),
	        (OldTotal - OldDelta) * sizeof(COcc));

    Conductor.Del(OldTicket);
}

long CAnswer::BeginMerge(long OldTicket)
{  long Total;

	CConductor &Conductor = GetConductor();
	Conductor.Get(OldTicket, HeadLen, &Total, sizeof(long));
	Total += MemOcc.GetCount();
	long NewTicket = Conductor.Add(HeadLen + sizeof(long) +
			Total * sizeof(COcc));
	Conductor.Put(NewTicket, HeadLen, &Total, sizeof(long));
	return NewTicket;
}

void CAnswer::MergeCycle(long NewTicket,
    long &From,
    COccList *NewOcc,
    long OldTicket,
    long &OldDelta)
{
    CConductor &Conductor = GetConductor();
    COccList OldOcc(Conductor,
	   OldTicket,
	   HeadLen,
	   OldDelta,
	   OutLimit);
    if (OldOcc.ListCount)
    {   OldDelta += OldOcc.ListCount;
	   MergeNonEmpty(NewTicket, From, NewOcc, &OldOcc);
    }
    else // new occurences are the biggest so far
    {   // append them to the end
	   int Len = NewOcc->ListCount * sizeof(COcc);
#if defined(DEBUG_answer)
	   NewOcc->Check();
#endif
	   Conductor.Put(NewTicket,
		  From,
		  NewOcc->OccList,
		  Len);
	   From += Len;
	   NewOcc->Subtract(NewOcc->ListCount);
    }
}

void CAnswer::MergeNonEmpty(long DestTicket,
	long &From,
	COccList *ShortList,
	COccList *ListSequence)
{
	int SeqCount = ListSequence->ListCount;
	int ShortCount = ShortList->ListCount;
	Assert(SeqCount > 0 && ShortCount > 0);
	DISArray<COcc> MergedArr(SeqCount + ShortCount, 0);
	for (int i = 0; i < SeqCount; i++)
		MergedArr.Add(new COcc(ListSequence->OccList[i]));
	int AddIndex = 0;
	while (AddIndex < ShortCount &&
		ShortList->OccList[AddIndex] < ListSequence->OccList[
				SeqCount - 1])
	{    int TestCount = MergedArr.GetCount();
		MergedArr.Add(new COcc(ShortList->OccList[AddIndex]));
		Assert(TestCount < MergedArr.GetCount());
			// all occurences must be unique; see BeginMerge()
		AddIndex++;
	}

	COccList MergedList(MergedArr);
	int Len = MergedList.ListCount * sizeof(COcc);
	CConductor &Conductor = GetConductor();
#if defined(DEBUG_answer)
     MergedList.Check();
#endif
	Conductor.Put(DestTicket,
		From,
		MergedList.OccList,
		Len);
	From += Len;
	ShortList->Subtract(AddIndex);
}

void CAnswer::Iterate(int Index, long &Iterator)
{
	if (Index < OutOcc->ListCount - 1)
		Iterator++;
	else
	{	long NewDelta = OutOcc->Delta + OutOcc->ListCount;
		delete OutOcc;
		OutOcc = NULL; // necessary (for exceptions)
		OutOcc = new COccList(GetConductor(),
			GetTicket(),
			HeadLen,
			NewDelta,
			OutLimit);
		if (OutOcc->ListCount == 0)
		{	delete OutOcc;
			OutOcc = NULL;
			Iterator = 0;
		}
		else
			Iterator = OutOcc->Delta + 1;
	}
}

#if defined(DEBUG_answer)
void CAnswer::Check()
{
    long Itr = InitIterator();
    if (!Itr)
	   return;

    COcc FirstHit = GetHit(Itr);
    ulong Occ = FirstHit.GetOccNum();
    while (Itr)
    {   COcc Hit = GetHit(Itr);
	   Assert(Occ <= Hit.GetOccNum());
	   Occ = Hit.GetOccNum();
    }
}
#endif

