#if !defined(__answer_h)
#define __answer_h

#include "mconduct.h"
#include "arrays.h"

//#define DEBUG_answer

class CRelevance;

class CAnswer : public CDatum
{
// consts & types
private:
	static const int OccLimit;
	static const int OccDelta;
	static const int MemLimit; // see Add()
	static const int OutLimit; // used in constructor of COccList

public:
	class COcc // stored on disk - do not change layout!
	{
	private:
		ulong Occ;
		ushort WordIndex;

	public:
		COcc(ulong aOcc, ushort aWordIndex);
		COcc(const COcc &Src);
		~COcc(); // not virtual

		ushort GetWordIndex() const { return WordIndex; }
		ulong GetOccNum() const { return Occ; }

		int operator == (const COcc &Other) const;
		int operator < (const COcc &Other) const;
	};

#if defined(DEBUG_answer)
    void Check();
#endif

private:
	class COccList
	{
	public:
		COcc *OccList;
		int ListCount;
		long Delta;

		COccList(CConductor &Conductor,
			long Ticket,
			int From,
			long aDelta,
			int Count);
		COccList(DISArray<COcc> &MemOcc);
		~COccList(); // not virtual

		void Subtract(int Count);

#if defined(DEBUG_answer)
	     void Check();
#endif
	};

// members
private:
	DISArray<COcc> MemOcc;
		// must be sorted
	COccList *OutOcc; // mostly null
	int HeadLen;

public:
	CAnswer(CStorage *aStorage, long aTicket);
protected:
	CAnswer(CStorage *aStorage, long aTicket, int aHeadLen);
public:
	virtual ~CAnswer();

	virtual long Finish();

	void Add(ushort WordIndex, ulong WordNum);
		// when MemOcc.GetItemsInContainer gets over MemLimit,
		// MemOcc is saved

	long InitIterator();
		// returns 0 => answer is empty
	void ResetIterator(long Iterator);
		// resets OutOcc where it once has been

/*	ulong GetOccNum(long &Iterator);
	ushort GetWordIndex(long &Iterator); implemented in CMaskingAnswer */
	COcc GetHit(long &Iterator); // iterates

	long GetHitCount();

     COcc Peek(long Iterator); // does not iterate
	void Bump(long &Iterator);

	void Fill(CAnswer *SrcAnswer);

	virtual CRelevance *GetRelevance();
		// always returns 0
		// should not be here at all

protected:
	virtual int IsDirty();
	virtual void Store();

private:
	long SaveMem();
	long MergeMem(long OldTicket);
	long BeginMerge(long OldTicket);
	void MergeCycle(long NewTicket,
		long &From, // in & out param
		COccList *NewOcc,
		long OldTicket,
		long &OldDelta);
	void MergeNonEmpty(long DestTicket,
		long &From, // in & out param
		COccList *ShortList, // in & out param
		COccList *ListSequence); // in param
     void EndMerge(long NewTicket, long From, long OldTicket, long OldDelta);

	void Iterate(int Index, long &Iterator);
     int GetIndex(long Iterator);
};

#endif

