#include "myfndt.h"
#include "docmask.h"
#include "filter.h"
#include "subdesc.h"
#include <algo.h>

//#define DEBUG_holes

CDocMask::CDocMask(CSubdocFilter &aFilter, CSubDesc &aSubDesc, long DocWordCount):
    Filter(aFilter),
    SubDesc(aSubDesc),
    VisibleCount(0)
{
    Base.reserve(SubDesc.Stairs.size());
    Init(DocWordCount);
}

void CDocMask::Init(long DocWordCount)
{
    Assert(VisibleCount == 0);  
    long PreviousTotal = 0;
    int n = SubDesc.Stairs.size();
    Assert(n > 0);
    for (int i = 0; i <= n; ++i)
    {   short CurrentId;
        long PreviousBase;
        if (i == 0)
        {   CurrentId = 0;
            PreviousBase = 0;
        }
        else
        {   CurrentId = SubDesc.Ids[i - 1];
            PreviousBase = Base[i - 1];
        }
        long CurrentTotal = i == n ? DocWordCount : SubDesc.Stairs[i];
        Assert(CurrentTotal >= PreviousTotal);

        long Delta = Filter.Has(CurrentId) ? CurrentTotal - PreviousTotal : 0;
        VisibleCount += Delta;
        if (i < n)
            Base[i] = PreviousBase + Delta;

        PreviousTotal = CurrentTotal;
    }
}

int CDocMask::Has(long LocalOcc)
{
#if defined(DEBUG_holes)
    printf("CDocMask::Has(%d)\n", LocalOcc);
    int n = SubDesc.Stairs.size();
    printf("%d parts:\n", n);
    for (int k = 0; k < n; ++k)
        printf("\tId=%d\tStair=%d\tBase=%d\n", 
            SubDesc.Ids[k],
            SubDesc.Stairs[k],
            Base[k]);
#endif
    vector<long>::iterator i = upper_bound(SubDesc.Stairs.begin(),
        SubDesc.Stairs.end(),
        LocalOcc);
    int j = i - SubDesc.Stairs.begin();
    short CurrentId = j == 0 ? 0 : SubDesc.Ids[j - 1];
#if defined(DEBUG_holes)
    printf("the answer is %s\n", Filter.Has(CurrentId) ? "yes" : "no");
#endif
    return Filter.Has(CurrentId);
}

long CDocMask::Constrict(long LocalOcc)
{
#if defined(DEBUG_holes)
    printf("CDocMask::Constrict(%d)\n", LocalOcc);
#endif
    Assert(Has(LocalOcc));
#if defined(DEBUG_holes)
    int n = SubDesc.Stairs.size();
    printf("%d parts:\n", n);
    for (int k = 0; k < n; ++k)
        printf("\tId=%d\tStair=%d\tBase=%d\n", 
            SubDesc.Ids[k],
            SubDesc.Stairs[k],
            Base[k]);
#endif

    vector<long>::iterator i = upper_bound(SubDesc.Stairs.begin(),
        SubDesc.Stairs.end(),
        LocalOcc);
    long CurrentBase;
    long CurrentStair;
    int j = i - SubDesc.Stairs.begin();
#if defined(DEBUG_holes)
    printf("upper_bound of %d is index %d\n", LocalOcc, j);
#endif
    if (!j)
    {   CurrentBase = 0;
        CurrentStair = 0;
    }
    else
    {   --j;
        CurrentBase = Base[j];
        CurrentStair = SubDesc.Stairs[j];
    }
#if defined(DEBUG_holes)
    printf("CurrentBase=%d,\tCurrentStair=%d\tresult=%d\n", CurrentBase, 
            CurrentStair, CurrentBase + (LocalOcc - CurrentStair));
#endif

    return CurrentBase + (LocalOcc - CurrentStair);
}

