#include "myfndt.h"
#include "formdisp.h"
#include <stdio.h>
#include "mirror.h"
#include "result.h"
#include "grazeopt.h"
#include "docum.h"
#include "messform.h"
#include "txtcolor.h"
#include "relevanc.h"
#include "clot.h"
#include "clotcurs.h"
#include "passover.h"
#include "puparse.h"
#include "dpnames.h"
#include "filecmp.h"
#include "rxw.h"

#include "odisptxt.h"
#include "ocolors.h"
#include "omirror.h"
#include "ooutform.h"
#include "oreader.h"
#include "opunct.h"
#include "osubdoc.h"

CFormDisplay::CFormDisplay(CMirror &aMirror,
    CResult &aResult,
    CGrazerOptions &aOptions):
    Mirror(aMirror),
    Result(aResult),
    Options(aOptions),
    OutFile(NULL),
    PunctParse(NULL),
    TextColor(GetColorEscape(aOptions.COptions<int>::Get(O_TEXTCOLOR))),
    OccColor(GetColorEscape(aOptions.COptions<int>::Get(O_OCCCOLOR))),
    HiddenHead(aOptions.Get(O_HIDDENFILEHEAD))
{
    DString OutFileName = Options.Get(O_OUTFILE);
    if (!OutFileName.IsEmpty())
	   OutFile.Open(OutFileName.c_str(), "w+b");

    SkipClots = IsEmpty(O_CLOTHEAD_1) && IsEmpty(O_CLOTBREAK_1) &&
	       IsEmpty(O_CLOTTAIL_1) && IsEmpty(O_HIGHPASS_4) &&
           IsEmpty(O_LOWPASS_4);

    DString Garbage = aOptions.Get(O_SEPARATORS) + aOptions.Get(O_PUNCT);
    PunctParse = new CPunctParse(Garbage);

    InitTitleParts();

    int Flags = CRegex::DefaultCompFlags;
    if (!aOptions.COptions<int>::Get(O_CRUFTCASE))
        Flags |= CRegex::Icase;
    TextCruft = CRegex(aOptions.Get(O_CRUFT), Flags, 0);
}

CFormDisplay::~CFormDisplay()
{
    delete PunctParse;
}

// This is a *very* coplicated way to print something... Isn't there
// an easier way?
void CFormDisplay::ErrPrint(const vector<char> &Buffer)
{
    vector<char>::const_iterator Stream = Buffer.begin();
    do
    {   vector<char>::const_iterator Escape = GetEscape(Stream,
            Buffer.end());
        PrintChars(Stream, Escape);
        if (Escape == Buffer.end())
            Stream = Buffer.end();
        else
        {   PrintEscape(Escape);
            Stream = Escape + 8;
        }
    }   while (Stream != Buffer.end());
}

// I *should* use STL algorithms here - if only I understood them...
vector<char>::const_iterator CFormDisplay::GetEscape(
	vector<char>::const_iterator Begin,
	vector<char>::const_iterator End)
{
    vector<char>::const_iterator p = Begin;
    while (p != End)
    {   if (IsEscape(p, End))
            return p;
        ++p;
    }
    return End;
}

bool CFormDisplay::IsEscape(
	vector<char>::const_iterator Begin,
	vector<char>::const_iterator End)
{
    return End - Begin >= 8 &&
        Begin[0] == 0x1b && Begin[1] == '[' &&
        Begin[3] == ';' && Begin[5] == ';' && Begin[7] == 'm';
}

void CFormDisplay::PrintChars(
	vector<char>::const_iterator Begin,
	vector<char>::const_iterator End)
{
    while (Begin != End)
    {   fputc(*Begin, stdout);
        ++Begin;
    }
}

void CFormDisplay::PrintEscape(vector<char>::const_iterator Escape)
{
    printf("%c[%d;%d;%dm", 0x1b, Escape[2], Escape[4], Escape[6]);
}

void CFormDisplay::OutPrint(const vector<char> &Buffer)
{
    if (OutFile.IsAssigned())
    {   vector<char>::const_iterator p = Buffer.begin();
        if (Buffer.size())
            OutFile.Write(p, Buffer.size());
    }
    else
        ErrPrint(Buffer);
}

void CFormDisplay::Display()
{
    DisplayUnindexed();

    DisplayProlog();

    CObjPtr<DDocument> Doc1 = Mirror.NewFirstDoc(&Result);
    while (Doc1 != NULL)
    {   CObjPtr<DDocument> Doc2 = Mirror.NewNextDoc(&Result, Doc1);
	   DisplayDoc(Doc1.Drop());
	   Doc1 = Doc2.Drop();
    }

    DisplayEpilog();
}

void CFormDisplay::DisplayUnindexed()
{
    DISArray<DString> &Unknown = Result.GetUnknown();
    if (Unknown.GetCount())
    {   DString NotFoundTemplate = Options.Get(O_NFLINE_1);

        CFormatRaw Format;
        ErrPrint(Format.GetMessage(Options.Get(O_NOTFOUNDHEADER_0)));
        DISArray<DString>::DIterator Itr = Unknown.InitIterator();
        while (Unknown.IsDereferenceable(Itr))
        {	DString *CurStr = Unknown.GetCurrent(Itr);
            Assert(!CurStr->IsEmpty());
            CFormatRaw Format;
            Format << const_cast<const DString &>(*CurStr);
            ErrPrint(Format.GetMessage(NotFoundTemplate));
            ++Itr;
        }
    }
}

void CFormDisplay::DisplayProlog()
{
    CFormatRaw Format;
    OutPrint(Format.GetMessage(Options.Get(O_PROLOG_0)));
}

void CFormDisplay::DisplayEpilog()
{
/*    CFormatRaw Format;
    vector<char> Message = Format.GetMessage(Options.Get(O_EPILOG));
    OutPrint(Message);
    if (!OutFile.IsAssigned())
        printf("%c[%d;m", 0x1b, 0);*/
    DString EpilogTemplate = Options.Get(O_EPILOG);
    vector<char> Buffer(1);
    const char *EscapeFormat = "%c[%d;m";
    const char *p = EpilogTemplate.c_str();
    while (*p)
    {   if (p[0] != '%' || p[1] != '0')
        {   Buffer[0] = *p;
            OutPrint(Buffer);
            ++p;
        }
        else
        {   if (OutFile.IsAssigned())
                OutFile.Print(EscapeFormat, 0x1b, 0);
            else
                printf(EscapeFormat, 0x1b, 0);
            p += 2;
        }
    }
}

void CFormDisplay::DisplayDoc(DDocument *InDoc)
{   CDocDesc DocDesc;
    const char *Template;

    CObjPtr<DDocument> Doc = InDoc;

    Mirror.GetDocName(Doc, DocDesc);

    CFormatRaw Format;
    Format << GetTail(DocDesc.GetFileName());
    if (DocDesc.GetStartOffs() == 0)
        Template = O_WHOLEFILE_X;
    else
    {   Template = O_PARTFILE_X;
        Format << DocDesc.GetStartOffs();
    }
    Format << GetColorEscape(Options.COptions<int>::Get(O_DOCNAMECOLOR));
    AddTitleText(Format, InDoc);
    OutPrint(Format.GetMessage(Options.Get(Template)));

    CRelevance *Relevance = Doc->GetRelObj();
    if (Relevance)
        DisplayRelevance(Relevance, Doc->GetRelRes());

    // optimization
    if (!SkipClots)
	   DisplayDocContents(Doc.Drop());

    CFormatRaw EndFormat;
    OutPrint(EndFormat.GetMessage(Options.Get(O_DOCEND)));
}

void CFormDisplay::DisplayDocContents(DDocument *InDoc)
{   
    try
    {   CClot *Clot;

	   CObjPtr<DDocument> Doc = InDoc;
	   CObjPtr<CClotCursor> ClotCursor = Mirror.NewClotCursor(Doc.Drop(),
            Options.COptions<int>::Get(O_SPOTDIST));
        int HighlightCount = 0;
        while ((Clot = ClotCursor->GetCurrent()) != NULL)
        {   if (DisplayClot(Clot, HighlightCount))
                break;
            Mirror.Iterate(ClotCursor);
        }
    }
    catch (CPassageOverflow &)
    {   CFormatRaw Format;
        Format << DString::MaxAllowedLength;
        OutPrint(Format.GetMessage(Options.Get(O_PASSAGEOVERFLOW_1)));
    }
}

void CFormDisplay::DisplayRelevance(CRelevance *Relevance, double RelRes)
{
    for (CRelevance::EFeature Feature = (CRelevance::EFeature)0;
        Feature < CRelevance::FeatureCount;
        Feature = (CRelevance::EFeature)(Feature + 1))
    {   CFormatRaw Format;
        Format << CRelevance::FeatureName[Feature] << 
            Relevance->GetMetric(Feature);
        OutPrint(Format.GetMessage(Options.Get(O_METRIC_2)));
    }
    CFormatRaw Format;
    Format << RelRes;
    OutPrint(Format.GetMessage(Options.Get(O_RELEVANCE_1)));
}

int CFormDisplay::DisplayClot(CClot *Clot, int &HighlightCount)
{   int HeadTailColor;
    int HighlightMax;
    DString Passage;

    HeadTailColor = Options.COptions<int>::Get(O_CLOTCOLOR);
    HighlightMax = Options.COptions<int>::Get(O_HIGHLIGHTMAX);

    CFormatRaw Format;
    Format << GetColorEscape(HeadTailColor);
    OutPrint(Format.GetMessage(Options.Get(O_CLOTHEAD_1)));

    int ColorIndex = Clot->GetFirstLow() ? 0 : 1;
    int i = 0;
    while (!Clot->GetPassage(i, Passage))
    {   if (ColorIndex == 1 && HighlightMax > 0)
        {   HighlightCount++;
            if (HighlightCount > HighlightMax)
            {   // reuse Format
		        OutPrint(Format.GetMessage(Options.Get(O_CLOTBREAK_1)));
                return 1;
            }
        }

        DisplayPassage(Passage,
            ColorIndex ? O_HIGHPASS_4 : O_LOWPASS_4);

	   ColorIndex = ColorIndex ? 0 : 1;
	   i++;
    }

    // reuse Format
    OutPrint(Format.GetMessage(Options.Get(O_CLOTTAIL_1)));
    return 0;
}

// O_FORMATSTRING no longer supported
void CFormDisplay::DisplayPassage(const DString &Passage, 
    const char *Template)
{   DString Before;
	DString After;
    DString ParsedPassage = Passage;
    PunctParse->ParsePunct(ParsedPassage, Before, After);

    CFormatRaw Format;
    Format << GetCleanText(Before, TextCruft) 
        << GetCleanText(ParsedPassage, TextCruft)
        << GetCleanText(After, TextCruft)
        << TextColor
        << OccColor;
    OutPrint(Format.GetMessage(Options.Get(Template)));
}

int CFormDisplay::IsEmpty(const char *Name)
{
    DString Value = Options.Get(Name);
    return Value.IsEmpty();
}

DString CFormDisplay::GetCleanText(const DString &RawText,
    const CRegex &Cruft)
{
    if (RawText.IsEmpty())
        return RawText;

    int From;
    int To;
    if (!Cruft.Matches(RawText.c_str(), From, To))
        return RawText;

    // From & To valid from here on

    int s = RawText.GetLength() + 1;
    Assert(To < s);
    CMemPtr<char> Buffer(s);
    memcpy((char *)Buffer, RawText.c_str(), s);

    do
    {   Assert(From <= To);
        if (From == To)
            break;

        memmove((char *)Buffer + From, (char *)Buffer + To, s - To);
        s -= To - From;
        Assert(s == strlen((char *)Buffer) + 1);
    }   while (Cruft.Matches((char *)Buffer, From, To));

    return DString((char *)Buffer);
}

DString CFormDisplay::GetTail(const DString &DocFileName)
{
    if (HiddenHead.IsEmpty() || 
        DocFileName.GetLength() <= HiddenHead.GetLength())
        return DocFileName;

    CMemPtr<char> b(HiddenHead.GetLength() + 1);
    memcpy((char *)b, DocFileName.c_str(), HiddenHead.GetLength());
    *(((char *)b) + HiddenHead.GetLength()) = 0;
    if (FileStrCmp((char *)b, HiddenHead.c_str()))
        return DocFileName;

    return DString(DocFileName.c_str() + HiddenHead.GetLength());
}

void CFormDisplay::AddTitleText(CFormatRaw &Format,
    DDocument *Doc)
{
    DString Out = "";
    for (vector<pair<short, CRegex> >::const_iterator i = TitleParts.begin();
         i != TitleParts.end();
         ++i)
         Format << GetCleanText(Mirror.GetSubText(Doc, i->first),
             i->second);
}

void CFormDisplay::InitTitleParts()
{
    CDocPartNames DocPartNames(Options);
    CStringListValue Parts = Options.COptions<CStringListValue>::Get(
        O_TITLEPARTS);

    for (int i = 0; i < Parts.GetCount(); ++i)
    {   short SubId = DocPartNames.GetId(Parts[i]);

        CRegex SubCruft;
        if (SubId == 0)
            SubCruft = TextCruft;
        else if (SubId > 0) // this assures GetUno is OK
        {   int Flags = CRegex::DefaultCompFlags;
            if (!Options.COptions<int>::GetUno(Parts[i], O_CRUFTCASE))
                Flags |= CRegex::Icase;
            SubCruft = CRegex(Options.COptions<DString>::GetUno(Parts[i],
                    O_CRUFT), Flags, 0);
        }

        TitleParts.push_back(pair<short, CRegex>(SubId, SubCruft));
    }
}
