#if !defined(__frontln_h)
#define __frontln_h

#include "arrays.h"
#include "fritem.h"
#include "linedoc.h"
#include "timeout.h"

class DDocument;
class CAnswer;
class CReporter;

class CFrontLine
{
public:
    CFrontLine(const DString &TempDir, 
        ulong aTotalDocCount,
	int TimeoutSec,
	int &aBreakFlag);
    ~CFrontLine();

    int AddVariation(CFrontItem *InItem);
		// returns 0 OK, 1 interrupted

    int GetFirstOcc(ulong &GlobOcc); // out
        // returns 0 OK, 1 no occurences found

    void AddCachedDoc(CLineDoc *InDoc);

    int SkipDoc(ulong &GlobOcc); // in & out
	   // returns 0 OK, GlobOcc in next doc
	   // 1 GlobOcc was in last doc (or interrupted)
	   // 2 document of GlobOcc was not cached
    int SkipDoc(ulong DocNum,
	   ulong &GlobOcc); // out
	   // returns 0 OK, GlobOcc in next doc
	   // 1 DocNum is last doc (or interrupted)
	   // 2 DocNum is not cached

    DDocument *NewDoc(ulong DocNum);
        // returns NULL => DocNum not cached or interrupted

    DISArray<DString> &GetUnknown();

    int IsBroken() { return Broken; }

    int GetVariationCount() { return Front.GetCount(); }
    CFrontItem *GetVariation(int Index) { return Front[Index]; }

    ushort GetFreeIndex() { return FreeIndex; }

    const ulong TotalDocCount;
private:
    static const ulong FOUninit;
    static const ulong FOEmpty;

    ulong FirstOcc;
    DISArray<CFrontItem> Front;
    DISArray<DString> Unknown;

    CTimeout Timeout;

    DISArray<CLineDoc> Cache;

    CStorage Storage; 

    int &BreakFlag;
    int Broken;

    ushort FreeIndex;

    int IsInitialized() { return FirstOcc != FOUninit; }
    void Init();

    int SkipDocIntern(CLineDoc *LineDoc, ulong &GlobOcc);
         // returns 0 OK, GlobOcc out param
	    // 1 no more docs

    DDocument *NewDocIntern(CLineDoc *LineDoc);
		// return NULL when interrupted

    int ShrinkCache(CLineDoc *LineDoc);
		// returns 0 OK, 1 interrupted

    void AddIndices(CAnswer *Answer, DArray<ushort> &Indices, ulong LocalOccNum);
};

#endif

