#include "myfndt.h"
#include "interval.h"
#include "grazeopt.h"
#include "docdesc.h"
#include "filebase.h"

#include "ointrvl.h"

CInterval::CInterval(CFileBase &FileBase,
    CGrazerOptions &Options):
    Low(FileBase,
	   Options,
	   O_INTERSTARTFILE,
	   O_INTERSTARTOFFS,
	   O_INCLUDEINTERSTART),
    High(FileBase,
	   Options,
	   O_INTERSTOPFILE,
	   O_INTERSTOPOFFS,
	   O_INCLUDEINTERSTOP)
{
    CBound::EMode LowMode = Low.GetMode();
    CBound::EMode HighMode = High.GetMode();
    if (LowMode == CBound::infinite || HighMode == CBound::infinite)
	   return;

    if (High.LessThan(Low) ||	(High.Equals(Low) && (HighMode == 
		  CBound::interval_opened || LowMode == CBound::interval_opened)))
    {   DString EmptyIntervalErr = Options.Get(T_EMPTYINTERVAL_N);
	   throw CGeneralException(EmptyIntervalErr.c_str());
    }
}

int CInterval::IsUnder(ulong Num)
{
    return Low.IsUnder(Num);
}

int CInterval::IsOver(ulong Num)
{
    return High.IsOver(Num);
}

CInterval::CBound::CBound(CFileBase &FileBase,
    CGrazerOptions &Options,
    const char *BDFileOptId,
    const char *BDOffsOptId,
    const char *BoundaryInIntervalOptId):
    Mode(infinite)
{
    DString BoundaryDocName = Options.Get(BDFileOptId);
    if (BoundaryDocName.IsEmpty())
        return;

    int BoundaryDocOffs = Options.COptions<int>::Get(BDOffsOptId);

    CFileBase::SFileRecord Rec;
    CDocDesc DocDesc(BoundaryDocName, BoundaryDocOffs);
    short Bogus = FileBase.Get(DocDesc, Rec, Num);
    if (Bogus)
    {   ThrowNotFound(Options.Get(T_BDNOTFOUND_SDN), 
		  BoundaryDocName, 
		  BoundaryDocOffs);
    }

    Mode = Options.COptions<int>::Get(BoundaryInIntervalOptId) ? interval_closed :
            interval_opened;
}

int CInterval::CBound::Equals(const CBound &Other)
{
    if (Mode == infinite)
	   return Other.Mode == infinite;
    return Num == Other.Num;
}

int CInterval::CBound::LessThan(const CBound &Other)
{
    if (Mode == infinite || Other.Mode == infinite)
       return 0;
    return Num < Other.Num;
}

CInterval::CBound::EMode CInterval::CBound::GetMode()
{
    return Mode;
}

int CInterval::CBound::IsUnder(ulong aNum)
{
    if (Mode == infinite ||
	   aNum > Num)
	   return 0;

    if (aNum < Num)
	   return 1;

    return Mode == interval_opened;
}

int CInterval::CBound::IsOver(ulong aNum)
{
    if (Mode == infinite ||
	   aNum < Num)
	   return 0;

    if (aNum > Num)
	   return 1;

    return Mode == interval_opened;
}

void CInterval::CBound::ThrowNotFound(const DString ErrFormatSD,
    const DString &BDFile, 
    int BDOffs)
{
    CMemPtr<char> Buffer(ErrFormatSD.GetLength() + BDFile.GetLength() + 40);
    sprintf((char *)Buffer, ErrFormatSD.c_str(), BDFile.c_str(), BDOffs);
    throw CGeneralException((char *)Buffer);
}
