#include "myfndt.h"
#include "mirror.h"
#include "docum.h"
#include "numclot.h"
#include "clotcurs.h"
#include "clot.h"
#include "cluster.h"
#include "spot.h"
#include "result.h"
#include "docwrap.h"
#include "softop.h"
#include "wordlist.h"
#include "reporter.h"
#include "preader.h"
#include "decompr.h"
#include "interval.h"
#include "frontln.h"
#include "fritem.h"
#include "linedoc.h"
#include "extras.h"
#include "query.h"
#include "formula.h"
#include "wordprov.h"
#include "idxprov.h"
#include "fileprov.h"
#include "compprov.h"
#include "filestat.h"
#include "watch.h"
#include "infra.h"
#include "subdesc.h"

#include "oindex.h"
#include "omirror.h"
#include "obase.h"
#include "otmp.h"
#include "omode.h"
#include "oreader.h"
#include "oinfra.h"

#include <string.h>
#if defined(__GNUG__)
#include <unistd.h>
#else
#if defined(_MSC_VER)
#include <io.h>
#define access _access
#endif
#endif

//#define DEBUG_stemhang
//#define DEBUG_frgline

CMirror::CMirror(CDecompressor *aDecompressor,
	CGrazerOptions &Options):
	FileBase(0, Options),
	GrazeIndex(Options.Get(O_WORDTREEEXT),
		Options.Get(O_OCCSTOREFREELISTEXT),
		Options.Get(O_OCCSTOREDATAEXT),
		Options.COptions<int>::Get(O_AVEWORDLEN)),
	ClusterBook(new CULongHandler(),
		Options.Get(O_CLUSTERFILEEXT),
		0),
	Separators(Options.Get(O_SEPARATORS)),
	Divider(CRegex(Options.Get(O_DIVIDER))),
    ReaderMargin(Options.COptions<int>::Get(O_READERMARGIN)),
	Decompressor(aDecompressor),
	HeadLength(Options.COptions<int>::Get(O_HEADLENGTH)),
    Infrastructure(NewInfrastructureOrNull(Options))
{
	Interval = new CInterval(FileBase, Options);
}

CMirror::~CMirror()
{
    delete Infrastructure;
	delete Interval;
	delete Decompressor;
}

CResult *CMirror::NewResult(CFormula &Formula,
    int DocLimit,
    const DString &ResultDir,
    CReporter &Reporter,
	int Timeout,
    int &BreakFlag)
{   int Broken;

    CFrontLine FrontLine(ResultDir, 
        GetTotalDocCount(),
        Timeout,
        BreakFlag);
    if (Formula.Collate(FrontLine))
	   return NULL;

    int Count = FrontLine.GetVariationCount();
    Assert(Count > 0);
    for (int i = 0; i < Count; i++)
    {   if (FrontLine.IsBroken())
		  return NULL;

	   CFrontItem *FrontItem = FrontLine.GetVariation(i);
	   FrontItem->SetOccList(GetOccList(FrontItem->WordVariation));
    }

	CExtras Extras(ResultDir, 
	    DocLimit,
	    Formula,
        Reporter);
	return NewSearchResult(FrontLine, Extras);
}

CResult *CMirror::NewSearchResult(CFrontLine &FrontLine,
	CExtras &Extras)
{
	CObjPtr<CResult> Result = new CResult(Extras.ResultDir,
		Extras.ResultDir,
		Extras.ResultDir,
		Extras.DocLimit); 

	if (!FrontLine.TotalDocCount)
	{   // untested
	    Result->SetUnknown(FrontLine.GetUnknown());
	    return Result.Drop();
	}

	CFormula &Formula = Extras.Formula;

	CObjPtr<DDocument> Doc1 = NewFirstDoc(FrontLine);
	while (Doc1 != NULL)
    {	Extras.Reporter.ReportIn();
	    Extras.Reporter.ReportProgress(1, 
            (short)((100 * Doc1->GetNum()) / FrontLine.TotalDocCount));

	    CObjPtr<DDocument> Doc2 = NewNextDoc(FrontLine, Doc1);
	    CObjPtr<CDocWrapper> DocWrapper = NewDocWrapper(Doc1.Drop());
	    Doc1 = Doc2.Drop();
	    Formula.AddRelevant(Result, DocWrapper);
	}

	if (!FrontLine.IsBroken())
	{	Result->SetUnknown(FrontLine.GetUnknown());
		return Result.Drop();
	}
	else
		return NULL;
}

DDocument *CMirror::NewFirstDoc(CFrontLine &FrontLine)
{	ulong GlobOcc;

	if (FrontLine.GetFirstOcc(GlobOcc))
	     return NULL;

	ulong Document = AddCachedDoc(FrontLine, GlobOcc);
	while (IsUnder(Document))
	{    int Ret = FrontLine.SkipDoc(GlobOcc);
	     Assert(Ret != 2);
	     if (Ret)
			return NULL;
		Document = AddCachedDoc(FrontLine, GlobOcc);
     }

	if (IsOver(Document))
		return NULL;

	return FrontLine.NewDoc(Document);
		// can be NULL - when interrupted
}

DDocument *CMirror::NewNextDoc(CFrontLine &FrontLine, DDocument *Doc)
{   ulong GlobOcc;

    int Ret = FrontLine.SkipDoc(Doc->GetNum(), GlobOcc);
    Assert(Ret != 2);
    if (Ret)
	   return NULL;

    ulong Document = AddCachedDoc(FrontLine, GlobOcc);
    if (IsOver(Document))
	   return NULL;

    return FrontLine.NewDoc(Document);
		// can be NULL - when interrupted
}

ulong CMirror::AddCachedDoc(CFrontLine &FrontLine, ulong GlobOcc)
{   ulong Document;
    CFileBase::SFileRecord Rec;
    short KnownRet;

    KnownRet = FileBase.GetNum(GlobOcc, Document);
    Assert(!KnownRet);

    CDocDesc DocDesc;
    KnownRet = FileBase.Get(Document, Rec, DocDesc);
    Assert(!KnownRet);

    FrontLine.AddCachedDoc(new CLineDoc(Document, Rec));
    return Document;
}

CDatumEnvelopeTemplate<CFixedOcc> CMirror::GetOccList(
	const DString &IndexedWord)
{
	return GrazeIndex.GetOcc(IndexedWord);
}

DDocument *CMirror::NewFirstDoc(CResult *Result)
{
	return Result->NewFirstDoc();
}

DDocument *CMirror::NewNextDoc(CResult *Result, DDocument *Doc)
{
	return Result->NewNextDoc(Doc);
}

void CMirror::GetDocName(DDocument *Doc, CDocDesc &DocDesc)
{	CFileBase::SFileRecord DummyRec;

	short KnownRet = FileBase.Get(Doc->GetNum(),
		DummyRec,
		DocDesc);
	Assert(!KnownRet);
}

CSpot *CMirror::NewFirstSpot(DDocument *Doc)
{
	CAnswer *Answer = Doc->GetAnswer();
	long Itr = Answer->InitIterator();
	Assert(Itr != 0);
	CAnswer::COcc Hit = Answer->GetHit(Itr);
	return NewSpot(Doc->GetNum(), Hit.GetOccNum(), Itr);
}

CSpot *CMirror::NewNextSpot(CSpot *Spot, DDocument *Doc)
{
	CAnswer *Answer = Doc->GetAnswer();
	long Itr = Spot->GetIterator();
	if (Itr == 0)
		return NULL;
	Answer->ResetIterator(Itr);
	CAnswer::COcc Hit = Answer->GetHit(Itr);
	return NewSpot(Doc->GetNum(), Hit.GetOccNum(), Itr);
}

CClotCursor *CMirror::NewClotCursor(DDocument *InDoc, int Distance)
{	CDocDesc DocDesc;

	CObjPtr<DDocument> Doc = InDoc;

	GetDocName(Doc, DocDesc);
	DocDesc.SetContentFile(Decompressor);
	CObjPtr<CReader> Reader = new CReader(Separators, new CWatch(Divider,
		ReaderMargin));
    Reader->SetIn(DocDesc.GetFileName(), DocDesc.GetStartOffs());

	CSpot *Spot = NewFirstSpot(Doc);
	Assert(Spot != NULL);
	CObjPtr<CClot> Clot = NewClot(Doc, Reader, Distance, Spot, 1);

	return new CClotCursor(Doc.Drop(), 
		Reader.Drop(),
		Distance,
		Clot.Drop());
}

void CMirror::Iterate(CClotCursor *ClotCursor)
{  CClot *NClot;

	CClot *Clot = ClotCursor->GetCurrent();
	Assert(Clot != NULL);
     if (Clot->IsDouble())
	{    Clot->Regularize();
		return;
	}

	CSpot *NextSpot = NewNextSpot(Clot->GetLastSpot(),
		ClotCursor->GetDoc());
	if (!NextSpot)
		NClot = NULL;
	else
		NClot = NewClot(ClotCursor->GetDoc(),
			ClotCursor->GetReader(),
			ClotCursor->GetDistance(),
			NextSpot,
			0);
	ClotCursor->ReplaceClot(NClot);
}

CSpot *CMirror::NewSpot(ulong DocNum, ulong LocalOcc, long Iterator)
{	CFileBase::SFileRecord Rec;
	CDocDesc DocDesc;
	long Offs;
	short KnownRet;

	KnownRet = FileBase.Get(DocNum, Rec, DocDesc);
	Assert(!KnownRet);
	ulong GlobalOcc = LocalOcc + Rec.FirstWord;

	CNewKey TestKey(CULongHandler::NewKey(GlobalOcc));
	CLockedKey FoundKey(ClusterBook.LockKeyBefore(TestKey),
			&ClusterBook);
	Assert(FoundKey != NULL);
	KnownRet = ClusterBook.GetData(FoundKey, Offs);
	Assert(!KnownRet);
	ulong ClusterNum = CULongHandler::GetULong(FoundKey);
	Assert(ClusterNum <= GlobalOcc);
	return new CSpot(Offs,
		ClusterNum,
		static_cast<ushort>(GlobalOcc - ClusterNum),
		Iterator);
}

CClot *CMirror::NewClot(DDocument *Doc,
	CReader *Reader,
	int Distance,
	CSpot *InSpot,
	int WithUserHead)
{
	CObjPtr<CSpot> Spot1 = InSpot;
	int HeadDist = Spot1->GetWordDelta();
	CCluster Cluster = Spot1->GetCluster();
	while (((Distance == 0 || HeadLength > 0) && HeadDist > 0) ||
		HeadDist <= Distance / 2)
	{	ulong OldWordNum = Cluster.GetFirstWordNum();
		if (IterateBack(Doc->GetNum(), Cluster))
			break;

		Assert(OldWordNum > Cluster.GetFirstWordNum());
		HeadDist += static_cast<int>(OldWordNum -
				Cluster.GetFirstWordNum());
	}

	CNumClot NumClot(Cluster);
	NumClot.Add(Spot1->GetWordNum());

	while (1)
	{	CObjPtr<CSpot> Spot2 = NewNextSpot(Spot1, Doc);
		if (!Spot2)
			break;
		Assert(Spot2->GetWordNum() >= Spot1->GetWordNum());
		if (Distance > 0 &&
			Spot2->GetWordNum() - Spot1->GetWordNum() > 
				(unsigned int)Distance)
			break;
		NumClot.Add(Spot2->GetWordNum());
		Spot1 = Spot2.Drop();
	}

	CObjPtr<CClot> Clot = new CClot(NumClot.IsFirstLow(),
		Spot1.Drop());
	FillClot(Clot,
		Reader,
		&NumClot,
		Distance,
		WithUserHead);
	return Clot.Drop();
}

void CMirror::FillClot(CClot *Clot,
	CReader *Reader,
	CNumClot *NumClot,
	int Distance,
	int WithUserHead)
{	ushort Count;
	DString Passage;
	ushort Retrieved;
	ushort TailCount;

	Reader->Seek(NumClot->GetOffs());
	ushort Index = FillClotHead(Clot, Reader, NumClot, Distance, WithUserHead);

	while (!NumClot->GetPassageCount(Index, Count))
	{  Assert(Count > 0);
		Retrieved = Reader->GetWords(Count, Passage);
		Assert(Retrieved == Count);
		Clot->Add(Passage);
		Index++;
	}

	if (Distance > 0)
		TailCount = Reader->GetWords(Distance / 2, Passage);
	else
		TailCount = Reader->GetAllWords(Passage);
	if (TailCount)
		Clot->Add(Passage);
}

short CMirror::IterateBack(ulong DocNum, CCluster &Cluster)
{	long Offs;
	short KnownRet;
	CFileBase::SFileRecord Rec;
	CDocDesc DocDesc;

	CNewKey TestKey(CULongHandler::NewKey(Cluster.GetFirstWordNum()));
	CLockedKey FoundKey(ClusterBook.LockPrevKey(TestKey),
			&ClusterBook);
	if (!FoundKey) // Cluster is first in base
		return 1;

	ulong ClusterNum = CULongHandler::GetULong(FoundKey);
	KnownRet = FileBase.Get(DocNum, Rec, DocDesc);
	Assert(!KnownRet);
	if (ClusterNum < Rec.FirstWord) // Cluster is first in document
		return 1;

	KnownRet = ClusterBook.GetData(FoundKey, Offs);
	Assert(!KnownRet);
	Cluster = CCluster(ClusterNum, Offs);
	return 0;
}

CDocWrapper *CMirror::NewDocWrapper(DDocument *InDoc)
{	short KnownRet;
	CFileBase::SFileRecord Rec;
	CDocDesc DocDesc;

	CObjPtr<DDocument> Doc = InDoc;
	KnownRet = FileBase.Get(InDoc->GetNum(), Rec, DocDesc);
	Assert(!KnownRet);
	Assert(Rec.WordCount > 0);
    CSubDesc *SubDesc = NULL;
    if (Infrastructure)
        SubDesc = Infrastructure->NewSubDesc(InDoc->GetNum());
	return new CDocWrapper(Doc.Drop(), 
        SubDesc,
        Rec.WordCount);
}

#if !defined(_MSC_VER)
CWordList *CMirror::NewWordList(CGrazerOptions &Options,
    CReporter &Reporter)
{   DString ItrWord = "";

    CObjPtr<CWordProvider> WordProvider = NewWordProvider(Options, Reporter);
    CObjPtr<CWordList> WordList = new CWordList(Options.Get(O_TEMPDIR));
    int Perverse = Options.COptions<int>::Get(O_PERVERSEWORDLIST);
	 
    if (WordProvider->GetWord(ItrWord))
	   return WordList.Drop();

    do
    {   long DocNum, OccCount;

	   CDatumEnvelopeTemplate<CFixedOcc> OccList =
		  GrazeIndex.GetOcc(ItrWord);
	   if (!Perverse)
	   {   if (!OccList.IsEmpty()) 
		  // explicitly requested word does not have to be indexed
		  {   GetDocCount(OccList, DocNum, OccCount);
			 WordList->AddWord(ItrWord, DocNum, OccCount);
		  }
	   }
	   else
	   {   if (OccList.IsEmpty())
			 WordList->AddWord(ItrWord, 0, 0);
	   }
#if defined(DEBUG_wordlist)
	   printf("got %s\n", ItrWord.c_str());
#endif
	   Reporter.ReportProgress(1, GetPerc(ItrWord));
    }   while (!WordProvider->GetWord(ItrWord));

    return WordList.Drop();
}

#endif

void CMirror::GetDocCount(CFixedOcc *FixedOcc,
	long &DocTotal,
	long &OccCount)
{	ulong Document;
	short KnownRet;

	long OccItr = FixedOcc->InitIterator();
	Assert(OccItr != 0);
	ulong CurDoc = ~(0ul);
	DocTotal = 0;
	OccCount = 0;
	do
	{	ulong GlobOcc = FixedOcc->GetNext(OccItr);
		KnownRet = FileBase.GetNum(GlobOcc, Document);
		Assert(!KnownRet);
		if (CurDoc != Document)
		{	CurDoc = Document;
			DocTotal++;
		}
		OccCount++;
	}	while (OccItr);
}

short CMirror::GetPerc(const DString &CurWord)
{
	Assert(!CurWord.IsEmpty());
	char c = CurWord[0];
	c -= 'A' + 1;
	return (100 * c) / ('z' - 'A');
}

void CMirror::DropDecompressedFiles()
{
	Decompressor->DropAll();
}

ushort CMirror::FillClotHead(CClot *Clot,
	CReader *Reader,
	CNumClot *NumClot,
	int Distance,
	int WithUserHead)
{    DString Passage;
	ushort Retrieved;

	if (Distance == 0)
	     return 0;

    int HeadDist = NumClot->GetHeadDist();
	Assert(Distance > 0);
	int Show2 = Distance / 2;
	if (HeadDist <= Show2)
	     return 0;

	Assert(HeadLength >= 0);
	int Show1 = WithUserHead ? HeadLength : 0;
	int Throw = 0;
	if (Show1 + Show2 < HeadDist)
	     Throw = HeadDist - (Show1 + Show2);
	else
	{    Show1 = 0;
		Show2 = HeadDist;
	}

	if (Show1)
	{    Retrieved = Reader->GetWords(Show1, Passage);
		Assert(Retrieved == Show1);
		Clot->AddUserHead(Passage);
	}
	if (Throw)
	{    Retrieved = Reader->GetWords(Throw, Passage);
		Assert(Retrieved == Throw);
	}
	if (Show2)
	{	Retrieved = Reader->GetWords(Show2, Passage);
		Assert(Retrieved == Show2);
		Clot->Add(Passage);
		return 1;
	}
	else
	     return 0;
}

ulong CMirror::GetTotalDocCount()
{
    return FileBase.GetTotalDocCount();
}

int CMirror::IsUnder(ulong Num)
{
	return Interval->IsUnder(Num);
}

int CMirror::IsOver(ulong Num)
{
	return Interval->IsOver(Num);
}

CWordProvider *CMirror::NewWordProvider(CGrazerOptions &Options,
    CReporter &Reporter)
{
    DString ExplicitFile = Options.Get(O_REQWORDFILE);
    if (ExplicitFile.IsEmpty())
    {   Assert(!Options.COptions<int>::Get(O_PERVERSEWORDLIST)); 
	       // see CGrazerOptions::FirstInvalidKey()
  	   return new CIndexProvider(GrazeIndex);
    }
    else
	   return new CComposedProvider(
		  new CFileProvider(ExplicitFile, Options),
		  Reporter,
		  Options.Get(O_TEMPDIR));
}

CInfrastructure *CMirror::NewInfrastructureOrNull(CGrazerOptions &Options)
{
    DString InfraFile = Options.Get(O_INFRAKEY);
	if (access(InfraFile.c_str(), 0))
        return NULL;
    return new CInfrastructure(0, Options);
}

DString CMirror::GetSubText(DDocument *Doc, short SubId)    
{
    if (!Infrastructure)
        return DString("");

    CObjPtr<CSubDesc> SubDesc = Infrastructure->NewSubDesc(Doc->GetNum());
    if (!SubDesc)
        return DString("");

    vector<short>::iterator i = SubDesc->Ids.begin();
    vector<short>::iterator e = SubDesc->Ids.end();
    while (i != e && *i != SubId)
        ++i;
    if (i == e)
        return DString("");

    int n = i - SubDesc->Ids.begin();
    int c = 0;
    if (n < SubDesc->Stairs.size() - 1)
    {   c = SubDesc->Stairs[n + 1] - SubDesc->Stairs[n];
        Assert(c > 0);
    }
        
    return ReadSubText(Doc, SubDesc->Stairs[n], c);
}

DString CMirror::ReadSubText(DDocument *Doc, int Start, int Count)
{   CDocDesc DocDesc;

	GetDocName(Doc, DocDesc);
	DocDesc.SetContentFile(Decompressor);
	CReader Reader(Separators, new CWatch(Divider, ReaderMargin));
    Reader.SetIn(DocDesc.GetFileName(), DocDesc.GetStartOffs());

    Assert(Start >= 0);
    if (Start)
    {   // there are more effective ways to seek, but they're
        // more complicated, and the title should be near the
        // doc start anyway
        DString Dummy;
        ushort Same = Reader.GetWords(Start, Dummy);
        Assert(Start == Same);
    }

    DString Words;
    if (Count)
    {   ushort Same = Reader.GetWords(Count, Words);
        Assert(Count == Same);
    }
    else
        Reader.GetAllWords(Words);
    return Words;
}

