#include "myfndt.h"
#include "qbuild.h"
#include "ande.h"
#include "ore.h"
#include "query.h"
#include "messform.h"
#include "filter.h"
#include "oquery.h"
#include "olang.h"
#include "osubdoc.h"

#include <stdio.h>
#include <string.h>

CQueryBuilder::CQueryBuilder(CTerm::FPrint aPrint, CGrazerOptions &aOptions):
    CQueryBuilderBase(aOptions),
    NamedSoftOps(aOptions.COptions<CStringListValue>::Get(O_SOFTOPS)),
    DocPartNames(aOptions),
    SoftOpName(""),
    Filter(NULL),
    Query(new CQuery),
    FormulaStore(4, 8),
    Print(aPrint),
    Options(aOptions)
{
}

CQueryBuilder::~CQueryBuilder()
{
    delete Query;
    delete Filter;
}

char *CQueryBuilder::Build(const DString &QueryFile,
    CFormula **NewFormula)
{
    *NewFormula = NULL;
    CFormulaBase *FormulaBase;
    char *Out = CQueryBuilderBase::Build(QueryFile, &FormulaBase);
    if (Out)
	   return Out;

    DetachFormula(FormulaBase);
    *NewFormula = static_cast<CFormula *>(FormulaBase);
    return NULL;
}

CQuery *CQueryBuilder::NewQuery()
{
    CQuery *Out = Query;
    Query = new CQuery;
    return Out;
}

CSubdocFilter *CQueryBuilder::YankFilter()
{
    CSubdocFilter *Out = Filter;
    Filter = NULL;
    return Out;
}

char *CQueryBuilder::SetSoftOpName(char *Name)
{
    Assert(Name != NULL);
    if (NamedSoftOps.Find(DString(Name)) == INT_MAX)
    {   DString UnknownOpFormat = Options.Get(T_UNKNOWNOP_S);
	   CMemPtr<char> Err(2 * (UnknownOpFormat.GetLength() + strlen(Name) + 10));
	   sprintf((char *)Err, UnknownOpFormat.c_str(), Name);
	   return Err.Drop();
    }

    SoftOpName = Name;
    return NULL;
}

char *CQueryBuilder::ResetSoftOpName()
{
    SoftOpName = "";
    return NULL;
}

char *CQueryBuilder::AddSoftOpArg(char *String)
{
    if (Query->Add(DString(String)))
    {   DString DuplWordErr = Options.Get(T_DUPLWORD);
	   CMemPtr<char> Err(DuplWordErr.GetLength() + 1);
	   strcpy((char *)Err, DuplWordErr.c_str());
	   return Err.Drop();
    }
    return NULL;
}

char *CQueryBuilder::CreateTerm(CFormulaBase **Out)
{
    if (!Query->GetWordCount())
    {   DString EmptyQuery = Options.Get(T_EMPTYQUERY);
	   int Size = EmptyQuery.GetLength() + 1;
	   CMemPtr<char> Err(Size);
	   memcpy((char *)Err, EmptyQuery.c_str(), Size);
	   return Err.Drop();
    }

    CTerm *Term = new CTerm(*this);
    FormulaStore.Add(Term);
    *Out = Term;
    return NULL;
}

char *CQueryBuilder::CreateBinary(int Op,
    CFormulaBase *Left,
    CFormulaBase *Right,
    CFormulaBase **Out)
{
    CBinaryFormula *Binary;
    Assert(Op == '&' || Op == '|');

    DetachFormula(Left);
    DetachFormula(Right);
    if (Op == '&')
	   Binary = new CAndExpr(static_cast<CSubFormula *>(Left), 
		  static_cast<CSubFormula *>(Right));
    else
	   Binary = new COrExpr(static_cast<CSubFormula *>(Left), 
		  static_cast<CSubFormula *>(Right));
    FormulaStore.Add(Binary);
    *Out = Binary;
    return NULL;
}

char *CQueryBuilder::AddSubdocName(CSubdocFilter *Filter, const char *Name)
{
    Assert(Filter != NULL && Name != NULL);

    short Id = DocPartNames.GetId(DString(Name));
    if (Id == -1)
    {   CFormatString Format;
        Format << Name;
        return Format.AllocMessage(Options.Get(O_UNKNOWNSUBDOC_1));
    }

    Filter->Add(Id);
    return NULL;
}

char *CQueryBuilder::CreateRestrictedTerm(CSubdocFilter *InFilter,
    CFormulaBase **Out)
{
    Assert(InFilter != NULL && Filter == NULL);
    Filter = InFilter;
    return CreateTerm(Out);
}

void CQueryBuilder::DetachFormula(CFormulaBase *Formula)
{
    int KnownRet = FormulaStore.Detach(static_cast<CSubFormula *>(Formula));
    Assert(!KnownRet);
}
