#if !defined(__qbuild_h)
#define __qbuild_h

#include "qebb.h"
#include "grazeopt.h"
#include "formula.h"
#include "arrays.h"
#include "term.h"
#include "dpnames.h"

class CQuery;

class CQueryBuilder : private CQueryBuilderBase
{
public:
    CQueryBuilder(CTerm::FPrint aPrint,
	   CGrazerOptions &aOptions); // keeps an alias
    virtual ~CQueryBuilder();

    // client interface
    char *Build(const DString &QueryFile,
            CFormula **NewFormula); // allocated on heap

    // CTerm interface
    CTerm::FPrint GetPrint() { return Print; }
    CGrazerOptions &GetOptions() { return Options; }
    CQuery *NewQuery();
    DString GetSoftOpName() { return SoftOpName; }
    CSubdocFilter *YankFilter(); // can return NULL

protected:
    // see buildqe.h for pointer memory management
    virtual char *SetSoftOpName(char *Name);
    virtual char *AddSoftOpArg(char *String);
    virtual char *CreateTerm(CFormulaBase **Out);
    virtual char *CreateBinary(int Op,
        CFormulaBase *Left,
        CFormulaBase *Right,
        CFormulaBase **Out);
    virtual char *ResetSoftOpName();
    virtual char *AddSubdocName(CSubdocFilter *Filter, const char *Name);
    virtual char *CreateRestrictedTerm(CSubdocFilter *InFilter,
        CFormulaBase **Out);

private:
    CStringListValue NamedSoftOps;
    DString SoftOpName;
    CSubdocFilter *Filter;
    CQuery *Query;
    CDocPartNames DocPartNames;

    DISArray<CSubFormula> FormulaStore;

    CGrazerOptions &Options;
    CTerm::FPrint Print;

    void DetachFormula(CFormulaBase *Formula);
};

#endif
