#include "myfndt.h"
#include "query.h"
#include "punctaxe.h"
#include "abstcase.h"
#include "abststem.h"

#include <string.h>
#include <stdlib.h>

const int CQuery::DefaultWordLimit = 10;
const int CQuery::DefaultWordDelta = 10;

CQuery::CQuery(int aWordLimit,
	int aWordDelta):
	WordLimit(aWordLimit),
	WordDelta(aWordDelta),
	MatrixCount(0)
{
	Assert(WordDelta > 0);
	int Size = WordLimit * sizeof(DIArray<DString> *);
	Matrix = (DIArray<DString> **)malloc(Size);
	if (!Matrix)
		throw DMemoryException(Size);
	memset(Matrix, 0, Size);
}

CQuery::~CQuery()
{
	for (int i = 0; i < MatrixCount; i++)
		delete Matrix[i];
	free(Matrix);
}

short CQuery::Add(const DString &QWord)
{
	for (int i = 0; i < MatrixCount; i++)
	{	DIArray<DString> *Var = Matrix[i];
		DString WordCopy = QWord;
		if (Var->Find(&WordCopy) != INT_MAX)
			return 1;
	}

	CObjPtr<DIArray<DString> > NVar = new DIArray<DString>(1, 0);
	NVar->Add(new DString(QWord));
	MatrixAdd(NVar.Drop());
	return 0;
}

int CQuery::GetWordCount()
{
	return MatrixCount;
}

int CQuery::AxePunct(CPunctAxe &PunctAxe)
{
	int Ret = 0;
	for (int i = MatrixCount - 1; i >= 0; i--)
	{  DIArray<DString> *Var = Matrix[i];
		int VarRet = AxePunctVar(Var, PunctAxe);
		if (VarRet == 2)
		{	MatrixDetach(i);
			delete Var;
			Ret = 2;
		}
		else
			if (VarRet == 1 && Ret == 0)
				Ret = 1;
	}
	return Ret;
}

int CQuery::AxePunctVar(DIArray<DString> *Var,
	CPunctAxe &PunctAxe)
{
	int Ret = 0;
	int Count = Var->GetCount();
	for (int i = Count - 1; i >= 0; i--)
	{	DString *CurWord = Var->operator[](i);
		DString OldWord = *CurWord;
		if (PunctAxe.AxePunct(*CurWord))
		{	Var->DetachByIndex(i);
			delete CurWord;
			Ret = 1;
		}
		else
		{	if (OldWord != (*CurWord))
				Ret = 1;
		}
	}
	if (Var->GetCount() == 0)
		return 2;
	return Ret;
}

int CQuery::Stem(CAbstractStemmer *Stemmer)
{
    int Ret = 0;
    for (int i = 0; i < MatrixCount; ++i)
    {   DIArray<DString> *Var = Matrix[i];
	   int Count = Var->GetCount();
	   for (int j = 0; j < Count; ++j)
	   {   DString *CurWord = Var->operator[](j);
		  DString Reference = *CurWord;
		  if (!Stemmer->Apply(*CurWord) &&
			 *CurWord != Reference)
		      Ret = 1;
	   }
    }
    return Ret;
}

void CQuery::VariateCase(CAbstractCase *AbstractCase)
{
	for (int i = MatrixCount - 1; i >= 0; i--)
	{	DIArray<DString> *OldVar = Matrix[i];
		Matrix[i] = NewVariate(OldVar, AbstractCase);
	}
}

DIArray<DString> *CQuery::NewVariate(DIArray<DString> *OldVarIn,
		CAbstractCase *AbstractCase)
{
	CObjPtr<DIArray<DString> > OldVar = OldVarIn;
	Assert(OldVarIn != NULL &&
		OldVarIn->GetCount() == 1);

	DString *OldWord = OldVarIn->operator[](0);
	DIArray<DString> *NewVar = AbstractCase->NewVariate(*OldWord);
	if (!NewVar)
		return OldVar.Drop();
	return NewVar;
}

DString CQuery::GetWord(int MajorIndex, int MinorIndex)
{
	Assert(MajorIndex >= 0 && MinorIndex >= 0);
	if (MajorIndex >= MatrixCount)
		return DString();

	DIArray<DString> *Var = Matrix[MajorIndex];
	if (MinorIndex >= Var->GetCount())
		return DString();

	return *(Var->operator[](MinorIndex));
}

void CQuery::MatrixAdd(DIArray<DString> *VarIn)
{
	CObjPtr<DIArray<DString> > Var = VarIn;

	if (MatrixCount == WordLimit)
	{	WordLimit += WordDelta;
		int Size = WordLimit * sizeof(DIArray<DString> *);
		DIArray<DString> **NMatrix = (DIArray<DString> **)malloc(Size);
		if (!NMatrix)
			throw DMemoryException(Size);
		memset(NMatrix, 0, Size);
		memcpy(NMatrix, Matrix, MatrixCount * sizeof(DIArray<DString> *));
		free(Matrix);
		Matrix = NMatrix;
	}

	Matrix[MatrixCount] = Var.Drop();
	MatrixCount++;
}

void CQuery::MatrixDetach(int Index)
{
	Assert(0 <= Index && Index < MatrixCount);
	if (Index < MatrixCount - 1)
		memmove(Matrix + Index,
			Matrix + Index + 1,
			(MatrixCount - Index - 1) * sizeof(DIArray<DString> *));
	Matrix[MatrixCount - 1] = 0;
	MatrixCount--;
}

int CQuery::Del(int MajorIndex)
{
     if (MajorIndex < 0 || MajorIndex >= MatrixCount)
	    return 1;
	
	DIArray<DString> *Col = Matrix[MajorIndex];
	delete Col;
	MatrixDetach(MajorIndex);
	return 0;
}
