#include "myfndt.h"
#include "result.h"
#include "relans.h"
#include "docum.h"
#include "relevanc.h"

CResult::CResult(const DString &aKeyFilePath,
	const DString &aFreeListFilePath,
	const DString &aDataFilePath,
	long aDocLimit):
	KeyHandlerAlias(new CTwoLongHandler()),
	Storage(KeyHandlerAlias,
		aKeyFilePath,
		new CDatumFactoryTemplate<CRelAnswer>(),
		aFreeListFilePath,
		aDataFilePath,
		IT_TEMP | IT_UPDATE),
	DocLimit(aDocLimit),
	CurrentTotal(0l),
	Unknown(5, 5)
{
}

CResult::~CResult()
{
}

void CResult::Add(DDocument *Doc,
	double RelRes,
	CRelevance *InRelevance)
{
	CObjPtr<CRelevance> Relevance = InRelevance;

	CNewKey Key(CTwoLongHandler::NewKey(
		ConvertRelevance(RelRes),
		Doc->GetNum()));

	if (DocLimit == 0 || CurrentTotal < DocLimit)
		CurrentTotal++;
	else
	{	CNewKey LastKey(Storage.NewLastKey());
		Assert(LastKey != NULL);
		if (KeyHandlerAlias->Compare(Key, LastKey) >= 0)
      	// see ConvertRelevance
			return;

		short KnownRet = Storage.Del(LastKey);
		Assert(!KnownRet);
	}

	// always creates new answer
	CRelAnswer *Answer = static_cast<CRelAnswer *>(
			Storage.LockAlways(Key));
	Assert(Answer != NULL);
	// wrap answer in document so that it's unlocked again
	DDocument DestDoc(Answer, Storage, Doc->GetNum());
	if (Relevance != NULL)
		Answer->SetRelevance(Relevance.Drop());
	Answer->Fill(Doc->GetAnswer());
}

DDocument *CResult::NewFirstDoc()
{	long ConvRel;
	long DocNum;

	CNewKey FirstKey(Storage.NewFirstKey());
	if (!FirstKey)
		return NULL;
	CTwoLongHandler::GetTwoLong(FirstKey, ConvRel, DocNum);
	CAnswer *Answer = static_cast<CAnswer *>(Storage.Lock(FirstKey));
	Assert(Answer != NULL);
	return new DDocument(Answer, Storage, DocNum, ConvRel);
}

DDocument *CResult::NewNextDoc(DDocument *Doc)
{  long NewConvRel;
	long NewDocNum;

	long ConvRel = Doc->GetRelevance();
	Assert(ConvRel != 0);
	CNewKey OldKey(CTwoLongHandler::NewKey(
		ConvRel,
		Doc->GetNum()));
	CNewKey NextKey(Storage.NewNextKey(OldKey));
	if (!NextKey)
		return NULL;
	CTwoLongHandler::GetTwoLong(NextKey, NewConvRel, NewDocNum);
	CAnswer *Answer = static_cast<CAnswer *>(Storage.Lock(NextKey));
	Assert(Answer != NULL);
	return new DDocument(Answer, Storage, NewDocNum, NewConvRel);
}

// inverse of DDocument::GetRelRes
long CResult::ConvertRelevance(double Relevance)
{
	if (Relevance < 0)
		Relevance = 0;
	else
		if (Relevance > 1)
			Relevance = 1;
	long PosRel = static_cast<long>(Relevance * 100000l);
	return -PosRel; // so that the highest relevance is first
}

void CResult::SetUnknown(DISArray<DString> &aUnknown)
{
	DISArray<DString>::DIterator Itr = aUnknown.InitIterator();
	while (aUnknown.IsDereferenceable(Itr))
	{	DString *CurStr = aUnknown.GetCurrent(Itr);
		Unknown.Add(new DString(*CurStr));
		++Itr;
	}
}
