#if !defined(__softop_h)
#define __softop_h

#include "relevanc.h"
#include "options.h"

class CResult;
class CDocWrapper;
class CWordMask;
class CSubdocFilter;
class CAnswer;
class CMaskingAnswer;

class CSoftOperator
{
// consts & types
private:
    class CRelAlgorithm;

// members
private:
    double *Weight;
    double RelevanceThreshold;
    int KeepRelevance;
    CSubdocFilter *Filter;

    void Init(double aFocusWeight,
        double aCorrelationWeight,
        double aProminenceWeight,
        double aLocalityWeight,
        double aOrderWeight,
        double aRelevanceThreshold,
        int aKeepRelevance);

public:
    static void SetUnoClass(CUnivOptions &Options, 
            const CStringListValue &NamedSoftOps);

    CSoftOperator(double aFocusWeight,
	double aCorrelationWeight,
	double aProminenceWeight,
	double aLocalityWeight,
	double aOrderWeight,
	double aRelevanceThreshold,
	int aKeepRelevance);
    // user-named operator
    CSoftOperator(const DString &Name, 
        CSubdocFilter *Filter, 
        CUnivOptions &Options);
    virtual ~CSoftOperator();

	void AddRelevant(CResult *Result, 
	    CDocWrapper *DocWrapper, 
         CWordMask &Mask);
     int IsRelevant(CDocWrapper *DocWrapper, CWordMask &Mask, double &Relevance);

private:
    double GetRelRes(CRelevance *Relevance);
    CMaskingAnswer *NewMaskingAnswer(CDocWrapper *DocWrapper, CWordMask &Mask);
        // returns NULL => doc not relevant
};

#endif

