#include "myfndt.h"
#include "unquote.h"
#include "tempfile.h"
#include <unistd.h>
#include <string.h>

//#define DEBUG_CUnquote

#if defined(DEBUG_CUnquote)
#include <stdio.h>
#endif

CUnquote::CUnquote(const DString &TempDir, const char *Quoted):
    Valid(0)
{
    if (!Quoted)
        return;

    int l = strlen(Quoted);
    if (l < 2 || Quoted[0] != '=')
        return;

    CMemPtr<char> Buf(l);
    memcpy((char *)Buf, Quoted + 1, l);
    if (Unquote(Buf))
        return;

#if defined(DEBUG_CUnquote)
    printf("*** from ***\n%s\n*** to ***\n%s\n", Quoted, (char *)Buf);
#endif

    QueryFileName = GetTemporaryFile(TempDir);
    CFileObj QueryFile(QueryFileName.c_str(), "wt");
    QueryFile.Write((char *)Buf, strlen((char *)Buf));

    Valid = 1;
}

CUnquote::~CUnquote()
{
    if (Valid)
        unlink(QueryFileName.c_str());
}

int CUnquote::IsValid()
{
    return Valid;
}

DString CUnquote::GetQueryFile()
{
    Assert(Valid);
    return QueryFileName;
}

int CUnquote::Unquote(char *Buf)
{
    char *p = Buf;
    while (*p)
    {   if (*p == '+')
            *p = ' ';
        ++p;
    }

    char *q;
    char c;
    while ((q = strchr(Buf, '%')) != NULL)
    {   if (GetChar(q + 1, c))
            return -1;
        if (!c)
            return -1;
        *q = c;
        memmove(q + 1, q + 3, strlen(q + 3) + 1);
        Buf = q + 1;
    }
    return 0;
}

int CUnquote::GetChar(const char *Hexa, char &Out)
{   int high, low;

    if (GetNum(Hexa[0], high))
        return -1;
    if (GetNum(Hexa[1], low))
        return -1;
    
    Out = (char)((16 * high) + low);
    return 0;
}

int CUnquote::GetNum(char Hexa, int &Out)
{
    if (Hexa >= 'a' && Hexa <= 'z')
        Hexa += 'A' - 'a';

    if (Hexa >= '0' && Hexa <= '9')
    {   Out = Hexa - '0';
        return 0;
    }
    else if (Hexa >= 'A' && Hexa <= 'F')
    {   Out = 10 + Hexa - 'A';
        return 0;
    }
    else
        return -1;
}
