#include "myfndt.h"
#include "wordlist.h"
#include "hthreel.h"

CWordList::CStringDatum::CStringDatum(CStorage *aStorage,
	long aTicket):
	CDatum(aStorage, aTicket),
	Str(""), Dirty(0)
{
}

CWordList::CStringDatum::~CStringDatum()
{
}

long CWordList::CStringDatum::Finish()
{
	if (!Dirty)
		return 0;

	CConductor &Conductor = GetConductor();
	long CurTicket = GetTicket();
	if (CurTicket)
		Conductor.Del(CurTicket);

	short StrLen = Str.GetLength();
	CurTicket = Conductor.Add(sizeof(short) + StrLen);
	Conductor.Put(CurTicket,
		0,
		&StrLen,
		sizeof(short));
	if (StrLen > 0)
		Conductor.Put(CurTicket,
			sizeof(short),
			Str.c_str(),
			StrLen);

	SetTicket(CurTicket);
	return CurTicket;
}

void CWordList::CStringDatum::SetStr(const DString &aStr)
{
	Str = aStr;
	Dirty = 1;
}

const DString &CWordList::CStringDatum::GetStr()
{	short StrLen;

	if (!Str.IsEmpty())
		return Str;
	long CurTicket = GetTicket();
	if (!CurTicket)
		return Str;

	CConductor &Conductor = GetConductor();
	Conductor.Get(CurTicket, 0, &StrLen, sizeof(short));
	if (!StrLen)
		return Str;

	CMemPtr<char> Buf(StrLen + 1);
	Conductor.Get(CurTicket,
		sizeof(short),
		(char *)Buf,
		StrLen);
	*((char *)Buf + StrLen) = 0;
	Str = (char *)Buf;
	return Str;
}

CWordList::CWordList(const DString &FPath):
	CStorage(new CThreeLongHandler,
		FPath,
		new CDatumFactoryTemplate<CWordList::CStringDatum>,
		FPath,
		FPath,
		IT_TEMP | IT_UPDATE),
	WordCount(0)
{
}

void CWordList::AddWord(const DString &CurWord,
	long DocCount,
	long OccCount)
{
	//Assert(DocCount > 0 && OccCount > 0);
	CNewKey Key(CThreeLongHandler::NewKey(-DocCount,
		-OccCount,
		WordCount));
	CDatumEnvelopeTemplate<CStringDatum> Datum = static_cast<
			CStringDatum *>(LockAlways(Key));
	Assert(Datum != NULL);
	Datum->SetStr(CurWord);
	WordCount++;
}

void CWordList::GetWord(KEY Key,
	DString &CurWord,
	long &DocCount,
	long &OccCount)
{	long Long[3];

	CThreeLongHandler::GetThreeLong(Key, Long[0], Long[1], Long[2]);
	DocCount = -Long[0];
	OccCount = -Long[1];
	//Assert(DocCount > 0 && OccCount > 0);
	CDatumEnvelopeTemplate<CStringDatum> Datum = static_cast<
			CStringDatum *>(Lock(Key));
	Assert(Datum != NULL);
	CurWord = Datum->GetStr();
}

