#if !defined(__array_h)
#define __array_h

#include <vector.h>
#include <algo.h>

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

template <class DItem>
class DArray : public vector<DItem>
{
// types
public:
	typedef iterator DIterator;

private:
	int Delta;

// not implemented
	operator == (const DArray<DItem> &) const;
	operator != (const DArray<DItem> &) const;
	operator < (const DArray<DItem> &) const;

// methods
public:
	inline DArray(int aLimit, int aDelta);

     // when this constructor is out of line, gcc will not generate it
	DArray<DItem>::DArray(const DArray<DItem> &Other)
//     	Delta(Other.Delta) not handled by Visual C++ 
     {
		Delta = Other.Delta;
	     DArray<DItem> &Kludge = const_cast<DArray<DItem> &>(Other);
		reserve(Kludge.capacity());
		DIterator Itr = Kludge.InitIterator();
		while (Kludge.IsDereferenceable(Itr))
		{	Add(GetCurrent(Itr));
			++Itr;
		}
	}

	virtual ~DArray() { }

	DArray<DItem> & operator = (const DArray<DItem> &Other)
	  { Delta = Other.Delta; vector<DItem>::operator = (Other); return *this; }

	inline int IsEmpty() const { return empty(); }
	inline int IsFull() const { return size() == capacity(); }
	inline int GetCount() const { return size(); }

	inline DIterator InitIterator() { return begin(); }
	inline int IsDereferenceable(DIterator &Itr)
		{ return !(Itr == end()); }

	virtual int Find(const DItem &Item);
	virtual void Add(const DItem &Item);
	virtual int Detach(const DItem &Item); // returns 0 OK, 1 Item not in array
	void DetachByIndex(int Index);
		// different name from Detach by analogy with DIArray 
	inline DItem &operator [] (int Index)
		{ Assert(0 <= Index && (unsigned int)Index < size()); 
			return *(begin() + Index); }

	static DItem &GetCurrent(DIterator &Itr) { return *Itr; }

protected:
	int GetDelta() { return Delta; }
};

template <class DItem>
DArray<DItem>::DArray(int aLimit, int aDelta):
	Delta(aDelta)
{
	reserve(aLimit);
}

template <class DItem>
int DArray<DItem>::Find(const DItem &Item)
{
	iterator iter = find(begin(), end(), Item);
	if (iter == end() ||
		!(*iter == Item))
		return INT_MAX;
	return iter - begin();
}

template <class DItem>
void DArray<DItem>::Add(const DItem &Item)
{
	if (IsFull())
	{	Assert(GetDelta() > 0);
		reserve(size() + GetDelta());
	}
	insert(end(), Item);
}

template <class DItem>
int DArray<DItem>::Detach(const DItem &Item)
{
	iterator iter = find(begin(), end(), Item);
	if (iter == end() ||
		!(*iter == Item))
		return 1;
	erase(iter);
	return 0;
}

template <class DItem>
void DArray<DItem>::DetachByIndex(int Index)
{
	Assert(0 <= Index && Index < size()); 
	iterator iter = begin() + Index;
	erase(iter);
}

#endif

