#if !defined(__arrayb_h)
#define __arrayb_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

template <class DItem>
class DArray : public TArrayAsVector<DItem>
{
public:
	typedef TArrayAsVectorIterator<DItem> DIterator;

private:
// not implemented
	operator == (const DArray<DItem> &) const;
	operator != (const DArray<DItem> &) const;
	operator < (const DArray<DItem> &) const;

public:
	DArray(int aLimit, int aDelta):
		TArrayAsVector<DItem>(aLimit - 1, 0, aDelta) { }
	DArray(const DArray<DItem> &Other);
	virtual ~DArray() { }

	DArray<DItem> &operator = (const DArray<DItem> &Other);

	inline int GetCount() const
		{ return GetItemsInContainer(); }

	inline DIterator InitIterator() const
		{ return DIterator(*this); }
	inline int IsDereferenceable(DIterator &Itr) const
		{ return Itr != NULL; }

	inline void Add(const DItem &Item);
	void DetachByIndex(int Index)
		{ Detach(Index); }

	static DItem GetCurrent(DIterator &Itr)
		{ return Itr.Current(); }
};

template <class DItem>
DArray<DItem>::DArray(const DArray<DItem> &Other):
	TArrayAsVector<DItem>(Other.ArraySize() - 1,
		0,
		Other.Data.GetDelta())
{
	DIterator Itr = Other.InitIterator();
	while (Other.IsDereferenceable(Itr))
	{	Add(Itr.Current());
		++Itr;
	}
}

template <class DItem>
DArray<DItem> &DArray<DItem>::operator = (const DArray<DItem> &Other)
{
	if (&Other == this)
		return *this;

	int Count = GetCount();
	for (int i = Count - 1; i >= 0; i--)
		Destroy(i);

	DIterator Itr = InitIterator();
	while (Other.IsDereferenceable(Itr))
	{	Add(Itr.Current());
		++Itr;
	}

	return *this;
}

template <class DItem>
void DArray<DItem>::Add(const DItem &Item)
{
	if (!TArrayAsVector<DItem>::Add(Item))
		throw DMemoryException();
}

#endif

