#if !defined(__iarray_h)
#define __iarray_h

#include "vector.h"
#include "algo.h"
#include "dumbptr.h"

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

template <class DItem>
class DIArray : public vector<DDumbPtr<DItem> >
{
// types
public:
	typedef iterator DIterator;

private:
	int Delta;

// not implemented
	DIArray<DItem> & operator = (const DIArray<DItem> &);
	operator == (const DIArray<DItem> &) const;
	operator != (const DIArray<DItem> &) const;
	operator < (const DIArray<DItem> &) const;

// methods
public:
	inline DIArray(int aLimit, int aDelta);
	inline DIArray(const DIArray<DItem> &Other)
		{ Assert(0); }
	virtual ~DIArray();

	DItem *GetCurrent(DIterator &Itr)
	    // cannot be static - gcc does not handle static template members
	{
	    DDumbPtr<DItem> &Item = *Itr;
	    return Item.p;
	}

	inline int IsEmpty() const { return empty(); }
	inline int IsFull() const { return size() == capacity(); }
	inline int GetCount() const { return size(); }

	inline DIterator Begin() { return begin(); }
	inline DIterator End() { return end(); }

	inline DIterator InitIterator() { return begin(); }
	inline int IsDereferenceable(DIterator &Itr)
		{ return !(Itr == end()); }

	virtual int Find(DItem *Item);
	virtual void Add(DItem *NewItem); // takes ownership of NewItem
	virtual int Detach(DItem *Item); // returns 0 OK, 1 Item not in array
	void DetachByIndex(int Index);
		// different name from Detach so that Detach(0) is unambiguous
	inline DItem *& operator [] (int Index);

protected:
	int GetDelta() const { return Delta; }
};

template <class DItem>
DIArray<DItem>::DIArray(int aLimit, int aDelta):
	Delta(aDelta)
{
	reserve(aLimit);
}

template <class DItem>
DIArray<DItem>::~DIArray()
{
	iterator iter = begin();
	while (iter != end())
	{	DDumbPtr<DItem> &ref = *iter;
		delete ref.p;
		ref.p = 0;
		++iter;
	}
}

template <class DItem>
int DIArray<DItem>::Find(DItem *Item)
{
	iterator iter = find(begin(), end(),
			DDumbPtr<DItem>(Item));
	if (iter == end() ||
		!(*iter == DDumbPtr<DItem>(Item)))
		return INT_MAX;
	return iter - begin();
}

template <class DItem>
void DIArray<DItem>::Add(DItem *NewItem)
{
	CObjPtr<DItem> safe_item = NewItem;
	if (IsFull())
	{	Assert(GetDelta() > 0);
		reserve(size() + GetDelta());
	}
	insert(end(), DDumbPtr<DItem>(NewItem));
	safe_item.Drop();
}

template <class DItem>
int DIArray<DItem>::Detach(DItem *Item)
{
	iterator iter = find(begin(), end(),
			DDumbPtr<DItem>(Item));
	if (iter == end() ||
		!(*iter == DDumbPtr<DItem>(Item)))
		return 1;
	erase(iter);
	return 0;
}

template <class DItem>
void DIArray<DItem>::DetachByIndex(int Index)
{
	Assert(0 <= Index && (unsigned int)Index < size());
	iterator iter = begin() + Index;
	erase(iter);
}

template <class DItem>
DItem *& DIArray<DItem>::operator [] (int Index)
{
	Assert(0 <= Index && (unsigned int)Index < size());
	DDumbPtr<DItem> &ref = *(begin() + Index);
	return ref.p;
}

#endif
