#if !defined(__iarrayb_h)
#define __iarrayb_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

template <class DItem>
class DIArray : public TIArrayAsVector<DItem>
{
public:
	typedef TIArrayAsVectorIterator<DItem> DIterator;

private:
// not implemented
	operator == (const DIArray<DItem> &) const;
	operator != (const DIArray<DItem> &) const;
	operator < (const DIArray<DItem> &) const;

public:
	DIArray(int aLimit, int aDelta):
		TIArrayAsVector<DItem>(aLimit - 1, 0, aDelta) { }
	DIArray(const DIArray<DItem> &Other);
	virtual ~DIArray() { }

	DIArray<DItem> &operator = (const DIArray<DItem> &Other);

	inline int GetCount() const
		{ return GetItemsInContainer(); }

	inline DIterator InitIterator() const
		{ return DIterator(*this); }
	inline int IsDereferenceable(DIterator &Itr) const
		{ return Itr != NULL; }

	inline void Add(DItem *Item);
	void DetachByIndex(int Index)
		{ Detach(Index); }

	static DItem *GetCurrent(DIterator &Itr)
		{ return Itr.Current(); }
};

template <class DItem>
DIArray<DItem>::DIArray(const DIArray<DItem> &Other):
	TIArrayAsVector<DItem>(Other.ArraySize() - 1,
		0,
		Other.Data.GetDelta())
{
	DIterator Itr = Other.InitIterator();
	while (Other.IsDereferenceable(Itr))
	{	Add(new DItem(*Itr.Current()));
		++Itr;
	}
}

template <class DItem>
DIArray<DItem> &DIArray<DItem>::operator = (const DIArray<DItem> &Other)
{
	if (&Other == this)
		return *this;

	int Count = GetCount();
	for (int i = Count - 1; i >= 0; i--)
		Destroy(i);

   DIterator Itr = InitIterator();
	while (Other.IsDereferenceable(Itr))
	{	Add(new DItem(*Itr.Current()));
		++Itr;
	}

	return *this;
}

template <class DItem>
void DIArray<DItem>::Add(DItem *Item)
{
	if (!TIArrayAsVector<DItem>::Add(Item))
		throw DMemoryException();
}

#endif
