#if !defined(__isarray_h)
#define __isarray_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

#include "vector.h"
#include "algo.h"
#include "dumbptr.h"
#include "iarray.h"

template <class DItem>
class DISArray : public DIArray<DItem>
{
// methods
public:
	inline DISArray(int aLimit, int aDelta):
		DIArray<DItem>(aLimit, aDelta) { }
	inline DISArray(const DISArray<DItem> &Other):
		DIArray<DItem>(Other) { }
	virtual ~DISArray() { }

	virtual int Find(DItem *Item);
	virtual void Add(DItem *NewItem); // takes ownership of NewItem
	virtual int Detach(DItem *Item); // returns 0 OK, 1 Item not in array
};

template <class DItem>
int DISArray<DItem>::Find(DItem *Item)
{
	iterator iter = lower_bound(begin(), end(),
			DDumbPtr<DItem>(Item));
	if (iter == end() ||
		!(*iter == DDumbPtr<DItem>(Item)))
		return INT_MAX;
	return iter - begin();
}

template <class DItem>
void DISArray<DItem>::Add(DItem *NewItem)
{
	CObjPtr<DItem> safe_item = NewItem;

	if (IsFull())
	{	Assert(GetDelta() > 0);
		reserve(size() + GetDelta());
	}

	iterator iter = lower_bound(begin(), end(),
			DDumbPtr<DItem>(NewItem));
	insert(iter, DDumbPtr<DItem>(NewItem));
	safe_item.Drop();
}

template <class DItem>
int DISArray<DItem>::Detach(DItem *Item)
{
	iterator iter = lower_bound(begin(), end(),
			DDumbPtr<DItem>(Item));
	if (iter == end() ||
		!(*iter == DDumbPtr<DItem>(Item)))
		return 1;
	erase(iter);
	return 0;
}

#endif

