#if !defined(__sarray_h)
#define __sarray_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

#include "vector.h"
#include "algo.h"
#include "array.h"

template <class DItem>
class DSArray : public DArray<DItem>
{
// methods
public:
	inline DSArray(int aLimit, int aDelta):
		DArray<DItem>(aLimit, aDelta) { }
	inline DSArray(const DSArray<DItem> &Other):
		DArray<DItem>(Other) { }
	virtual ~DSArray() { }

	virtual void Add(const DItem &Item);
	virtual int Detach(const DItem &Item); // returns 0 OK, 1 Item not in array
	virtual int Find(const DItem &Item);
};

template <class DItem>
void DSArray<DItem>::Add(const DItem &Item)
{
	if (IsFull())
	{	Assert(GetDelta() > 0);
		reserve(size() + GetDelta());
	}

	iterator iter = lower_bound(begin(), end(), Item);
	insert(iter, Item);
}

template <class DItem>
int DSArray<DItem>::Detach(const DItem &Item)
{
	iterator iter = lower_bound(begin(), end(), Item);
	if (iter == end() ||
		!(*iter == Item))
		return 1;
	erase(iter);
	return 0;
}

template <class DItem>
int DSArray<DItem>::Find(const DItem &Item)
{
	iterator iter = lower_bound(begin(), end(), Item);
	if (iter == end() ||
		!(*iter == Item))
		return INT_MAX;
	return iter - begin();
}

#endif

