#include "myfndt.h"
#include "conduct.h"
#include "htwolong.h"

#include <stdlib.h>
#if defined(__GNUG__)
#include <unistd.h>
#endif

/* profile 27. 12. 96 (Linux): 1 file, 949 250 bytes
	free version: 9 min, occstore.dat 399 424 bytes 
	checked version: 9 min, occstore.dat 458 744 bytes */

#define VF_CHECKED	1

struct CConductor::SDataHead
{	unsigned int VerFlags;
	int Stuff;
};

const unsigned long CConductor::FreeFlag = 1;
const unsigned long CConductor::PrecFreeFlag = 2;
#if defined(CHECKED_VERSION)
const unsigned int CConductor::FreeSig = 0x4642;
const unsigned int CConductor::UsedSig = 0x5542;
#endif

const long CConductor::InitialSize = 1000;
const long CConductor::ShrinkMin = sizeof(SArena) +
		2 * sizeof(long);
	// occurence is at least 3 * sizeof(long); see Resize
const short CConductor::MinTicketSize = sizeof(SArena) +
		sizeof(ulong);
const long CConductor::GrowDelta = 1000;
const short CConductor::MinCarved = sizeof(SArena) +
		3 * sizeof(long);
	// occurence is at least 3 * sizeof(long)

CConductor::CConductor(const DString &aFreeListFile, 
     const DString &aDataFile,
	ushort Opt):
	FreeList(new CTwoLongHandler(), aFreeListFile, Opt),
	DataFile(NULL),
	DataFileName(aDataFile),
	CreateOptions(Opt)
{
	// check const params
	Assert(!(InitialSize % 4) &&
		!(GrowDelta % 4) &&
		ShrinkMin >= MinTicketSize && // see Resize
		GrowDelta >= MinTicketSize && // see GrowList
		MinCarved >= MinTicketSize); // see Carve

	if (OpenData(aDataFile, Opt, DataFileName, DataFile))
		CreateList();
	else
		ReadHead();
}

CConductor::~CConductor()
{
	if (CreateOptions & IT_TEMP)
	{	DataFile.Close();
		unlink(DataFileName.c_str());
	}
}

long CConductor::Add(long TicketLen)
{	long CurSize;
	long Offs;

#if defined(DEBUG_stem)
	Check();
#endif
	ulong Size = RoundTicketLen(TicketLen);

	CNewKey TestKey(CTwoLongHandler::NewKey(Size, 0));
	while (1)
	{    CLockedKey Key(FreeList.LockKeyAfter(TestKey), &FreeList);
		if (Key)
		{    CTwoLongHandler::GetTwoLong(Key, CurSize, Offs);
			Key = NULL;
			return SplitBlock(Offs, Size);
		}
		GrowList(Size);
	}
}

void CConductor::Del(long Ticket)
{  long Offs;
	long FreeOffs;
	SArena UsedBlock;
	SArena PrevBlock;

#if defined(DEBUG_stem)
	Check();
#endif
	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0);
	ulong UsedOffs = Ticket - sizeof(SArena);
	ReadUsedArena(UsedBlock, UsedOffs, &Offs);
	ulong Size = GetSize(UsedBlock.ArenaInfo);
	if (UsedBlock.ArenaInfo & PrecFreeFlag)
	{	// Offs valid
		ReadFreeArena(PrevBlock, Offs);
		Assert(Offs + GetSize(PrevBlock.ArenaInfo) == UsedOffs);
		ulong PrevSize = GetSize(PrevBlock.ArenaInfo);
		Size += PrevSize;
		FreeFreeBlock(Offs, PrevSize);
		FreeOffs = Offs;
	}
	else
		FreeOffs = UsedOffs;

	FreeSub(FreeOffs, Size);
}

long CConductor::Resize(long Ticket, long TicketLen)
{	SArena UsedArena;
	ulong UsedOffs;
	ulong OldSize;

#if defined(DEBUG_stem)
	Check();
#endif
	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0);
	InitResizeLocals(Ticket, UsedArena,	UsedOffs, OldSize);
	ulong NewSize = RoundTicketLen(TicketLen);
	if (NewSize <= OldSize) // shrink
	{	ulong Delta = OldSize - NewSize;
		if (Delta > ShrinkMin)
			FreeSub(UsedOffs + NewSize, Delta);
		return Ticket;
	}

	ulong RealSize = GrowInPlace(UsedArena,
		UsedOffs,
		OldSize,
		NewSize,
		0);
	if (RealSize >= NewSize)
		return Ticket;

	long Ticket2 = Add(TicketLen);
	CopyContent(Ticket2, Ticket, OldSize - sizeof(SArena));
	Del(Ticket);
#if defined(DEBUG_stem)
	Check();
#endif
	return Ticket2;
}

long CConductor::SoftGrow(long Ticket, long TicketLen)
{	SArena UsedArena;
	ulong UsedOffs;
	ulong OldSize;

	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0);
	InitResizeLocals(Ticket, UsedArena,	UsedOffs, OldSize);
	ulong NewSize = RoundTicketLen(TicketLen);
	// never shrinks
	if (NewSize <= OldSize)
		return SizeToLen(OldSize);

	long RealSize = GrowInPlace(UsedArena,
		UsedOffs,
		OldSize,
		NewSize,
		1);
	return SizeToLen(RealSize);
}

void *CConductor::Alloc(long Ticket)
{	SArena Arena;

	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0);
	ReadUsedArena(Arena, Ticket - sizeof(SArena), NULL);
	ulong Size = GetSize(Arena.ArenaInfo);

	short StuffLen = static_cast<short>(Size);
	Assert((ulong)StuffLen == Size);
	CMemPtr<void> Stuff(StuffLen);

	ReadData(Ticket, (void *)Stuff, StuffLen);
	return Stuff.Drop();
}

void CConductor::Get(long Ticket,
	long From,
	void *Stuff,
	short Len)
{
#if defined(DEBUG_stem)
	Check();
#endif
	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0 &&
		From >= 0 && Len > 0);
	ReadData(Ticket + From, Stuff, Len);
}

void CConductor::Put(long Ticket,
	long From,
	const void *Stuff,
	short Len)
{
#if defined(DEBUG_stem)
	Check();
#endif
	Assert(Ticket > sizeof(SArena) &&
		(Ticket % sizeof(long)) == 0 &&
		From >= 0 && Len >= 0);
	WriteData(Ticket + From, Stuff, Len);
#if defined(DEBUG_stem)
	Check();
#endif
}

void CConductor::Copy(long DestTicket, 
    long DestFrom,
    long SrcTicket,
    long SrcFrom,
    long Size)
{ 
    Assert(DestTicket > sizeof(SArena) &&
	   (DestTicket % sizeof(long)) == 0 &&
	   DestFrom >= 0 && 
	   SrcTicket > sizeof(SArena) &&
	   (SrcTicket % sizeof(long)) == 0 &&
	   SrcFrom >= 0 &&	   
	   Size >= 0);
    CopyContent(DestTicket + DestFrom, SrcTicket + SrcFrom, Size); 
}

void CConductor::StoreAll()
{
#if defined(DEBUG_stem)
	Check();
#endif
	FreeList.StoreAll();
#if defined(DEBUG_stem)
	Check();
#endif
	DataFile.Flush(); //returns error ???
}

#if defined(DEBUG_stem)
void CConductor::Check()
{	long CurSize1;
	long CurSize2;
	long Offs;
	SArena FreeArena;

	CLockedKey Key1(FreeList.LockFirstKey(), &FreeList);
	while (Key1 != NULL)
	{    CTwoLongHandler::GetTwoLong(Key1, CurSize1, Offs);
		ReadFreeArena(FreeArena, Offs);
		CurSize2 = GetSize(FreeArena.ArenaInfo);
		Assert(CurSize1 == CurSize2);
		CLockedKey Key2(FreeList.LockNextKey(Key1), &FreeList);
		Key1 = Key2.Drop();
	}

}
#endif

void CConductor::ReadData(long Offs, void *Stuff, short Len)
{
	DataFile.Seek(Offs);
	DataFile.Read(Stuff, Len);
}

void CConductor::WriteData(long Offs, const void *Stuff, short Len)
{
	DataFile.Seek(Offs);
	DataFile.Write(Stuff, Len);
}

void CConductor::CreateList()
{	SDataHead DataHead;
	SArena FList;
	SArena Sentinel;

	Assert(FreeList.IsEmpty());
	Assert(InitialSize > sizeof(SDataHead) +
		sizeof(SArena) + // FList
		MinTicketSize + // for users
		sizeof(SArena)); // sentinel
	ulong BigBlockSize = InitialSize - sizeof(SDataHead) -
			2 * sizeof(SArena);

	DataFile.Seek(0);
	DataHead.VerFlags = 0;
#if defined(CHECKED_VERSION)
	DataHead.VerFlags |= VF_CHECKED;
#endif
	DataHead.Stuff = 0;
	DataFile.Write(&DataHead, sizeof(SDataHead));
	AddFreeArena(FList, 0, FreeFlag, sizeof(SDataHead));
	DataFile.Write(&FList, sizeof(SArena));

	DataFile.WriteZero(BigBlockSize);

	FillUsedArena(Sentinel, 0, 0);
	DataFile.Write(&Sentinel, sizeof(SArena));

	FreeSub(sizeof(SDataHead) + sizeof(SArena), BigBlockSize);
}

void CConductor::ReadHead()
{  SDataHead DataHead;

	DataFile.Seek(0);
	DataFile.Read(&DataHead, sizeof(SDataHead));
#if defined(CHECKED_VERSION)
	if (!(DataHead.VerFlags & VF_CHECKED))
#else
	if (DataHead.VerFlags & VF_CHECKED)
#endif
		throw CReadException();
}

void CConductor::GrowList(ulong MinSize)
{	SArena NewSentinel;
	SArena OldSentinel;

	DataFile.Seek(0, SEEK_END);
	long OldSentinelTicket = DataFile.Tell();

	ulong FreeSize = GrowDelta >= MinSize ? GrowDelta : MinSize;
	DataFile.WriteZero(FreeSize - sizeof(SArena));

	FillUsedArena(NewSentinel, 0, 0);
	DataFile.Write(&NewSentinel, sizeof(SArena));

	long OldSentinelOffs = OldSentinelTicket - sizeof(SArena);
	ReadUsedArena(OldSentinel, OldSentinelOffs, NULL);
	OldSentinel.ArenaInfo = GetArenaInfo(FreeSize,
		GetFlags(OldSentinel.ArenaInfo));
	WriteArena(OldSentinel, OldSentinelOffs);
	Del(OldSentinelTicket);
}

long CConductor::SplitBlock(long Offs, ulong Len)
{	SArena FreeArena;

	ReadFreeArena(FreeArena, Offs);
	ulong CarvedSize = Carve(Offs, FreeArena, Len);
	FillUsedArena(FreeArena, CarvedSize, 0);
	WriteArena(FreeArena, Offs);
	return Offs + sizeof(SArena);
}

ulong CConductor::Carve(long Offs, SArena &Arena, ulong Len)
{
	ulong CurSize = GetSize(Arena.ArenaInfo);
	Assert(CurSize >= Len);
	FreeFreeBlock(Offs, CurSize);
	if (Len + MinCarved <= CurSize)
	{	FreeSub(Offs + Len, CurSize - Len);
		return Len;
	}
	else
	{	SArena NextArena;

		long NextOffs = Offs + CurSize;
		ReadUsedArena(NextArena, NextOffs, NULL);
		NextArena.ArenaInfo &= ~PrecFreeFlag;
		WriteArena(NextArena, NextOffs);
		return CurSize;
	}
}

void CConductor::FreeSub(long Offs, ulong Len)
{  SArena NextBlock;
	SArena Arena;

	Assert(Len >= MinTicketSize);
	long NextOffs = Offs + Len;
	ReadArena(NextBlock, NextOffs);
	if (NextBlock.ArenaInfo & FreeFlag)
	{	ulong NextBlockSize = GetSize(NextBlock.ArenaInfo);
		FreeFreeBlock(NextOffs, NextBlockSize);
		NextOffs += NextBlockSize;
		ReadUsedArena(NextBlock, NextOffs, NULL);
		Len += NextBlockSize;
	}

	// now block before NextBlock will be freed
	WriteOffs(Offs, NextOffs - sizeof(long));
	NextBlock.ArenaInfo |= PrecFreeFlag;
	WriteArena(NextBlock, NextOffs);

	AddFreeArena(Arena, Len, FreeFlag, Offs);
   WriteArena(Arena, Offs);
}

void CConductor::FreeFreeBlock(long Offs, ulong Size)
{
#if defined(DEBUG_CConduct)
	LONG dummy;
#endif

	CNewKey Key(CTwoLongHandler::NewKey(Size, Offs));
#if defined(DEBUG_invalnode)
     FreeList.Check();
#endif
	short KnownRet = FreeList.Del(Key);
	Assert(!KnownRet);
#if defined(DEBUG_invalnode)
     FreeList.Check();
#endif
#if defined(DEBUG_CConduct)
	KnownRet = FreeList.GetData(Key, dummy);
	Assert(KnownRet);
#endif
}

void CConductor::ReadUsedArena(SArena &Arena, long Offs,
		long *PrevOffs)
{
	Assert(Offs >= sizeof(SArena));
	if (PrevOffs)
	{	DataFile.Seek(Offs - sizeof(long));
		DataFile.Read(PrevOffs, sizeof(long));
	}
	else
		DataFile.Seek(Offs);
	DataFile.Read(&Arena, sizeof(SArena));
#if defined(CHECKED_VERSION)
	if (!(Arena.ArenaInfo & PrecFreeFlag))
	{	if (PrevOffs)
			*PrevOffs = -1;
   }
	Assert(Arena.Signature == UsedSig);
#endif
}

void CConductor::ReadFreeArena(SArena &Arena, long Offs)
{
	DataFile.Seek(Offs);
	DataFile.Read(&Arena, sizeof(SArena));
#if defined(CHECKED_VERSION)
	Assert(Arena.Signature == FreeSig);
#endif
}

void CConductor::ReadArena(SArena &Arena, long Offs)
{
	DataFile.Seek(Offs);
	DataFile.Read(&Arena, sizeof(SArena));
#if defined(CHECKED_VERSION)
	Assert((!(Arena.ArenaInfo & FreeFlag) && Arena.Signature ==
			UsedSig) || ((Arena.ArenaInfo & FreeFlag) &&
			Arena.Signature == FreeSig));
#endif
}

void CConductor::FillUsedArena(SArena &Arena, ulong Size,
		ulong Flags)
{
	Assert(!(Flags & FreeFlag));
	Arena.ArenaInfo = GetArenaInfo(Size, Flags);
#if defined(CHECKED_VERSION)
	Arena.Signature = UsedSig;
	Arena.Stuff = 0;
#endif
}

void CConductor::AddFreeArena(SArena &Arena, ulong Size,
		ulong Flags, long Offs)
{
#if defined(DEBUG_CConduct)
	static LONG Dif = 0;
#endif

	Assert((Flags & FreeFlag) ? 1 : 0);
	Arena.ArenaInfo = GetArenaInfo(Size, Flags);
#if defined(CHECKED_VERSION)
	Arena.Signature = FreeSig;
	Arena.Stuff = 0;
#endif

	CNewKey Key(CTwoLongHandler::NewKey(Size, Offs));
#if defined(DEBUG_invalnode)
     FreeList.Check();
#endif
	FreeList.Add(Key,
#if !defined(DEBUG_CConduct)
		0);
#else
		Dif++);
#endif
#if defined(DEBUG_invalnode)
     FreeList.Check();
#endif
}

ulong CConductor::RoundTicketLen(long TicketLen)
{
	Assert(TicketLen > 0);
	long Delta = TicketLen % 4;
	if (Delta)
		TicketLen += 4 - Delta;
	TicketLen += sizeof(SArena);
	Assert(TicketLen > 0);
	if (TicketLen < MinTicketSize)
		TicketLen = MinTicketSize;
	return TicketLen;
}

long CConductor::SizeToLen(ulong Size)
{
	long Len = Size;
	Len -= sizeof(SArena);
	Assert(Len > 0);
	return Len;
}

void CConductor::CopyContent(long Dest, long Src, long Size)
{
	Assert(Size > 0);
	short CurSize = 1024;
	CMemPtr<char> Buf(CurSize);

	while (Size > 0)
	{	if (Size < CurSize)
			CurSize = static_cast<short>(Size);

		ReadData(Src, (char *)Buf, CurSize);
		Src += CurSize;
		WriteData(Dest, (char *)Buf, CurSize);
		Dest += CurSize;

		Size -= CurSize;
	}
}

void CConductor::InitResizeLocals(long Ticket,
	SArena &UsedArena,
	ulong &UsedOffs,
	ulong &OldSize)
{
	Assert(Ticket > sizeof(SArena));
	UsedOffs = Ticket - sizeof(SArena);
	ReadUsedArena(UsedArena, UsedOffs, NULL);
	OldSize = GetSize(UsedArena.ArenaInfo);
	Assert(OldSize > sizeof(SArena));
}

ulong CConductor::GrowInPlace(SArena &UsedArena,
	ulong UsedOffs,
	ulong OldSize,
	ulong NewSize,
	short Soft)
{	SArena NextArena;

	long NextOffs = UsedOffs + OldSize;
	ReadArena(NextArena, NextOffs);
	if (NextArena.ArenaInfo & FreeFlag)
	{  ulong ExtraNeeded = NewSize - OldSize;
		ulong NextSize = GetSize(NextArena.ArenaInfo);
		if (Soft && ExtraNeeded > NextSize)
			ExtraNeeded = NextSize;
		if (ExtraNeeded <= NextSize)
		{	ulong CarvedSize = Carve(NextOffs,
				NextArena,
				ExtraNeeded);
			FillUsedArena(UsedArena,
				OldSize + CarvedSize,
				GetFlags(UsedArena.ArenaInfo));
			WriteArena(UsedArena, UsedOffs);
			return OldSize + CarvedSize;
		}
	}
	return OldSize;
}


