#if !defined(__conduct_h)
#define __conduct_h

#include "tl.h"

//#define CHECKED_VERSION
//#define DEBUG_stem

class CConductor
{
private:
// types & const params

	struct SDataHead;

	struct SArena
	{	unsigned long ArenaInfo;
#if defined(CHECKED_VERSION)
		unsigned int Signature;
		int Stuff;
#endif
	};

	static const ulong FreeFlag;
	static const ulong PrecFreeFlag;
#if defined(CHECKED_VERSION)
	static const unsigned int FreeSig;
	static const unsigned int UsedSig;
#endif

	static const long InitialSize;
	static const long ShrinkMin;
	static const short MinTicketSize;
	static const long GrowDelta;
	static const short MinCarved;

// data members
private:
	CBook FreeList;
	CFileObj DataFile;
	DString DataFileName;
	ushort CreateOptions;

private:
	CConductor(const CConductor &); // not implemented
	CConductor & operator = (const CConductor &); // not implemented

public:
	CConductor(const DString &aFreeListFile, // only path if Opt & IT_TEMP
		const DString &aDataFile, // only path if Opt & IT_TEMP
		ushort Opt);
	virtual ~CConductor();

	long Add(long TicketLen);
	void Del(long Ticket);
	long Resize(long Ticket, long TicketLen);
	long SoftGrow(long Ticket, long TicketLen);
		// returns real new size

	void Get(long Ticket, long From, void *Stuff, short Len);
	void Put(long Ticket, long From, const void *Stuff,
			short Len);
	void *Alloc(long Ticket);
	void Copy(long DestTicket, 
	    long DestFrom,
	    long SrcTicket,
	    long SrcFrom,
         long Size);

	void StoreAll();

protected: // for CSegmentConductor
	void ReadData(long Offs, void *Stuff, short Len);
	void WriteData(long Offs, const void *Stuff, short Len);

private:
	void CreateList();
	void ReadHead();

	void GrowList(ulong MinSize);
	long SplitBlock(long Offs, ulong Len);
	ulong Carve(long Offs, SArena &Arena, ulong Len);
	void FreeSub(long Offs, ulong Len);

	void FreeFreeBlock(long Offs, ulong Size);

	// reads valid used arena from disk
	// Arena is out param
	// if PrevOffs is not NULL, it out param;
	// it is valid only when returned arena has PrecFreeFlag
	void ReadUsedArena(SArena &Arena, long Offs, long *PrevOffs);

	// reads valid free arena from disk
	// Arena is out param
	void ReadFreeArena(SArena &Arena, long Offs);
		// inline?

	void ReadArena(SArena &Arena, long Offs);
		// inline?

	void FillUsedArena(SArena &Arena, ulong Size, ulong Flags);

	// fills arena and adds it into free list;
	// does no write into data file!
	void AddFreeArena(SArena &Arena, ulong Size,	ulong Flags,
			long Offs);

	void WriteArena(SArena &Arena, long Offs)
		{ DataFile.Seek(Offs); DataFile.Write(&Arena, sizeof(SArena)); }

	void WriteOffs(long What, long Where)
		{ DataFile.Seek(Where); DataFile.Write(&What, sizeof(long)); }

	static ulong GetArenaInfo(ulong Size, ulong Flags)
		{ return Size | Flags; }
	static ulong GetSize(ulong ArenaInfo)
		{ return ArenaInfo & ~(FreeFlag | PrecFreeFlag); }
	static ulong GetFlags(ulong ArenaInfo)
		{ return ArenaInfo & (FreeFlag | PrecFreeFlag); }

	static ulong RoundTicketLen(long TicketLen);
	static long SizeToLen(ulong Size);
	void CopyContent(long Dest, long Src, long Size);

	void InitResizeLocals(long Ticket,
		SArena &UsedArena, // out param
		ulong &UsedOffs, // out param
		ulong &OldSize); // out param
	ulong GrowInPlace(SArena &UsedArena, // in param, modified
		ulong UsedOffs,
		ulong OldSize,
		ulong NewSize,
		short Soft);
		// returns real new size
};

#endif

