#if !defined(__datum_h)
#define __datum_h

#if defined(EXTERNAL_TEMPLATES)
#pragma interface
#endif

class CConductor;
class CStorage;
class CDatumEnvelope;

class CDatum
{
friend class CDatumEnvelope; // uses Count & Storage

private:
	CStorage *Storage;
	long Ticket; // 0 means no data on disk yet
	int Count;

public:
	CDatum(CStorage *aStorage, long aTicket = 0);
	virtual ~CDatum();

	virtual long Finish() = 0;
		// called by CStorage
		// returns new ticket or 0 if it does not want to be saved

	virtual void Remove();
		// called by CStorage
		// deletes ticket from conductor; ancestors may implement
		// fancier management

protected:
	CConductor &GetConductor();
	long GetTicket() { return Ticket; }
	void SetTicket(long aTicket) { Ticket = aTicket; }
};

class CDatumFactory
{
public:
	CDatumFactory() { }
	virtual ~CDatumFactory() { }

	virtual CDatum *Create(CStorage *aStorage,
		long aTicket) = 0;
};

template <class CConcreteDatum>
class CDatumFactoryTemplate : public CDatumFactory
{
public:
	CDatumFactoryTemplate() { }
	virtual ~CDatumFactoryTemplate() { }

	virtual CDatum *Create(CStorage *aStorage,
			long aTicket) { return new CConcreteDatum(
			aStorage, aTicket); }
};

class CDatumEnvelope
{
protected:
	CDatum *Datum;

public:
	CDatumEnvelope(CDatum *aDatum);
	CDatumEnvelope(const CDatumEnvelope &Src);
	virtual ~CDatumEnvelope();
	CDatumEnvelope & operator = (const CDatumEnvelope &Src);

   short IsEmpty() { return (short)(Datum == NULL ? 1 : 0); }
};

template <class CConcreteDatum>
class CDatumEnvelopeTemplate : public CDatumEnvelope
{
public:
	CDatumEnvelopeTemplate(CConcreteDatum *aDatum):
		CDatumEnvelope(aDatum) { }
	CDatumEnvelopeTemplate(const CDatumEnvelopeTemplate &Src):
   			CDatumEnvelope(Src) { }
	virtual ~CDatumEnvelopeTemplate() { }
	CDatumEnvelopeTemplate & operator = (const CDatumEnvelopeTemplate &Src)
		{ CDatumEnvelope::operator = (Src); return *this; }

	CConcreteDatum * operator -> () { return
			static_cast<CConcreteDatum *>(Datum); }
	operator CConcreteDatum *() { return
			static_cast<CConcreteDatum *>(Datum); }
};

#endif

