#include "myfndt.h"
#include "remtab.h"
#include "storag.h"

const int CRemTable::DefaultLimit = 10;
const int CRemTable::DefaultDelta = 10;

CRemTable::CRemTable(int Limit, int Delta, CKeyHandler *aKeyHandlerAlias):
	Table(Limit, Delta),
	InvertedTable(Limit, Delta),
	KeyHandlerAlias(aKeyHandlerAlias)
{
}

CRemTable::~CRemTable()
{
}

CAssoc *CRemTable::Get(KEY Key)
{
	int Index = GetTableIndex(Key);
	return Index == -1 ? 0 : Table[Index];
}

CAssoc *CRemTable::Get(CDatum *Datum)
{
	CInvertedAssoc *Assoc = GetInverted(Datum);
	if (!Assoc)
		return NULL;
	return Get(Assoc->GetKey());
}

void CRemTable::Add(CAssoc *Assoc)
{
	CInvertedAssoc *Inverted = new CInvertedAssoc(KeyHandlerAlias,
		Assoc->GetKey(),
		Assoc->GetDatum());
	Table.Add(Assoc);
	InvertedTable.Add(Inverted);
}

void CRemTable::Detach(CAssoc *Assoc)
{
	Assert(Assoc != NULL);
	CInvertedAssoc *Inverted = GetInverted(Assoc->GetDatum());
	Assert(Inverted != NULL);
	Table.Detach(Assoc);
	InvertedTable.Detach(Inverted);
   delete Inverted;
}

void CRemTable::StoreAll(CStorage *Storage)
{
	DISArray<CAssoc>::DIterator Itr = Table.InitIterator();
	while (Table.IsDereferenceable(Itr))
	{	CAssoc *Current = Table.GetCurrent(Itr);
		Storage->Store(Current);
		++Itr;
	}
}

CInvertedAssoc *CRemTable::GetInverted(CDatum *Datum)
{
	int Count = InvertedTable.GetCount();
	if (!Count)
		return NULL;

	int Min = 0;
	int Max = Count;
	while (Min + 1 < Max)
	{	int Ave = (Min + Max) / 2;
		if (((ulong)Datum) < ((ulong)InvertedTable[
				Ave]->GetDatum()))
			Max = Ave;
		else
			Min = Ave;
	}
	if (Datum != InvertedTable[Min]->GetDatum())
		return NULL;

	return InvertedTable[Min];
}

int CRemTable::GetTableIndex(KEY Key)
{
	int Count = Table.GetCount();
	if (!Count)
		return -1;

	int Min = 0;
	int Max = Count;
	while (Min + 1 < Max)
	{	int Ave = (Min + Max) / 2;
		if (KeyHandlerAlias->Compare(Key, Table[Ave]->GetKey()) < 0)
			Max = Ave;
		else
			Min = Ave;
	}
	if (KeyHandlerAlias->Compare(Key, Table[Min]->GetKey()))
		return -1;
	return Min;
}

KEY CRemTable::GetFirstKey()
{
	if (!Table.GetCount())
		return NULL;
	return Table[0]->GetKey();
}

KEY CRemTable::GetLastKey()
{
	int Count = Table.GetCount();
	if (!Count)
		return NULL;
	return Table[Count - 1]->GetKey();
}

KEY CRemTable::GetNextKey(KEY Key)
{
	int Index = GetTableIndex(Key);
	if (Index == -1 ||
		Index == Table.GetCount() - 1)
		return NULL;
	return Table[++Index]->GetKey();
}

KEY CRemTable::GetKeyAfter(KEY Key)
{	int Index;
	short t;

	int Count = Table.GetCount();
	if (!Count)
		return NULL;

	int min = 0;
	int max = Count;
	int ave; // = (max - min) / 2;
	while (min + 1 < max)
	{	ave = (min + max) / 2;
		if (KeyHandlerAlias->Compare(Key, Table[ave]->GetKey()) < 0)
			max = ave;
		else
			min = ave;
	}

	if ((t = KeyHandlerAlias->Compare(Key, Table[min]->GetKey())) <= 0)
	{	if (!t)
		{	if (min == Count - 1)
				return NULL;
			Index = min + 1;
		}
		else
			Index = min;
	}
	else
	{	if (max == Count)
			return NULL;
		Index = max;
	}
	return Table[Index]->GetKey();
}

int CRemTable::IsEmpty()
{
	return Table.GetCount() == 0;
}

/*#if defined(MEMCHECK)
int CRemTable::GetCount()
{
	int Count1 = Table.GetItemsInContainer();
	int Count2 = InvertedTable.GetItemsInContainer();
	Assert(Count1 == Count2);
	return Count1;
}
#endif*/

