#include "myfndt.h"
#include "stcache.h"
#include "storag.h"

const int CStorageCache::DefaultLimit = 10;

CStorageCache::CStorageCache(int Limit,
	CKeyHandler *aKeyHandlerAlias):
	CacheArr(Limit, 0),
	FirstKeyAlias(NULL),
	LastKeyAlias(NULL),
	KeyHandlerAlias(aKeyHandlerAlias)
{
}

CStorageCache::~CStorageCache()
{
}

CAssoc *CStorageCache::Add(CAssoc *NewAssoc)
{
	CObjPtr<CAssoc> OldAssoc = NULL;

	if (CacheArr.IsFull())
	{	Assert(LastKeyAlias != NULL);
		OldAssoc = Yank(LastKeyAlias);
	}

	KEY OldFirst = FirstKeyAlias;
	CCacheMember *Member = new CCacheMember(NewAssoc,
		NULL,
		OldFirst,
		KeyHandlerAlias);

	CacheArr.Add(Member);
	FirstKeyAlias = Member->GetKey();

	if (LastKeyAlias)
	{	int OldFirstIdx = FindIndex(OldFirst);
		Assert(OldFirstIdx != INT_MAX);
		CacheArr[OldFirstIdx]->PrevKeyAlias = FirstKeyAlias;
	}
	else
		LastKeyAlias = FirstKeyAlias;

	return OldAssoc.Drop();
}

int CStorageCache::IsKnown(KEY Key)
{
    return FindIndex(Key) != INT_MAX;
}

CAssoc *CStorageCache::Yank(KEY Key)
{
	int FindIdx = FindIndex(Key);
	if (FindIdx == INT_MAX)
		return NULL;

	CObjPtr<CCacheMember> Member = Detach(Key, FindIdx);
	CAssoc *Assoc = Member->DropAssoc();
	return Assoc;
}

void CStorageCache::StoreAll(CStorage *Storage)
{
	DISArray<CCacheMember>::DIterator Itr = CacheArr.InitIterator();
	while (CacheArr.IsDereferenceable(Itr))
	{	CCacheMember *Current = CacheArr.GetCurrent(Itr);
		Storage->Store(Current->Assoc);
		++Itr;
	}
}

KEY CStorageCache::GetFirstKey()
{
	if (!CacheArr.GetCount())
		return NULL;
	return CacheArr[0]->GetKey();
}

KEY CStorageCache::GetLastKey()
{
	int Count = CacheArr.GetCount();
	if (!Count)
		return NULL;
	return CacheArr[Count - 1]->GetKey();
}

KEY CStorageCache::GetNextKey(KEY Key)
{
	int Index = FindIndex(Key);
	if (Index == INT_MAX ||
		Index == CacheArr.GetCount() - 1)
		return NULL;
	return CacheArr[++Index]->GetKey();
}

KEY CStorageCache::GetKeyAfter(KEY Key)
{	int Index;
	short t;

	int Count = CacheArr.GetCount();
	if (!Count)
		return NULL;

	int min = 0;
	int max = Count;
	int ave; //= (max - min) / 2;
	while (min + 1 < max)
	{	ave = (min + max) / 2;
		if (KeyHandlerAlias->Compare(Key, CacheArr[ave]->GetKey()) < 0)
			max = ave;
		else
			min = ave;
	}

	if ((t = KeyHandlerAlias->Compare(Key, CacheArr[min]->GetKey())) <= 0)
	{	if (!t)
		{	if (min == Count - 1)
				return NULL;
			Index = min + 1;
		}
		else
			Index = min;
	}
	else
	{	if (max == Count)
			return NULL;
		Index = max;
	}
	return CacheArr[Index]->GetKey();
}

int CStorageCache::FindIndex(KEY Key)
{
	CAssoc *TestAssoc = new CAssoc(KeyHandlerAlias,
		Key,
		NULL);
	CObjPtr<CCacheMember> TestMember = new CCacheMember(TestAssoc,
		NULL,
		NULL,
		KeyHandlerAlias);
	return CacheArr.Find(TestMember);
}

CStorageCache::CCacheMember *CStorageCache::Detach(KEY Key, int Idx)
{	CCacheMember *Member;

	// update ends of linked list
	if (!KeyHandlerAlias->Compare(Key, FirstKeyAlias))
		FirstKeyAlias = CacheArr[Idx]->NextKeyAlias;
	if (!KeyHandlerAlias->Compare(Key, LastKeyAlias))
		LastKeyAlias = CacheArr[Idx]->PrevKeyAlias;

	// update linked list
	if ((Member = GetPrev(Idx)) != NULL)
		SkipNext(Member);
	if ((Member = GetNext(Idx)) != NULL)
		SkipPrev(Member);

	Member = CacheArr[Idx];
	CacheArr.DetachByIndex(Idx);
	return Member;
}

CStorageCache::CCacheMember *CStorageCache::GetPrev(int Idx)
{
	CCacheMember *Member = CacheArr[Idx];
	if (!Member->PrevKeyAlias)
		return NULL;

	int PrevIdx = FindIndex(Member->PrevKeyAlias);
	Assert(PrevIdx != INT_MAX);
	return CacheArr[PrevIdx];
}

CStorageCache::CCacheMember *CStorageCache::GetNext(int Idx)
{
	CCacheMember *Member = CacheArr[Idx];
	if (!Member->NextKeyAlias)
		return NULL;

	int NextIdx = FindIndex(Member->NextKeyAlias);
	Assert(NextIdx != INT_MAX);
	return CacheArr[NextIdx];
}

void CStorageCache::SkipPrev(CStorageCache::CCacheMember *Member)
{
	if (!Member->PrevKeyAlias)
		return;

	int PrevIdx = FindIndex(Member->PrevKeyAlias);
	Assert(PrevIdx != INT_MAX);
	Member->PrevKeyAlias = CacheArr[PrevIdx]->PrevKeyAlias;
}

void CStorageCache::SkipNext(CStorageCache::CCacheMember *Member)
{
	if (!Member->NextKeyAlias)
		return;

	int NextIdx = FindIndex(Member->NextKeyAlias);
	Assert(NextIdx != INT_MAX);
	Member->NextKeyAlias = CacheArr[NextIdx]->NextKeyAlias;
}

CStorageCache::CCacheMember::CCacheMember(CAssoc *aAssoc,
	KEY aPrevKeyAlias,
	KEY aNextKeyAlias,
	CKeyHandler *aKeyHandlerAlias):
	Assoc(aAssoc),
	PrevKeyAlias(aPrevKeyAlias),
	NextKeyAlias(aNextKeyAlias),
	KeyHandlerAlias(aKeyHandlerAlias)
{
}

CStorageCache::CCacheMember::~CCacheMember()
{
	delete Assoc;
}

CAssoc *CStorageCache::CCacheMember::DropAssoc()
{
	CAssoc *FreeAssoc = Assoc;
	Assoc = NULL;
	return FreeAssoc;
}


