#if !defined(__stcache_h)
#define __stcache_h

#include "assoc.h"
#include "tl.h"
#include "arrays.h"

class CStorage;

class CStorageCache
{
public:
	static const int DefaultLimit;
private:
	class CCacheMember;

// members
private:
	DISArray<CCacheMember> CacheArr;
	KEY FirstKeyAlias;
	KEY LastKeyAlias;
	CKeyHandler *KeyHandlerAlias;

public:
	CStorageCache(int Limit,
		CKeyHandler *aKeyHandlerAlias);
	virtual ~CStorageCache();

	CAssoc *Add(CAssoc *NewAssoc);
		// can return old association; caller is responsible for
		// storing (and deleting) it

	CAssoc *Yank(KEY Key);
		// returns NULL when no association with Key exists

	void StoreAll(CStorage *Storage);

	KEY GetFirstKey();
	KEY GetLastKey();
	KEY GetNextKey(KEY Key);
	KEY GetKeyAfter(KEY Key);
	int IsKnown(KEY Key);

private:
	int FindIndex(KEY Key);

	CCacheMember *Detach(KEY Key, int Idx);

	CCacheMember *GetPrev(int Idx);
	CCacheMember *GetNext(int Idx);
	void SkipPrev(CCacheMember *Member);
	void SkipNext(CCacheMember *Member);

private:
	class CCacheMember
	{
	public:
		CAssoc *Assoc;
		KEY PrevKeyAlias;
		KEY NextKeyAlias;
		CKeyHandler *KeyHandlerAlias;

		CCacheMember(CCacheMember &Other); // not implemented
	public:
		CCacheMember(CAssoc *aAssoc,
			KEY aPrevKeyAlias,
			KEY aNextKeyAlias,
			CKeyHandler *aKeyHandlerAlias);
		virtual ~CCacheMember();
			// deletes Assoc but no aliases

		KEY GetKey() const { return Assoc->GetKey(); }

		CAssoc *DropAssoc();

#if defined(NOCPCONSTRUCT)
		int operator == (const CCacheMember *Other)
			{ return KeyHandlerAlias->Compare(GetKey(),
					Other->GetKey()) == 0; }
		int operator < (const CCacheMember *Other)
			{ return KeyHandlerAlias->Compare(GetKey(),
					Other->GetKey()) < 0; }
		int operator > (const CCacheMember *Other)
			{ return KeyHandlerAlias->Compare(GetKey(),
					Other->GetKey()) > 0; }
#else
		int operator == (const CCacheMember &Other) const
			{ return KeyHandlerAlias->Compare(GetKey(),
					Other.GetKey()) == 0; }
		int operator < (const CCacheMember &Other) const
			{ return KeyHandlerAlias->Compare(GetKey(),
					Other.GetKey()) < 0; }
#endif
	}; // end CCacheMember declaration
};

#endif

