#if !defined(__storage_h)
#define __storage_h

#include "tl.h"
#include "conduct.h"
#include "remtab.h"
#include "stcache.h"

class CDatum;
class CDatumFactory;

class CStorage
{
friend class CStorageCache;
friend class CRemTable;
	// use Store

private:
	CBook KeyStorage;
	CConductor DataStorage;
	CKeyHandler *KeyHandlerAlias;
	CDatumFactory *Factory;

	CRemTable Remembered;
	CStorageCache Cache;

public:
	CStorage(CKeyHandler *aKeyHandler,
		const DString &aKeyFile, // only path if Opt & IT_TEMP
		CDatumFactory *aFactory,
		const DString &aFreeListFile, // only path if Opt & IT_TEMP
		const DString &aDataFile, // only path if Opt & IT_TEMP
		short Opt,
		int RemLimit = CRemTable::DefaultLimit,
		int RemDelta = CRemTable::DefaultDelta,
		int CacheLimit = CStorageCache::DefaultLimit);
	virtual ~CStorage();

	CDatum *Lock(KEY Key);
		// returns NULL => Key not stored

	int IsKnown(KEY Key);
	CDatum *LockAlways(KEY Key);
		// cannot return NULL - if the key does not exist,
		// new CDatum object is created and returned

	void Unlock(CDatum *Datum);

	KEY NewFirstKey();
		// returns NULL => storage empty

	KEY NewLastKey();
		// returns NULL => storage empty

	KEY NewNextKey(KEY Key);
		// returns NULL => Key is last (or not in storage)

	short Del(KEY Key);
		// 0 OK, 1 key not found

	void StoreAll();

	CConductor &GetConductor() { return DataStorage; }

	unsigned long GetTotal() { return KeyStorage.GetTotal(); }

private:
	CAssoc *GetIntern(KEY Key);
         // returns association in Remembered
	CAssoc *GetInternUnlocked(KEY Key);
	CDatum *LockIntern(CAssoc *Assoc);

	void Store(CAssoc *Assoc);

	KEY GetExtremKey(KEY Key1, KEY Key2, short Min);
		// handles NULL keys

	KEY LockGreaterStoredKey(KEY Key);
	KEY GetGreaterRememberedKey(KEY Key);
	KEY GetGreaterCachedKey(KEY Key);
};

#endif

