#include "myfndt.h"
#include "actfile.h"
#include "dircont.h"
#include "archcmd.h"

#include <stdlib.h>
#if defined(__BORLANDC__)
#include <io.h>
#endif
#if defined(__GNUG__)
#include <unistd.h>
#endif
/*#if !defined(_MSC_VER)
#include <mem.h>
#endif*/
#include <string.h>

CActiveFile::CActiveFile(CArchiverCommand &aArchiverCommand,
	const DString &aStoredName,
	const DString &aContentDir):
	StoredName(aStoredName),
	RefCount(1)
{
	DString CommandLine = aArchiverCommand.Parse(aStoredName,
		aContentDir);
	RunArchiver(CommandLine);
	ContentName = LocateFile(aContentDir, GetMask(aStoredName));
}

CActiveFile::CActiveFile(const DString &aStoredName):
	StoredName(aStoredName),
	ContentName(""),
	RefCount(0)
{
}

CActiveFile::~CActiveFile()
{
	if (!ContentName.IsEmpty())
		unlink(ContentName.c_str());
}

int CActiveFile::DecRefCount()
{
	Assert(RefCount > 0);
	return --RefCount == 0 ? 1 : 0;
}

void CActiveFile::RunArchiver(const DString &CommandLine)
{
	if (system(CommandLine.c_str()))
	{	const char *Comment = "could not execute command: ";
		CMemPtr<char> ErrBuf((short)(strlen(Comment) +
      		CommandLine.GetLength() + 1));
		strcpy((char *)ErrBuf, Comment);
		strcat((char *)ErrBuf, CommandLine.c_str());
		throw CGeneralException(ErrBuf);
	}
}

DString CActiveFile::LocateFile(const DString &aContentDir,
	const DString &Mask)
{
	CDirectory Directory(aContentDir, Mask);

	DString ExName = Directory.ReadNext();
	if (ExName.IsEmpty())
	{	const char *Comment = "decompressor did not create file ";
		CMemPtr<char> ErrBuf((short)(strlen(Comment) +
				aContentDir.GetLength() + Mask.GetLength() + 1));
		strcpy((char *)ErrBuf, Comment);
		strcat((char *)ErrBuf, aContentDir.c_str());
		strcat((char *)ErrBuf, Mask.c_str());
		throw CGeneralException(ErrBuf);
	}

	DString NonexName = Directory.ReadNext();
	if (!NonexName.IsEmpty())
	{	const char *Comment1 = "file ";
		const char *Comment2 = " is ambiguous";
		CMemPtr<char> ErrBuf((short)(strlen(Comment1) +
				aContentDir.GetLength() + Mask.GetLength() +
				strlen(Comment2) + 1));
		strcpy((char *)ErrBuf, Comment1);
		strcat((char *)ErrBuf, aContentDir.c_str());
		strcat((char *)ErrBuf, Mask.c_str());
		strcat((char *)ErrBuf, Comment2);
		throw CGeneralException(ErrBuf);
	}

	return ExName;
}

DString CActiveFile::GetMask(const DString &aStoredName)
{
	const char *StoredNameStr = aStoredName.c_str();
	const char *p = strrchr(StoredNameStr, CDirectory::PathBreak);
	if (p)
		p++;
	else
		p = StoredNameStr;

	int l = strlen(p);
	CMemPtr<char> MaskBuf((short)(l + 3));
	memcpy((char *)MaskBuf, p, l + 1);

	char *q = strrchr((char *)MaskBuf, '.');
	if (!q)
		strcat((char *)MaskBuf, ".*");
	else
		strcpy(q + 1, "*");

	return DString((char *)MaskBuf);
}


