#include "myfndt.h"
#include "archcmd.h"

#include <string.h>

CArchiverCommand::CArchiverCommand(const DString &aCommand,
	const DString &NameMask,
	const DString &DirMask):
	Part0(""),
	Part1(""),
	Part2(""),
	Type(tInvalid)
{
	if (!aCommand.IsEmpty())
	{	const char *p;

		const char *CommandStr = aCommand.c_str();
		const char *NameMaskStr = NameMask.c_str();
		const char *DirMaskStr = DirMask.c_str();

		int NameOffs = -1;
		if (NameMaskStr != NULL &&
			(p = strstr(CommandStr, NameMaskStr)) != NULL)
			NameOffs = static_cast<int>(p - CommandStr);

		int DirOffs = -1;
		if (DirMaskStr != NULL &&
			(p = strstr(CommandStr, DirMaskStr)) != NULL)
			DirOffs = static_cast<int>(p - CommandStr);

		if (NameOffs == -1 &&
			DirOffs == -1)
			SetTypeNoPar(aCommand);
		else if (NameOffs == -1)
			SetSimpleType(DirOffs > 0 ? tStr_Dir : tDir,
				aCommand,
				DirOffs,
				strlen(DirMaskStr));
		else if (DirOffs == -1)
			SetSimpleType(NameOffs > 0 ? tStr_Name : tName,
				aCommand,
				NameOffs,
				strlen(NameMaskStr));
		else
			SetFullType(aCommand, NameOffs, strlen(NameMaskStr),
					DirOffs, strlen(DirMaskStr));
	}
}

DString CArchiverCommand::Parse(const DString &StoredName,
	const DString &ContentDir)
{
	Assert(!StoredName.IsEmpty() && !ContentDir.IsEmpty());
	switch (Type)
	{	case tNoPar:
			return ComposeTypeNoPar();
		case tStr_Name:
			return ComposeSimpleType(StoredName, 1);
		case tName:
			return ComposeSimpleType(StoredName, 0);
		case tStr_Dir:
			return ComposeSimpleType(ContentDir, 1);
		case tDir:
			return ComposeSimpleType(ContentDir, 0);
		case tStr_Name_Dir:
			return ComposeFullTypeSorted(StoredName,
				ContentDir,
				1);
		case tName_Dir:
			return ComposeFullTypeSorted(StoredName,
				ContentDir,
				0);
		case tStr_Dir_Name:
			return ComposeFullTypeSorted(ContentDir,
				StoredName,
				1);
		case tDir_Name:
			return ComposeFullTypeSorted(ContentDir,
				StoredName,
				0);
	}
	Assert(0);
	return DString("");
}

int CArchiverCommand::IsWellFormed()
{
	return Type != tInvalid;
}

void CArchiverCommand::SetTypeNoPar(const DString &aCommand)
{
	Type = tNoPar;
	Part0 = aCommand;
}

void CArchiverCommand::SetSimpleType(EType aType,
	const DString &aCommand,
	int Offs,
	int Len)
{
	Assert(aType == tName || aType == tStr_Name ||
		aType == tDir || aType == tStr_Dir);

	int CommLen = aCommand.GetLength();
	Assert(Offs >= 0 &&
		Offs + Len <= CommLen);
	if (Len == CommLen)
	{	Type = tInvalid;
		return;
	}

	Type = aType;

	CMemPtr<char> CommandCopy((short)(CommLen + 1));
	memcpy((char *)CommandCopy, aCommand.c_str(), CommLen + 1);
	*((char *)CommandCopy + Offs) = 0;

	if (Offs > 0)
	{	Part0 = (char *)CommandCopy;
		if (Offs + Len < CommLen)
			Part1 = (char *)CommandCopy + Offs + Len;
	}
	else
		Part0 = (char *)CommandCopy + Len;
}

void CArchiverCommand::SetFullType(const DString &aCommand,
	int NameOffs,
	int NameLen,
	int DirOffs,
	int DirLen)
{
	if (NameOffs < DirOffs)
		SetFullTypeSorted(NameOffs > 0 ? tStr_Name_Dir : tName_Dir,
			aCommand,
			NameOffs,
			NameLen,
			DirOffs,
			DirLen);
	else
		SetFullTypeSorted(DirOffs > 0 ? tStr_Dir_Name : tDir_Name,
			aCommand,
			DirOffs,
			DirLen,
			NameOffs,
			NameLen);
}

void CArchiverCommand::SetFullTypeSorted(EType aType,
	const DString &aCommand,
	int FirstOffs,
	int FirstLen,
	int SecondOffs,
	int SecondLen)
{
	Assert(aType == tStr_Name_Dir || aType == tName_Dir ||
		aType == tStr_Dir_Name || aType == tDir_Name);

	int CommLen = aCommand.GetLength();
	if (FirstOffs + FirstLen >= SecondOffs ||
		FirstLen + SecondLen >= CommLen)
	{	Type = tInvalid;
		return;
	}

	Type = aType;

	CMemPtr<char> CommandCopy((short)(CommLen + 1));
	memcpy((char *)CommandCopy, aCommand.c_str(), CommLen + 1);
	*((char *)CommandCopy + FirstOffs) = 0;
	*((char *)CommandCopy + SecondOffs) = 0;

	if (FirstOffs > 0)
	{	Part0 = (char *)CommandCopy;
		Part1 = (char *)CommandCopy + FirstOffs + FirstLen;
		if (SecondOffs + SecondLen < CommLen)
			Part2 = (char *)CommandCopy + SecondOffs + SecondLen;
	}
	else
	{	Part0 = (char *)CommandCopy + FirstOffs + FirstLen;
		if (SecondOffs + SecondLen < CommLen)
			Part1 = (char *)CommandCopy + SecondOffs + SecondLen;
	}
}

DString CArchiverCommand::ComposeTypeNoPar()
{
	Assert(!Part0.IsEmpty() && Part1.IsEmpty() &&
			Part2.IsEmpty());
	return Part0;
}

DString CArchiverCommand::ComposeSimpleType(const DString &InsertStr,
	int StartFixed)
{	DString Sum;

	if (StartFixed)
	{	Assert(!Part0.IsEmpty() && Part2.IsEmpty());
		Sum = Part0 + InsertStr;
		if (!Part1.IsEmpty())
			Sum += Part1;
	}
	else
	{	Assert(Part1.IsEmpty() && Part2.IsEmpty());
		Sum = InsertStr;
		if (!Part0.IsEmpty())
			Sum += Part0;
	}
	return Sum;
}

DString CArchiverCommand::ComposeFullTypeSorted(const DString &FirstInsertStr,
	const DString &SecondInsertStr,
	int StartFixed)
{	DString Sum;

	if (StartFixed)
	{	Assert(!Part0.IsEmpty() && !Part1.IsEmpty());
		Sum = Part0 + FirstInsertStr + Part1 + SecondInsertStr;
		if (!Part2.IsEmpty())
			Sum += Part2;
	}
	else
	{	Assert(!Part0.IsEmpty() && Part2.IsEmpty());
		Sum = FirstInsertStr + Part0 + SecondInsertStr;
		if (!Part1.IsEmpty())
			Sum += Part1;
	}
	return Sum;
}

