#include "myfndt.h"
#include "buildp.h"
#include "packer.h"
#include <algo.h>

const int COccBuilder::MarginThreshold = 4;
const int COccBuilder::AQuotient = 2;
const double COccBuilder::BQuotient = 0.5;
	/* profile 13. 9.: introducing margin cut time to process
		1 file, 780 884b from 100 min. to 45 min. while indexes
		_shrank_ from 1 658 049b to 1379 349b */
const int COccBuilder::SettledThreshold = 500;
	/* profile 20. 9.: 1 file, 780 884b:
			1000 ... 40 min
			 100 ... 45 min
			  10 ... 55 min
		with optimizations elsewhere:
			 500 ... 35 min
				0 ... 50 min */

const int COccBuilder::MaxBlockCount = (int)((0xffffL / 2) / sizeof(long));

CPacker COccBuilder::Packer;

COccBuilder::COccBuilder(CStorage *aStorage, 
	long aTicket):
	COccDatum(aStorage, aTicket)
{
}

COccBuilder::~COccBuilder()
{
}

long COccBuilder::Finish()
{
	return GetTicket();
		// always ready
}

/*void COccBuilder::Add(const vector<ulong> &Occs)
{	long NewTicket;

    size_t Size = Occs.size() * sizeof(ulong);
	Assert(Size < 32767);
    CMemPtr<ulong> OccArray(Size);
    copy(Occs.begin(), Occs.end(), (ulong *)OccArray);
	if (GetTicket() == 0)
		NewTicket = AddNew((ulong *)OccArray, Occs.size());
	else
		NewTicket = AddEx((ulong *)OccArray, Occs.size());
	SetTicket(NewTicket);
}*/

void COccBuilder::Add(const vector<ulong> &Occs)
{
    Assert(!Occs.empty());
    CMemPtr<ulong> OccArray(MaxBlockCount * sizeof(long));
    vector<ulong>::const_iterator Block = Occs.begin();
    while (Occs.end() - Block > MaxBlockCount)
    {   copy(Block, Block + MaxBlockCount, (ulong *)OccArray);
        AddBlock((ulong *)OccArray, MaxBlockCount);
        Block += MaxBlockCount;
    }
    if (Block != Occs.end())
    {   copy(Block, Occs.end(), (ulong *)OccArray);
        AddBlock((ulong *)OccArray, Occs.end() - Block);
    }
}

void COccBuilder::AddBlock(ulong *OccArray, int Count)
{   long NewTicket;

	if (GetTicket() == 0)
		NewTicket = AddNew(OccArray, Count);
	else
		NewTicket = AddEx(OccArray, Count);
	SetTicket(NewTicket);
}

long COccBuilder::AddNew(ulong *OccArray, int Count)
{
	if (Count == 1)
		return OccArray[0] | 0x80000000l;

    Packer.Pack(OccArray, Count);

	CConductor &Conductor = GetConductor();
	long Ticket = Conductor.Add(sizeof(long) + Count * sizeof(ulong));
	long Total = Count;
	Conductor.Put(Ticket, 0, &Total, sizeof(long));
	Conductor.Put(Ticket,
		sizeof(long),
		OccArray,
		(short)(Count * sizeof(ulong)));
	return Ticket;
}

long COccBuilder::AddEx(ulong *OccArray, int Count)
{   long Head;

    long OldTicket = GetTicket();
    if (OldTicket & 0x80000000l)
    {   OldTicket &= ~0x80000000l;
	   ulong OldTicketRepre = OldTicket;

	   if (!Packer.MergeOverflows(1, Count))
	   {   int PackedCount;
		  CMemPtr<ulong> PackedArray = Packer.AllocMerge(&OldTicketRepre,
			 1,
			 OccArray,
		      Count,
		      PackedCount);
		  return AddNewTicket((ulong *)PackedArray, PackedCount, NULL, 0);
	   }
	   else // should be *really* rare
	   {   Packer.Pack(OccArray, Count);
	       return AddNewTicket(&OldTicketRepre, 1, OccArray, Count);
	   }
    }

	CConductor &Conductor = GetConductor();
	Conductor.Get(OldTicket, 0, &Head, sizeof(long));
	if (!(Head & 0x80000000l))
	{	if (!SettledThreshold || Head < SettledThreshold)
			return AddOnTip(OldTicket,
				Head,
				OccArray,
				Count);
		else
		{	BuildSecond(OldTicket,
				Head,
				OccArray,
				Count);
			return OldTicket;
		}
	}

	BuildSegmented(OldTicket,
		Head & ~0x80000000l,
		OccArray,
		Count);
	return OldTicket;
}

long COccBuilder::AddNewTicket(ulong *Occ1,
	int Count1,
	ulong *Occ2,
	int Count2)
{    
	long Count = Count1 + Count2;
	Assert(Count > 0);
	long NewLen = (1 + Count) * sizeof(ulong);
	CConductor &Conductor = GetConductor();
	long NewTicket = Conductor.Add(NewLen);
	Conductor.Put(NewTicket, 0, &Count, sizeof(ulong));
	if (Occ1)
		Conductor.Put(NewTicket,
			sizeof(ulong),
			Occ1,
			(short)(Count1 * sizeof(ulong)));
	if (Occ2)
		Conductor.Put(NewTicket,
			(1 + Count1) * sizeof(ulong),
			Occ2,
			(short)(Count2 * sizeof(ulong)));
	return NewTicket;
}

/* profile 8. 2. 97: (pgw)
	pack new only		repack
time:	9:30			9:03
space:	36 169 168		31 605 328 */
long COccBuilder::AddOnTip(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	long Margin;

	int PackedCount = Packer.GetPackedCount(OccArray, Count);

	CConductor &Conductor = GetConductor();
	ulong NewBigCount = OldCount + PackedCount;
	long NewBigSize = (1 + NewBigCount) * sizeof(ulong);

	if (OldCount >= MarginThreshold)
		Margin = static_cast<long>((AQuotient + BQuotient * 
		        OldCount) * sizeof(ulong));
	else
		Margin = 0;
	if (Margin > 0)
		Margin = RoundMargin(NewBigSize, Margin);

	long RealSize = Conductor.SoftGrow(OldTicket, NewBigSize + Margin);
	if (RealSize >= NewBigSize)
	{    // don't touch what's already on disk
	     Packer.Pack(OccArray, Count);
		Assert(PackedCount == Count);

		Conductor.Put(OldTicket, 0, &NewBigCount, sizeof(ulong));
		Conductor.Put(OldTicket,
			(1 + OldCount) * sizeof(ulong),
			OccArray,
			(short)(Count * sizeof(ulong)));

		return OldTicket;
	}
    else
    {   // pack the whole segment
	   if (Packer.AddOverflows(OldCount, Count))
		  return AddBig(OldTicket, OldCount, OccArray, Count);

	   int SegmentSize = OldCount * sizeof(ulong);
	   CMemPtr<ulong> Segment(SegmentSize);
	   Conductor.Get(OldTicket, sizeof(ulong), (ulong *)Segment, SegmentSize);
	   int RepackedCount;
	   CMemPtr<ulong> RepackedSegment = Packer.AllocAdd((ulong *)Segment,
            OldCount,
            OccArray,
   		  Count,
	       RepackedCount);

	   long NewTicket; 
	   long NewSmallSize = (1 + RepackedCount) * sizeof(ulong);
	   if (RealSize >= NewSmallSize)
	       NewTicket = OldTicket;
	   else
	   {   Conductor.Del(OldTicket);
		  NewTicket = Conductor.Add(NewSmallSize + Margin);
	   }

	   ulong NewSmallCount = RepackedCount;
	   Conductor.Put(NewTicket, 0, &NewSmallCount, sizeof(ulong));
	   Conductor.Put(NewTicket,
		  sizeof(ulong),
		  (ulong *)RepackedSegment,
		  (short)(RepackedCount * sizeof(ulong)));

	   return NewTicket;
    }
}

long COccBuilder::AddBig(long OldTicket,
    long OldCount,
    ulong *OccArray,
    int Count)
{
    CConductor &Conductor = GetConductor();

    Packer.Pack(OccArray, Count);
    long NewCount = OldCount + Count;
    // optimization: if old ticket is small enough, it could pack it separately
    // requires new function in CPacket (Pack is only for unpacked arrays)
    long NewTicket = Conductor.Add((1 + NewCount) * sizeof(ulong));

    Conductor.Put(NewTicket, 0, &NewCount, sizeof(ulong));
    Conductor.Copy(NewTicket, sizeof(ulong), OldTicket, sizeof(ulong),
	       OldCount * sizeof(ulong));
    Conductor.Put(NewTicket,
	   (1 + OldCount) * sizeof(ulong),
	   OccArray,
	   (short)(Count * sizeof(ulong)));

    Conductor.Del(OldTicket);
    return NewTicket;
}

void COccBuilder::BuildSecond(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	STail MovedOcc;
	STail NewTail;

	//	we need sizeof(STail) % sizeof(ulong) == 0
	int MovedCount = sizeof(STail) / sizeof(ulong);
	Assert(OldCount >= MovedCount);
	CConductor &Conductor = GetConductor();
	ulong *MovedOccAlias = reinterpret_cast<ulong *>(&MovedOcc);

	Packer.Pack(OccArray, Count);

	long StartSegCount = MakeRoom(OldTicket,
		OldCount,
		MovedOccAlias,
		MovedCount,
		OccArray,
		Count);
	if (!MovedCount && !Count)
		return;

	long TipTicket = AddNewTicket(MovedCount ? MovedOccAlias : 0,
		MovedCount,
		OccArray,
		Count);
	NewTail.InnerFirst = NewTail.InnerLast = NewTail.Tip = TipTicket;
	long NewCount = StartSegCount | 0x80000000l;
	Conductor.Put(OldTicket,
		0,
		&NewCount,
		sizeof(long));
	Conductor.Put(OldTicket,
		(1 + StartSegCount) * sizeof(ulong),
		&NewTail,
		sizeof(STail));
}

void COccBuilder::BuildSegmented(long OldTicket,
	long OldCount,
	ulong *OccArray,
	int Count)
{	STail Tail;
	STail OldTail;
	long TipCount;

	CConductor &Conductor = GetConductor();
	Conductor.Get(OldTicket,
		(1 + OldCount) * sizeof(ulong),
		&Tail,
		sizeof(STail));
	memcpy(&OldTail, &Tail, sizeof(STail));
	Conductor.Get(Tail.Tip, 0, &TipCount, sizeof(long));
	if (!SettledThreshold || TipCount < SettledThreshold)
		BuildExSegment(Tail, TipCount, OccArray, Count);
	else
		BuildNewSegment(Tail, OccArray, Count);
	if (Tail != OldTail)
		Conductor.Put(OldTicket,
			(1 + OldCount) * sizeof(ulong),
			&Tail,
			sizeof(STail));
}

void COccBuilder::BuildExSegment(STail &Tail,
	long TipCount,
	ulong *OccArray,
	int Count)
{
	long NewTip = AddOnTip(Tail.Tip, TipCount, OccArray, Count);
	if (Tail.InnerFirst == Tail.Tip)
		Tail.InnerFirst = NewTip;
	if (Tail.InnerLast == Tail.Tip)
		Tail.InnerLast = NewTip;
	Tail.Tip =NewTip;
}

void COccBuilder::BuildNewSegment(STail &Tail,
	ulong *OccArray,
	int Count)
{
	ulong OldTipMoved = GetOldTipMoved(Tail.Tip);
	long NewTip = AddNewTicket(&OldTipMoved, 1, OccArray, Count);
	SetOldInnerLastNext(Tail.InnerLast,	Tail.Tip);
	Tail.InnerLast = Tail.Tip;
	Tail.Tip = NewTip;
}

ulong COccBuilder::GetOldTipMoved(long OldTipTicket)
{	long OldTipCount;
	ulong Moved;

	CConductor &Conductor = GetConductor();
	Conductor.Get(OldTipTicket, 0, &OldTipCount, sizeof(long));
	Assert(OldTipCount > 1);
	Conductor.Get(OldTipTicket,
		OldTipCount * sizeof(ulong),
			// (1 + OldTipCount - 1) * sizeof(ulong)
		&Moved,
		sizeof(ulong));
	OldTipCount--;
	Conductor.Put(OldTipTicket, 0, &OldTipCount, sizeof(long));
	// now there is free place at the end of OldTipTicket;
	// it will be filled by SetOldInnerLastNext
	return Moved;
}

void COccBuilder::SetOldInnerLastNext(long InnerLastTicket,
		long NextTicket)
{	long InnerLastCount;

	CConductor &Conductor = GetConductor();
	Conductor.Get(InnerLastTicket, 0, &InnerLastCount, sizeof(long));
	Conductor.Put(InnerLastTicket,
		(1 + InnerLastCount) * sizeof(ulong),
		&NextTicket,
		sizeof(long));
}

long COccBuilder::RoundMargin(long Size, long OldMargin)
{
	long Num = 128;
	while (Num > 1)
	{	if (OldMargin > Num)
			return OldMargin + (Size + OldMargin) % Num;
		Num /= 2;
	}
	return OldMargin;
}

long COccBuilder::MakeRoom(long CurTicket,
	long CurCount,
	ulong *DestBuf,
	int &DestBufCount,
	ulong *&SrcPadding,
	int &SrcPaddingCount)
{
     long StartSegCount = CurCount;

	CConductor &Conductor = GetConductor();
	long MaxSize = 1 + CurCount;
	if (SrcPaddingCount > DestBufCount)
		MaxSize += SrcPaddingCount;
	else
		MaxSize += DestBufCount;
	MaxSize *= sizeof(long);
	long RealSize = Conductor.SoftGrow(CurTicket, MaxSize);

	if ((ulong)RealSize >= (1 + CurCount + SrcPaddingCount) *
			sizeof(long))
	{	// put in whole padding
		StartSegCount += SrcPaddingCount;
		Conductor.Put(CurTicket,
			0,
			&StartSegCount,
			sizeof(long));
		Conductor.Put(CurTicket,
			(1 + CurCount) * sizeof(ulong),
			SrcPadding,
			(short)(SrcPaddingCount * sizeof(ulong)));
		DestBufCount = SrcPaddingCount = 0;
		SrcPadding = NULL;
	}
	else if ((ulong)RealSize >= (1 + CurCount + DestBufCount) *
			sizeof(long))
	{	// put in some padding and leave room
		long InCount = (RealSize / sizeof(long)) -
				(1 + CurCount + DestBufCount);
		Assert(InCount < SrcPaddingCount);
		if (InCount > 0)
		{	StartSegCount += InCount;
		     Conductor.Put(CurTicket,
				(1 + CurCount) * sizeof(ulong),
				SrcPadding,
				(short)(InCount * sizeof(ulong)));
			SrcPaddingCount -= static_cast<int>(InCount);
			SrcPadding += static_cast<int>(InCount);
		}
		DestBufCount = 0;
	}
	else
	{	// make room
		long OutCount = (1 + CurCount + DestBufCount) -
				(RealSize / sizeof(long));
		Assert(0 < OutCount && OutCount <= DestBufCount);
		StartSegCount -= OutCount;
		Conductor.Get(CurTicket,
			(1 + StartSegCount) * sizeof(ulong),
			DestBuf,
			(short)(OutCount * sizeof(ulong)));
		DestBufCount = static_cast<int>(OutCount);
	}

	return StartSegCount;
}

