#include "myfndt.h"
#include "decompr.h"
#include "locusopt.h"
#include "filecmp.h"
#include "exedrive.h"

#include "oarchive.h"
#include "odrive.h"

#include <string.h>

int CDecompressor::ActiveFileLimit = 10;
int CDecompressor::ActiveFileDelta = 10;

CDecompressor::CDecompressor(CLocusOptions &Options,
	int aActiveFileLimit,
	int aActiveFileDelta):
	ExeDrive(Options.Get(O_EXEDRIVE)),
	ExeDriveName(Options.Get(O_EXEDRIVENAME)),
	CompressedExt(Options.Get(O_COMPRESSEDEXT)),
	ContentDir(Options.Get(O_CONTENTDIR)),
	ArchiverCommand(Options.Get(O_ARCHIVERCOMMAND),
		Options.Get(O_ARNAMEMASK),
		Options.Get(O_ARDIRMASK)),
	ActiveFiles(aActiveFileLimit, aActiveFileDelta)
{
}

CDecompressor::~CDecompressor()
{
}

DString CDecompressor::GrabContentFile(const DString &StoredFileName)
{
	return Decompress(GetLocalName(ExeDriveName,
			ExeDrive, StoredFileName));
}

DString CDecompressor::Decompress(const DString &StoredFileName)
{
	if (!MatchExt(StoredFileName, CompressedExt))
		return StoredFileName;

	CActiveFile TestFile(StoredFileName);
	int Index = ActiveFiles.Find(&TestFile);
	if (Index != INT_MAX)
	{	CActiveFile *FoundFile = ActiveFiles[Index];
		FoundFile->IncRefCount();
		return FoundFile->GetContentName();
	}
	else
	{	CActiveFile *NewFile = new CActiveFile(ArchiverCommand,
			StoredFileName,
			ContentDir);
		DString ContentName = NewFile->GetContentName();
		ActiveFiles.Add(NewFile);
		return ContentName;
	}
}

void CDecompressor::DropContentFile(const DString &StoredFileName)
{
	if (!MatchExt(StoredFileName, CompressedExt))
		return;

	CActiveFile TestFile(StoredFileName);
	int Index = ActiveFiles.Find(&TestFile);
	Assert(Index != INT_MAX);
	CActiveFile *FoundFile = ActiveFiles[Index];
	if (FoundFile->DecRefCount())
	{	ActiveFiles.Detach(FoundFile);
		delete FoundFile;
	}
}

void CDecompressor::DropAll()
{
	int Count = ActiveFiles.GetCount();
	for (int i = Count - 1; i >= 0; i--)
	{	CActiveFile *ActiveFile = ActiveFiles[i];
		ActiveFiles.Detach(ActiveFile);
		delete ActiveFile;
	}
}

int CDecompressor::MatchExt(const DString &FName, const DString &Ext)
{
	const char *NameExt = strrchr(FName.c_str(), '.');
	if (!NameExt)
		return 0;

	const char *ExtStr = Ext.c_str();
	if (FileStrCmp(NameExt, ExtStr))
		return 0;

	return 1;
}

