#include "myfndt.h"
#include "dircont.h"

//#define DEBUG_find

#if defined(__BORLANDC__)
#include <dir.h>
#include <dos.h>
#else
#if defined(__GNUG__)
#include <sys/stat.h>
#include <fnmatch.h>
#endif
#endif

#if defined(DEBUG_find)
#include <stdio.h>
#endif

#if defined(PLATFORM_DOS) || defined(PLATFORM_WIN32)
const char CDirectory::PathBreak = '\\';
const char *CDirectory::PathBreakStr = "\\";
#else
#if defined(PLATFORM_LINUX)
const char CDirectory::PathBreak = '/';
const char *CDirectory::PathBreakStr = "/";
#else
#error define CDirectory::PathBreak & CDirectory::PathBreakStr
#endif
#endif

CDirectory::CDirectory(const DString &aDirName,
	const DString &aMask):
#if defined(__GNUG__)
	Mask(aMask),
#else
#if defined(__WIN32__)
	FFHandle(INVALID_HANDLE_VALUE),
#endif
#endif
	DirName(aDirName)
{
#if defined(__GNUG__)
	if ((Directory = opendir(aDirName.c_str())) == NULL)
		throw COpenException(aDirName, DString("opendir"));
#else
#if !defined(PLATFORM_WIN32)
	if ((FileFindData = (struct ffblk *)malloc(sizeof(struct
			ffblk))) == NULL)
		throw DMemoryException(sizeof(struct ffblk));
#else
	if ((FileFindData = (WIN32_FIND_DATA *)malloc(sizeof(
			WIN32_FIND_DATA))) == NULL)
#endif
		throw DMemoryException(sizeof(WIN32_FIND_DATA));

	DString Templ = GetPathName(aDirName, aMask);
#if !defined(PLATFORM_WIN32)
	if (findfirst(Templ.c_str(), FileFindData, 0))
#else
	if ((FFHandle = FindFirstFile(Templ.c_str(),
			FileFindData)) == INVALID_HANDLE_VALUE)
#endif
		NextFile = "";
	else
	{
#if !defined(PLATFORM_WIN32)
		if (!(FileFindData->ff_attrib & FA_DIREC))
			NextFile = GetPathName(DirName, FileFindData->ff_name);
#else
		if (!(FileFindData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			NextFile = GetPathName(DirName, FileFindData->cFileName);
#endif
		else
			NextFile = GetNextFile();
	}
#endif
}

CDirectory::~CDirectory()
{
#if defined(__GNUG__)
	closedir(Directory);
#else
#if defined(__WIN32__)
	if (FFHandle != INVALID_HANDLE_VALUE)
		FindClose(FFHandle);
#endif
	free(FileFindData);
#endif
}

DString CDirectory::ReadNext()
{
#if defined(__GNUG__)
	return GetNextFile();
#else
	if (NextFile.IsEmpty())
		return NextFile;

	DString Ret = NextFile;
	NextFile = GetNextFile();
	return Ret;
#endif
}

DString CDirectory::GetNextFile()
{
#if defined(__GNUG__)
	struct dirent *Entry;

	while ((Entry = readdir(Directory)) != NULL)
	{    if (IsMatch(Entry->d_name))
		    return GetPathName(DirName, Entry->d_name);
	}
#else
#if !defined(PLATFORM_WIN32)
	while (!findnext(FileFindData))
#else
	while (FindNextFile(FFHandle, FileFindData))
#endif
	{
#if !defined(PLATFORM_WIN32)
		if (!(FileFindData->ff_attrib & FA_DIREC))
			return GetPathName(DirName, FileFindData->ff_name);
#else
		if (!(FileFindData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
#endif
			return GetPathName(DirName, FileFindData->cFileName);
	}
#endif
	return DString();
}

#if defined(__GNUG__)
int CDirectory::IsMatch(const char *Name)
{	struct stat Buf;

	DString FName = GetPathName(DirName, Name);

	if (stat(FName.c_str(), &Buf))
		return 0;

	if (!S_ISREG(Buf.st_mode))
		return 0;

#if !defined(__DJGPP__)
	if (fnmatch(Mask.c_str(), FName.c_str(), FNM_PATHNAME | FNM_PERIOD))
#else
	if (fnmatch(Mask.c_str(), Name, FNM_PATHNAME | FNM_PERIOD))
		return 0;
#endif

	return 1;
}
#endif

CDirectory::EType CDirectory::NextFileOrDir(DString &Name)
{
#if defined(__GNUG__)
    return GetNextFile2(Name);
#else
    if (NextFile.IsEmpty())
	   return nomore;    

    Name = NextFile;
    EType Ret = GetNextFile2(NextFile);
    if (Ret == nomore) 
	   NextFile = "";
    return Ret;
#endif
}

CDirectory::EType CDirectory::GetNextFile2(DString &Name)
{
#if defined(__GNUG__)
	struct dirent *Entry;

	while ((Entry = readdir(Directory)) != NULL)
	{
#if defined(DEBUG_find)
                printf("testing %s\n", Entry->d_name);
#endif
                EType Ret = IsMatch2(Entry->d_name);
		if (Ret != nomore)
		{       Name = GetPathName(DirName, Entry->d_name);
#if defined(DEBUG_find)
                        printf("found %s\n", Name.c_str());
#endif
                        return Ret;
		}
	}
#else
#if !defined(PLATFORM_WIN32)
	while (!findnext(FileFindData))
#else
	while (FindNextFile(FFHandle, FileFindData))
#endif
	{
#if !defined(PLATFORM_WIN32)
		if (!strcmp(FileFindData->ff_name, ".") ||
			!strcmp(FileFindData->ff_name, ".."))
			return nomore;
		Name = GetPathName(DirName, FileFindData->ff_name);
	    return (FileFindData->ff_attrib & FA_DIREC) ? dir : file;
#else
		if (!strcmp(FileFindData->cFileName, ".") ||
			!strcmp(FileFindData->cFileName, ".."))
			return nomore;
		Name = GetPathName(DirName, FileFindData->cFileName);
		return (FileFindData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? 
	             dir : file;
#endif
	}
#endif
    return nomore;
}

#if defined(__GNUG__)
CDirectory::EType CDirectory::IsMatch2(const char *Name)
{   struct stat Buf;
    static const char *MatchAll = 
#if defined(PLATFORM_LINUX)
	"*";
#else
	"*.*";
#endif
 
    if (!strcmp(Name, ".") || !strcmp(Name, ".."))
	return nomore;

    DString FName = GetPathName(DirName, Name);

    if (stat(FName.c_str(), &Buf))
    {
#if defined(DEBUG_find)
           printf("stat failed on %s\n", FName.c_str());
#endif
	   return nomore;
    }

    if (S_ISDIR(Buf.st_mode))
    {
#if defined(DEBUG_find)
           printf("%s is a directory\n", FName.c_str());
#endif
	   return dir;
    }

    if (!S_ISREG(Buf.st_mode))
    {
#if defined(DEBUG_find)
           printf("%s is not a regular file\n", FName.c_str());
#endif
	   return nomore;
    }

    // kludge - fnmatch does not necessarily work
    if (!strcmp(Mask.c_str(), MatchAll))
	return file;

#if defined(DEBUG_find)
    printf("testing %s against %s\n", FName.c_str(), Mask.c_str());
#endif

    DString PathMask = GetPathName(DirName, Mask.c_str());
//#if !defined(__DJGPP__)
    if (fnmatch(PathMask.c_str(), FName.c_str(), FNM_PATHNAME | FNM_PERIOD))
/*#else
    if (fnmatch(Mask.c_str(), Name, FNM_PATHNAME | FNM_PERIOD))
#endif*/
	return nomore;

    return file;
}
#endif

DString CDirectory::GetPathName(const DString &Path, const char *Name)
{
    Assert(!Path.IsEmpty());
    DString FName = Path;
    if (FName[FName.GetLength() - 1] != PathBreak)
	   FName += PathBreak;
    FName += Name;
    return FName;
}
