#include "myfndt.h"
#include "dualmap.h"

CDualMap::CDualMap(ushort Opt,
		ushort AveStrLen,
		const DString &NumTreeName,
		const DString &StrTreeName,
		const DString &ConductListName,
		const DString &ConductDataName):
	CBook(new CCompressedStrHandler(AveStrLen), StrTreeName, Opt),
	NumToStr(new CULongHandler(), NumTreeName, Opt),
	Conductor(ConductListName, ConductDataName, Opt)
{
}

CDualMap::~CDualMap()
{
}

short CDualMap::Add(ulong Num, const DString &Str)
{	LONG Data;
	short KnownRet;

	// check duplicities
	short RetVal = 0;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (!NumToStr.GetData(NumKey, Data))
		RetVal |= 1;

	CNewKey StrKey(CCompressedStrHandler::NewKey(Str.c_str()));
	if (!CBook::GetData(StrKey, Data))
		RetVal |= 2;

	if (RetVal)
		return RetVal;

	// add
	short Len = (short)(Str.GetLength() + 1);
	long Ticket = Conductor.Add(Len);
	Conductor.Put(Ticket, 0, Str.c_str(), Len);
	KnownRet = NumToStr.Add(NumKey, Ticket);
	Assert(KnownRet == 0);
	KnownRet = CBook::Add(StrKey, Num);
	Assert(KnownRet == 0);
	return 0;
}

short CDualMap::Del(ulong Num)
{	long Ticket;
	short KnownRet;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToStr.GetData(NumKey, Ticket))
		 return 1;

	CMemPtr<char> Data = reinterpret_cast<char *>(
			Conductor.Alloc(Ticket));
	CNewKey StrKey(CCompressedStrHandler::NewKey((char *)Data));
	KnownRet = CBook::Del(StrKey);
	Assert(KnownRet == 0);
	Conductor.Del(Ticket);

	KnownRet = NumToStr.Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CDualMap::Del(const DString &Str)
{	LONG Data;
	long Ticket;
	short KnownRet;

	CNewKey StrKey(CCompressedStrHandler::NewKey(Str.c_str()));
	if (CBook::GetData(StrKey, Data))
		return 1;

	ulong Num = Data;
	KnownRet = CBook::Del(StrKey);
	Assert(!KnownRet);

	CNewKey NumKey(CULongHandler::NewKey(Num));
	KnownRet = NumToStr.GetData(NumKey, Ticket);
	Assert(KnownRet == 0);
	Conductor.Del(Ticket);

	KnownRet = NumToStr.Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CDualMap::Get(ulong Num, DString &Str)
{	long Ticket;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToStr.GetData(NumKey, Ticket))
		return 1;

	CMemPtr<char> NewData = reinterpret_cast<char *>(
			Conductor.Alloc(Ticket));
	Str = (char *)NewData;
	return 0;
}

short CDualMap::Get(const DString &Str, ulong &Num)
{	long Data;

	CNewKey StrKey(CCompressedStrHandler::NewKey(Str.c_str()));
	if (CBook::GetData(StrKey, Data))
		return 1;

	Num = Data;
	return 0;
}

int CDualMap::Has(const DString &Str)
{	LONG DummyData;

	CNewKey StrKey(CCompressedStrHandler::NewKey(Str.c_str()));
	return CBook::GetData(StrKey, DummyData) ? 0 : 1;
}

short CDualMap::GetMax(ulong &Num)
{
	CLockedKey LastKey(NumToStr.LockLastKey(), &NumToStr);
	if (!LastKey)
		return 1;
	Num = CULongHandler::GetULong(LastKey);
	return 0;
}

