#include "myfndt.h"
#include "filebase.h"
#include "fbdual.h"
#include "oflbase.h"
#include "obase.h"

//#define DEBUG_add

CFileBase::CFileBase(ushort Opt,
	CLocusOptions &Options)
{
	BaseIndex = new CFBDual(Opt, Options);
	BaseData = new CRecBase(Options.Get(O_FILEBASEEXT),
		Opt,
		sizeof(SFileRecord));
	FirstWordTree = new CBook(new CULongHandler,
		Options.Get(O_FIRSTWORDTREE),
		Opt);
}

CFileBase::~CFileBase()
{
	delete FirstWordTree;
	delete BaseData;
	delete BaseIndex;
}

ulong CFileBase::GetGlobalDelta()
{  CDocDesc DocDesc;
	SFileRecord Rec;
	ulong Num;

	if (!GetLast(DocDesc, Rec, Num))
		return Rec.FirstWord + Rec.WordCount;
	else
		return 0;
}

short CFileBase::Add(const CDocDesc &DocDesc,
	const SFileRecord &Rec,
	ulong &Num)
{   short KnownRet;

#if defined(DEBUG_add)
    printf("CFileBase::Add\n");
#endif

	if (!BaseIndex->Get(DocDesc, Num))
		return 1;
    Num = BaseData->Add();

#if defined(DEBUG_add)
    printf("CFileBase::Add: %s from %d (%d words)\n",
        DocDesc.GetFileName().c_str(),
        DocDesc.GetStartOffs(),
        Rec.WordCount);
#endif
    KnownRet = BaseIndex->Add(Num, DocDesc);
	Assert(!KnownRet);

	BaseData->Put(Num, &Rec);
	CNewKey Key(CULongHandler::NewKey(Rec.FirstWord));
	KnownRet = FirstWordTree->Add(Key, Num);
	Assert(!KnownRet);

	return 0;
}

short CFileBase::Get(const CDocDesc &DocDesc,
	SFileRecord &Rec,
	ulong &Num)
{
	if (BaseIndex->Get(DocDesc, Num))
		return 1;
	CMemPtr<void> RecPtr = BaseData->Alloc(Num);
	memcpy(&Rec, (void *)RecPtr, sizeof(SFileRecord));
	return 0;
}

short CFileBase::Get(ulong Num,
	SFileRecord &Rec,
	CDocDesc &DocDesc)
{
	if (BaseIndex->Get(Num, DocDesc))
		return 1;
	CMemPtr<void> RecPtr = BaseData->Alloc(Num);
	memcpy(&Rec, (void *)RecPtr, sizeof(SFileRecord));
	return 0;
}

short CFileBase::GetFirst(CDocDesc &DocDesc,
	SFileRecord &Rec,
	ulong &Num)
{
	if (BaseData->GetFirst(Num))
		return 1;
	short KnownRet = Get(Num, Rec, DocDesc);
	Assert(!KnownRet);
	return 0;
}

short CFileBase::GetLast(CDocDesc &DocDesc,
	SFileRecord &Rec,
	ulong &Num)
{
	if (BaseData->GetLast(Num))
		return 1;
	short KnownRet = Get(Num, Rec, DocDesc);
	Assert(!KnownRet);
	return 0;
}

// supposes there are no holes in CRecBase
ulong CFileBase::GetTotalDocCount()
{    ulong Num;

	if (BaseData->GetLast(Num))
		return 0;
	return Num + 1;
}

void CFileBase::StoreAll()
{
	BaseIndex->StoreAll();
	BaseData->StoreAll();
	FirstWordTree->StoreAll();
}

short CFileBase::GetNum(ulong Occ, ulong &Num)
{	LONG Data;

	CNewKey TestKey(CULongHandler::NewKey(Occ));
	CLockedKey FoundKey(FirstWordTree->LockKeyBefore(TestKey),
			FirstWordTree);
	if (!FoundKey)
		return 1;
	short KnownRet = FirstWordTree->GetData(FoundKey, Data);
	Assert(!KnownRet);
	Num = Data;
	return 0;
}

short CFileBase::Has(const DString &DocName)
{
	return BaseIndex->Has(DocName);
}

