#include "myfndt.h"
#include "fixedocc.h"

/*#if !defined(_MSC_VER)
#include <alloc.h>
#endif*/
#include <stdlib.h>

const int CFixedOcc::OutLimit = 300;
const long CFixedOcc::CompressCacheOnlyIterator = -1;

CFixedOcc::COccList::COccList(CConductor &Conductor,
	long Ticket,
	long aDelta,
	int Count):
	OccList(NULL),
	ListCount(0),
	Delta(aDelta)
{	long TotalCount;

	Assert(Ticket != 0);
	Conductor.Get(Ticket, 0, &TotalCount, sizeof(long));
	TotalCount &= ~0x80000000l;
	Assert(TotalCount > 0 &&
		Delta >= 0 &&
		Delta <= TotalCount);
	if (Delta == TotalCount)
		return;

	if (TotalCount - Delta < Count)
		ListCount = static_cast<int>(TotalCount - Delta);
	else
		ListCount = Count;

	OccList = (ulong *)malloc(ListCount * sizeof(ulong));
	if (!OccList)
		throw DMemoryException(ListCount * sizeof(ulong));
	Conductor.Get(Ticket,
		sizeof(long) + Delta * sizeof(ulong),
		OccList,
		ListCount * sizeof(ulong));
}

CFixedOcc::COccList::~COccList()
{
	if (OccList)
		free(OccList);
}

CFixedOcc::CFixedOcc(CStorage *aStorage, long aTicket):
	COccDatum(aStorage, aTicket),
	OutOcc(NULL),
	CurrentTicket(GetTicket()),
	InternalTail(NULL),
	TailUnknown(1),
	ItrStatus(get_next),
	CurBase(0xffffffffl),
	LastCompressedOcc(0xffffffffl)
{
}

CFixedOcc::~CFixedOcc()
{
	delete OutOcc;
	delete InternalTail;
}

long CFixedOcc::Finish()
{
	return 0; // never changes
}

long CFixedOcc::InitIterator()
{
     ItrStatus = get_next;

	CurrentTicket = GetTicket();
     return InitSegmentIterator();
}

ulong CFixedOcc::GetNext(long &Iterator)
{    
     if (Iterator == CompressCacheOnlyIterator)
	{    Assert(ItrStatus == use_second);
		Iterator = 0;
		return ExtractSecond();
	}

     if (ItrStatus == get_next)
     {    LastCompressedOcc = GetNextCompressed(Iterator);
	   	if (!(LastCompressedOcc & 0x80000000l))
	    	     CurBase = LastCompressedOcc;
	    	else
	    	{    CurBase = ExtractFirst();
	    		ItrStatus = use_second;
			if (!Iterator)
			     Iterator = CompressCacheOnlyIterator;
	    	}
	}
	else
	{    CurBase = ExtractSecond();
		ItrStatus = get_next;
	}
	return CurBase;
}

ulong CFixedOcc::GetNextCompressed(long &Iterator)
{
	if (CurrentTicket & 0x80000000l)
	{	Assert(Iterator == 1 &&
			CurrentTicket == GetTicket());
		Iterator = 0;
		return CurrentTicket & ~0x80000000l;
	}

	Assert(OutOcc != NULL &&
		Iterator > OutOcc->Delta);
	int Index = static_cast<int>(Iterator - OutOcc->Delta - 1);
	Assert(Index < OutOcc->ListCount);
	ulong Occ = OutOcc->OccList[Index];
	Iterate(Index, Iterator);
	return Occ;
}

#if defined(DEBUG_CFixedOcc)
void CFixedOcc::Check()
{
	long Itr = InitIterator();
	Assert(Itr != 0);
	ulong Occ1 = GetNext(Itr);
	long CountedCount = 1;
	while (Itr != 0)
	{	ulong Occ2 = GetNext(Itr);
		Assert(Occ2 >= Occ1);
		Occ1 = Occ2;
		CountedCount++;
	}
}
#endif

/*long CFixedOcc::GetTotalCount()
{	long TotalCount;

	STail *Tail = GetTail();
	if (!Tail)
	{	long Ticket = GetTicket();
		if (Ticket == 0)
			TotalCount = 0;
		else
		{	if (Ticket & 0x80000000l)
				TotalCount = 1;
			else
			{	CConductor &Conductor = GetConductor();
				Conductor.Get(Ticket, 0, &TotalCount, sizeof(long));
				Assert(!(TotalCount & 0x80000000l));
			}
		}
	}
	else
		TotalCount = Tail->TotalCount;
	return TotalCount;
}*/

void CFixedOcc::Iterate(int Index, long &Iterator)
{
	IterateInSegment(Index, Iterator);

	if (Iterator != 0)
		return;
	STail *Tail = (STail *)GetTail();
	if (!Tail)
		return;
	if (CurrentTicket == Tail->Tip)
		return;

	IterateTicket();
	Iterator = InitSegmentIterator();
}

void CFixedOcc::IterateInSegment(int Index, long &Iterator)
{
	if (Index < OutOcc->ListCount - 1)
		Iterator++;
	else
	{	long NewDelta = OutOcc->Delta + OutOcc->ListCount;
		delete OutOcc;
		OutOcc = NULL; // necessary (for exceptions)
		OutOcc = new COccList(GetConductor(),
			CurrentTicket,
			NewDelta,
			OutLimit);
		if (OutOcc->ListCount == 0)
		{	delete OutOcc;
			OutOcc = NULL;
			Iterator = 0;
		}
		else
			Iterator = OutOcc->Delta + 1;
	}
}

void CFixedOcc::IterateTicket()
{	long Count;
	long Next;

	CConductor &Conductor = GetConductor();
	STail *Tail = (STail *)GetTail();
	Assert(Tail != NULL);
	Conductor.Get(CurrentTicket, 0, &Count, sizeof(long));
	if (Count & 0x80000000l)
		CurrentTicket = Tail->InnerFirst;
	else
	{  if (CurrentTicket == Tail->InnerLast)
			CurrentTicket = Tail->Tip;
		else
		{	Conductor.Get(CurrentTicket,
				(1 + Count) * sizeof(ulong),
				&Next,
				sizeof(long));
			CurrentTicket = Next;
      }
	}
}

long CFixedOcc::InitSegmentIterator()
{
	if (CurrentTicket == 0)
		return 0;

	delete OutOcc;
	OutOcc = NULL;
	if (!(CurrentTicket & 0x80000000l))
		OutOcc = new COccList(GetConductor(),
			CurrentTicket,
			0,
			OutLimit);
	return 1;
}

void *CFixedOcc::GetTail()
{	long Count;

	if (!TailUnknown)
		return InternalTail;
	TailUnknown = 0;

	long Ticket = GetTicket();
	if (Ticket == 0 || Ticket & 0x80000000l)
		return NULL;

	CConductor &Conductor = GetConductor();
	Conductor.Get(Ticket, 0, &Count, sizeof(long));
	if (!(Count & 0x80000000l))
		return NULL;

	Assert(!InternalTail);
	InternalTail = new STail;
	Count &= ~0x80000000l;
	Conductor.Get(Ticket,
		(1 + Count) * sizeof(ulong),
		InternalTail,
		sizeof(STail));
	return InternalTail;
}

