#include "myfndt.h"
#include <syslog.h>
#include "dircont.h"
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

extern void Run(int Count, char *Argv[]);

void LogEntry(int Count, const char * const Argv[]);
void LogError(const char *Message);

int main(int argc, char *argv[])
{   char ErrorBuffer[512];

	try
	{	LogEntry(argc, argv);
		Assert1(1, __FILE__, __LINE__);
			// generate assert fnc even in liblocus.so
			// without asserts - clients may need it
		Run(argc, argv);
		syslog(LOG_DEBUG, "exit success");
		return 0;
	}
	catch (DAssertException &x)
	{	DString File = x.GetFile();
        sprintf(ErrorBuffer, 
            "assertion failed in file %s on line %d; aborting...\n",
            File.c_str(),
            x.GetLine());
        LogError(ErrorBuffer);
		return -1;
	}
	catch (DMemoryException &x)
	{	sprintf(ErrorBuffer,
            "allocation of %d bytes failed; aborting...\n",
			x.requested());
        LogError(ErrorBuffer);
		return -1;
	}
	catch (COpenException &x)
	{	sprintf(ErrorBuffer,
            "could not open file (directory) %s with mode %s; "
			   "aborting...\n",
            x.GetFile(),
            x.GetMode());
        LogError(ErrorBuffer);
		return -1;
	}
	catch (CReadException &x)
	{	LogError("could not read from file; aborting...\n");
		return -1;
	}
	catch (CWriteException &x)
	{	LogError("could not write to file; aborting...\n");
		return -1;
	}
	catch (DException &x)
		// under gcc, this will _not_ catch COpenException nor any
		// other descendants
	{	DString Why = 
#if !defined(_MSC_VER)
			x.why();
#else
			x.what();
#endif
	    sprintf(ErrorBuffer, "exception: %s; aborting...\n", Why.c_str());
        LogError(ErrorBuffer);
		return -1;
	}
	catch (...)
	{	LogError("unknown exception! aborting...\n");
		return -1;
	}
}

void LogEntry(int Count, const char * const Argv[])
{
    Assert(Count > 0);
    const char *p = Argv[0];
    const char *q = strrchr(p, CDirectory::PathBreak);
    if (q)
        ++q;
    else
        q = p;
    openlog(q, LOG_PID | LOG_CONS, LOG_USER);

    DString With = "";
    if (Count > 1)
    {   for (int i = 1; i < Count; ++i)
        {   With += " \"";
            With += Argv[i];
            With += "\"";
        }
    }
    else
        With = " no parameters";

    char dir[256];
    if (!getcwd(dir, sizeof(dir)))
        strcpy(dir, "???");

    syslog(LOG_DEBUG, "started in %s with%s\n", dir, With.c_str());
}

void LogError(const char *Message)
{
    fprintf(stderr, "%s", Message);
    syslog(LOG_ERR, "%s", Message);
}
