#include "myfndt.h"
#include "locusopt.h"
#include "archcmd.h"
#include "dircont.h"
#include "exedrive.h"

#include "obase.h"
#include "oindex.h"
#include "oflbase.h"
#include "opunct.h"
#include "oarchive.h"
#include "odocfile.h"
#include "ocase.h"
#include "odrive.h"
#include "opresent.h"
#include "oarchw.h"
#include "ostem.h"
#include "oreader.h"
#include "ouno.h"
#include "osubdoc.h"
#include "oinfra.h"
#include "odlerr.h"
#include "ooutform.h"

static DString GetPunctuation(COptions<DString> *Options, const char *Key);
static DString GetTamedDirectory(COptions<DString> *Options, const char *Key);
static DString GetStemFileName(COptions<DString> *Options, const char *Key);
static DString GetSeparators(COptions<DString> *Options, const char *);
static DString GetUnoFile(COptions<DString> *Options, const char *);

CLocusOptions::CLocusOptions(const DString &ExeDrive)
{	DString Text;
	char Kludge[2];

	Add(O_BASEDIR, DString(""));

	Add(O_FILEBASEEXT, GetFileName);
	Add(O_BASECONLISTEXT, GetFileName);
	Add(O_BASECONDATAEXT, GetFileName);
	Add(O_BASENUMTREE, GetFileName);
	Add(O_BASESTRTREE, GetFileName);
	Add(O_CLUSTERFILEEXT, GetFileName);
	Add(O_WORDTREEEXT, GetFileName);
	Add(O_OCCSTOREFREELISTEXT, GetFileName);
	Add(O_OCCSTOREDATAEXT, GetFileName);
	Add(O_FIRSTWORDTREE, GetFileName);

	Add(O_DIVIDER, DString(""));
    Add(O_READERMARGIN, 20); // how big should be?

	Text = "\r\t.,;:\'\"()!?";
	Kludge[0] = 0x1a;
	Kludge[1] = 0;
	Text += Kludge;
	Add(O_BASICPUNCT, Text);
	Text = "[]{}@#$%^&*-_=|\\/";
	Add(O_ADDITIONALPUNCT, Text);
	Add(O_PUNCT, GetPunctuation);

	Add(O_COMPRESSEDEXT, DString(""));
	Add(O_CONTENTDIRWILD, DString(""));
	Add(O_CONTENTDIR, GetTamedDirectory);
	Add(O_ARCHIVERCOMMAND, DString(""));
	Add(O_ARNAMEMASK, DString("%src"));
	Add(O_ARDIRMASK, DString("%dest"));

	Add(O_DFNUMTREE, GetFileName);
	Add(O_DFSTRTREE, GetFileName);
	Add(O_DFCONDLIST, GetFileName);
	Add(O_DFCONDDATA, GetFileName);

	Add(O_AVEWORDLEN, 2);
	/* should be used only by cruncher and read by grazer from
			tree on disk
		profile 17. 10.: 184735 bytes in 9 files
		AverageWordLen		word tree size
		8						185724
		6						175772
		5						168860
		4						173292
		19. 10. (with compression): 97392 bytes in 10 files
		AverageWordLen		word tree size
		5 no compress		105420
		5						90780
		4						84828
		3						82268
		2						78412
		1						79212 */
	Add(O_AVEFILELEN, 10); //30);

	Add(O_EARLYCASECONV, 2);

	Add(O_EXEDRIVENAME, "");
	Add(O_EXEDRIVE, ExeDrive);

	Add(O_CONFNAME, "");

	Add(O_STEMDIR, "");
	Add(O_SDKEY, GetStemFileName);
	Add(O_SDSUPP, GetStemFileName);
	Add(O_SDDATA, GetStemFileName);
	Add(O_SDTABLE, GetStemFileName);

	Add(O_BASICSEPAR, " \n\r\t");
	Add(O_ADDSEPAR, "");
	Add(O_SEPARATORS, GetSeparators);

	Add(O_WILDUNOFILE, "");
	Add(O_UNOFILE, GetUnoFile);
	Add(T_PARSEUNOERRFORMAT_SDSN, "error parsing named objects in file %s "
             "on line %d: %s\n");

    Add(O_DOCPARTS, CStringListValue(4, 8));
    Add(O_PUSHPART, "");
    Add(O_POPPART, "");
    Add(O_PUSHCASESENSITIVE, 1);
    Add(O_POPCASESENSITIVE, 1);

    Add(O_CRUFT, "");
    Add(O_CRUFTCASE, 1);
    Add(O_SUBTEMPL_1, "%0");

    Add(O_INFRAKEY, GetFileName);
    Add(O_INFRAFREE, GetFileName);
    Add(O_INFRADAT, GetFileName);

    Add(T_DLOPENERR_SS, "dlopen(\"%s\") failed (error: %s)");
    Add(T_DLSYMERR_SS, "dlsym(\"%s\") failed (error: %s)");

    Add(O_HIDDENFILEHEAD, "");

	WatchSection("locus");
}

CLocusOptions::~CLocusOptions()
{
}

DString CLocusOptions::FirstInvalidKey()
{
	DString Key = CUnivOptions::FirstInvalidKey();
	if (!Key.IsEmpty())
		return Key;

	DString BaseDir = Get(O_BASEDIR);
	if (BaseDir.IsEmpty())
		return DString(O_BASEDIR);

	Key = FirstArchiverInvalidKey();
	if (!Key.IsEmpty())
		return Key;

	if (!IsDocPartsValid())
		return DString(O_DOCPARTS);

#if defined(STATIC_LIBS_ONLY)
    DString StemDir = Get(O_STEMDIR);
    if (!StemDir.IsEmpty())
        return DString(O_STEMDIR);
#endif

	return DString("");
}

bool CLocusOptions::IsDocPartsValid()
{
    // "plain" is a reserved word
    CStringListValue DocParts = COptions<CStringListValue>::Get(O_DOCPARTS);
    if (DocParts.Find(DString(PLAINTEXT_SUBDOC_NAME)) != INT_MAX)
        return false;

    // doc parts must be valid names as defined in lib/query/quexpr.l
    CStringListValue::DIterator Itr = DocParts.InitIterator();
    while (DocParts.IsDereferenceable(Itr))
    {   if (!IsValidDocPart(*Itr))
            return false;
        Itr++;
    }

    return true;
}

bool CLocusOptions::IsValidDocPart(const DString &Name)
{
    if (Name.IsEmpty())
        return false;

    char f = Name[0];
    if (!((f >= 'a' && f <= 'z') || (f >= 'A' && f <= 'Z')))
        return false;

    int l = Name.GetLength();
    for (int i = 1; i < l; ++i)
    {   char c = Name[i];
        if (!((c >= 'a' && c <= 'z') || 
            (c >= 'A' && c <= 'Z') ||
            (c >= '0' && c <= '9') ||
            (c == '_')))
            return false;
    }

    return true;
}

DString CLocusOptions::FirstArchiverInvalidKey()
{
	DString CompressedExt = Get(O_COMPRESSEDEXT);
	DString ContentDir = Get(O_CONTENTDIR);
	DString ArchiverCommand = Get(O_ARCHIVERCOMMAND);
	if (CompressedExt.IsEmpty() &&
		ContentDir.IsEmpty() &&
		ArchiverCommand.IsEmpty())
		return DString("");

	if (CompressedExt.IsEmpty())
		return CompressedExt;
	if (ContentDir.IsEmpty())
		return ContentDir;
	if (ArchiverCommand.IsEmpty())
		return ArchiverCommand;

	CArchiverCommand TestCommand(ArchiverCommand,
		Get(O_ARNAMEMASK),
		Get(O_ARDIRMASK));
	if (!TestCommand.IsWellFormed())
		return ArchiverCommand;

	return DString("");
}

DString GetFileName(COptions<DString> *Options, const char *Key)
{       static DString FileName = "";

        if (FileName.IsEmpty())
        {       FileName = GetDirectoryFromKey(Options, O_BASEDIR);
           	Assert(!FileName.IsEmpty());
        }
	return FileName + Key;
}

static DString GetPunctuation(COptions<DString> *Options, const char *)
{
	return Options->Get(O_BASICPUNCT) +
		Options->Get(O_ADDITIONALPUNCT);
}

static DString GetSeparators(COptions<DString> *Options, const char *)
{
	return Options->Get(O_BASICSEPAR) + Options->Get(O_ADDSEPAR);
}

static DString GetTamedDirectory(COptions<DString> *Options, const char *Key)
{
        return GetDirectoryFromKey(Options, Key + 1);
}

DString GetDirectoryFromKey(COptions<DString> *Options, const char *Key)
{
        DString Name = GetLocalName(Options->Get(O_EXEDRIVENAME),
                Options->Get(O_EXEDRIVE),
                Options->Get(Key));
        int Len = Name.GetLength();
        if (!Len)
                return DString("");

   	if (Name[Len - 1] != CDirectory::PathBreak)
		Name += CDirectory::PathBreakStr;
        return Name;
}

static DString GetStemFileName(COptions<DString> *Options, const char *Key)
{
    DString Name = GetDirectoryFromKey(Options, O_STEMDIR);
    Assert(!Name.IsEmpty());
    return Name + Key;
}

static DString GetUnoFile(COptions<DString> *Options,
    const char *)
{
#if defined(PLATFORM_WIN32)
#error do not use file name without path
#endif
    DString UnoFile = Options->Get(O_WILDUNOFILE);
    if (UnoFile.IsEmpty())
	   return DString("locus.opt");
    else
	   return UnoFile;
}
