#include "myfndt.h"
#include "occdatum.h"

int COccDatum::STail::operator == (const STail &Other) const
{
	return Tip == Other.Tip &&
		InnerFirst == Other.InnerFirst &&
		InnerLast == Other.InnerLast;
}

int COccDatum::STail::operator != (const STail &Other) const
{
	return Tip != Other.Tip ||
		InnerFirst != Other.InnerFirst ||
		InnerLast != Other.InnerLast;
}

COccDatum::COccDatum(CStorage *aStorage, long aTicket):
	CDatum(aStorage, aTicket)
{
}

COccDatum::~COccDatum()
{
}

void COccDatum::Remove()
{	long Head;

	long Ticket = GetTicket();
	if (!Ticket ||
   	(Ticket & 0x80000000l))
		return;

	CConductor &Conductor = GetConductor();
	Conductor.Get(Ticket, 0, &Head, sizeof(long));
	if (Head & 0x80000000l)
		RemoveSegments(Head & ~0x80000000l);
	Conductor.Del(Ticket);
	SetTicket(0);
}

void COccDatum::RemoveSegments(long FirstSegmentCount)
{  STail Tail;

	long Ticket = GetTicket();
	Assert(Ticket != 0);
	CConductor &Conductor = GetConductor();
	Conductor.Get(Ticket,
		(1 + FirstSegmentCount) * sizeof(ulong),
		&Tail,
		sizeof(STail));

	long Next = Tail.InnerFirst;
	while (Next != Tail.InnerLast)
		Next = RemoveOne(Next);

	Conductor.Del(Tail.InnerLast);
	if (Tail.Tip != Tail.InnerLast)
		Conductor.Del(Tail.Tip);
}

long COccDatum::RemoveOne(long Segment)
{	long SegmentCount;
	long Next;

	CConductor &Conductor = GetConductor();
	Conductor.Get(Segment, 0, &SegmentCount, sizeof(long));
	Conductor.Get(Segment,
		(1 + SegmentCount) * sizeof(ulong),
		&Next,
		sizeof(long));
	Conductor.Del(Segment);
	return Next;
}

