#include "myfndt.h"
#include "packer.h"
#include "spider.h"

CPacker::CPacker()
{
}

int CPacker::GetPackedCount(ulong *OccArray, int Count)
{
    Assert(Count > 0);
    if (Count <= 2)
	   return Count;

    CSpider Spider(OccArray, Count, 0);
    Spider.Pack();
    return Spider.GetCount();
}

void CPacker::Pack(ulong *OccArray, int &Count)
{
    Assert(Count > 0);
    if (Count <= 2)
	   return;

    CSpider Spider(OccArray, Count, 1);
    Spider.Pack();
    Count = Spider.GetCount();
}

ulong *CPacker::AllocMerge(ulong *Occ1, int Count1, ulong *Occ2, int Count2, 
        int &OutCount)
{
    OutCount = Count1 + Count2;
    CMemPtr<ulong> OutOcc((short)(OutCount * sizeof(ulong)));
    if (Occ1)
	   memcpy((ulong *)OutOcc, Occ1, Count1 * sizeof(ulong));
    if (Occ2)
	   memcpy((ulong *)OutOcc + Count1, Occ2, Count2 * sizeof(ulong));
    Pack((ulong *)OutOcc, OutCount);
    return OutOcc.Drop();
}

ulong *CPacker::AllocAdd(ulong *PackedArr, int PackedCount, ulong *SparseArr,
	   int SparseCount, int &OutCount)
{   int i;
    ulong *Anchor;

    Assert(SparseArr != NULL && SparseCount > 0);

    for (i = 0, Anchor = PackedArr; 
	   i < PackedCount;
	   i++, Anchor++)
    {   if (!((*Anchor) & 0x80000000l))
		  return AllocAddComposed(PackedArr, i, Anchor, PackedCount - i,
		          SparseArr, SparseCount, OutCount);
    }

    Pack(SparseArr, SparseCount);
    OutCount = PackedCount + SparseCount;
    CMemPtr<ulong> OutOcc((short)(OutCount * sizeof(ulong)));
    if (PackedArr)
	   memcpy((ulong *)OutOcc, PackedArr, PackedCount * sizeof(ulong));
    memcpy((ulong *)OutOcc + PackedCount, SparseArr, SparseCount * sizeof(ulong));
    return OutOcc.Drop();
}

ulong *CPacker::AllocAddComposed(ulong *Squeezed, int SqueezedCount, 
	   ulong *Potencial, int PotencialCount, ulong *Sparse, int SparseCount, 
	   int &OutCount)
{
    Assert(Potencial != NULL && PotencialCount > 0);

    CMemPtr<ulong> OutOcc((short)(SqueezedCount +
	   GetUpperBound(PotencialCount) +
	   SparseCount) * sizeof(ulong));
    if (SqueezedCount)
	   memcpy((ulong *)OutOcc, Squeezed, SqueezedCount * sizeof(ulong));

    ulong *RepackArr = (ulong *)OutOcc + SqueezedCount;
    Assert(!((*Potencial) & 0x80000000l));
    *RepackArr = *Potencial;

    ulong *Src = Potencial + 1;
    ulong *Dest = RepackArr + 1;
    ulong *Last = Potencial + PotencialCount - 1;
//    while (Last - Src > 0)
	// 28. 2.
    while (Last - Src >= 0)
	   UnpackStep(Dest, Src);

    int RepackCount = (int)(Dest - RepackArr);
    if (Sparse)
    {   memcpy(Dest, Sparse, SparseCount * sizeof(ulong));
	    RepackCount += SparseCount;
    }

    Pack(RepackArr, RepackCount);
    OutCount = SqueezedCount + RepackCount;
    return OutOcc.Drop();
}

void CPacker::UnpackStep(ulong *&Dest, ulong *&Src)
{
    ulong Occ = *Src;
    if (Occ & 0x80000000l)
    {    *Dest = ExtractFirst(*(Dest - 1), Occ);
	    *(Dest + 1) = ExtractSecond(*Dest, Occ);
	    Dest += 2;
    }
    else
    {    *Dest = Occ;
	    Dest++;
    }
    Src++;
}
	    
