#include "myfndt.h"
#include "preader.h"
#include "passover.h"
#include "watch.h"

//#define DEBUG_Pietrek

const short CReader::MaxReadSize = 3000;

CReader::CReader(const DString &aBreaks,
    CWatch *InWatch):
    Source(NULL),
    DocStartOffs(-1),
    SeekOffs(-1),
    FileSize(-1),
    ChunkStartOffs(-1),
    Chunk(MaxReadSize + 1), // place for terminating 0
    CurrentPos(-1),
    ReadSize(-1),
    Watch(InWatch),
    Margin(InWatch->GetMargin()),
    CurrentSubdoc(0),
    Breaks(aBreaks.GetLength(), 0)
{
    Assert(Margin < MaxReadSize / 2);
    int Len = aBreaks.GetLength();
    for (int i = 0; i < Len; ++i)
        Breaks.Add(aBreaks[i]);
}

CReader::~CReader()
{
    delete Watch;
}

void CReader::SetIn(const DString &FName, long Offs)
{
    Source.Open(FName.c_str(), "rb");
    Assert(Offs >= 0);
    DocStartOffs = Offs;
    SeekOffs = DocStartOffs;
    Source.Seek(0, SEEK_END);
    FileSize = Source.Tell();
    Assert(Offs <= FileSize);
    SeekOffs = Offs;
    InitPos();
}

short CReader::GetPerc()
{   long Offs;

    if (ChunkStartOffs == -1)
    {   if (DocStartOffs >= 0)
            Offs = DocStartOffs;
        else
            Offs = 0;
    }
    else
        Offs = GetOffsAfter(ChunkStartOffs);

    short Perc = (100 * Offs) / (FileSize + 1);
    if (Perc > 100)
        return 100;

    return Perc;
}

void CReader::Seek(long Offs)
{
    Assert(DocStartOffs >= 0 && Offs >= DocStartOffs);
    SeekOffs = Offs;
    InitPos();
}

void CReader::InitPos()
{
    Change.Reset();
    ChunkStartOffs = -1;
    CurrentPos = -1;
    ReadSize = 0;
}

CReader::EShiftRes CReader::SkipWhite(long &Offs)
{	EInternRes docend;

    if (ChunkStartOffs == -1)
    {   if (DocStartOffs >= 0)
            Offs = DocStartOffs;
        else
            Offs = 0;
    }
    else
        Offs = GetOffsAfter(ChunkStartOffs);

	int NewLine = 0;
	while ((docend = GetChar()) == ir_ok)
    {   char c = *(((char *)Chunk) + CurrentPos);
        if (!IsSeparator(c))
            break;
		if (c == '\n')
			NewLine = 1;
	}
	if (docend != ir_ok)
		return EndDoc(docend);

	Offs = ChunkStartOffs + CurrentPos;
    if (Change.IsAfterSign(CurrentPos))
    {   // theoretically, there can be more than one matching regex between
        // adjacent words; not actually tested, though
        int ReallyChanged = 0;
        do
        {   short OldSubdoc = CurrentSubdoc;
            CurrentSubdoc = Change.GetChangedSubdoc();
            if (OldSubdoc != CurrentSubdoc)
                ReallyChanged = 1;
            Watch->Iterate((char *)Chunk, Change);
        }   while (Change.IsAfterSign(CurrentPos));
        if (ReallyChanged)
            return at_sign;
    }
	return NewLine ? over_newline : ok;
}

CReader::EShiftRes CReader::GetWord(DString &WordStr)
{   EInternRes docend;

	WordStr = "";
    Assert(CurrentPos >= 0 && CurrentPos < Change.GetValidSize());

	CheckedCharCat(WordStr, *(((char *)Chunk) + CurrentPos));
	while ((docend = GetChar()) == ir_ok)
    {   char c = *(((char *)Chunk) + CurrentPos);
        if (IsSeparator(c))
        {
#if defined(DEBUG_Pietrek)
            printf("*** read %s\n", WordStr.c_str());
#endif
            return c == '\n'? over_newline : ok;
        }
		CheckedCharCat(WordStr, c);
	}
#if defined(DEBUG_Pietrek)
    printf("*** read %s\n", WordStr.c_str());
#endif
	return EndDoc(docend);
}

CReader::EShiftRes CReader::EndDoc(EInternRes Reason)
{
#if defined(DEBUG_Pietrek)
    printf("*** in EndValue: ChunkStartOffs = %d, FileSize = %d, "
            "ReadSize = %d\n", ChunkStartOffs, FileSize, ReadSize);
#endif
    Assert(0 <= ChunkStartOffs && ChunkStartOffs <= FileSize);
    Assert(Reason != ir_ok);

    if (Reason == ir_eof)
        return at_eof;

    DocStartOffs = GetOffsAfter(ChunkStartOffs);
    SeekOffs = DocStartOffs;
    InitPos();
    return at_docend;
}

CReader::EInternRes CReader::GetChar()
{
    ++CurrentPos;
    if (CurrentPos < Change.GetValidSize() - Margin)
        return ir_ok;

    if (Change.GetValidSize() == MaxReadSize)
        return ReadChunk();
    else
    {   if (CurrentPos < Change.GetValidSize())
            return ir_ok;
        else
        {   if (!Change.GetValidSize())
                return ReadChunk();
            else
                return ir_docend;
        }
    }
}

CReader::EInternRes CReader::ReadChunk()
{
    if (!FileSize)
    {   ChunkStartOffs = 0;
        return ir_eof;
    }

    if (ChunkStartOffs == -1)
        ChunkStartOffs = SeekOffs;
    else
    {   if (FileSize - ChunkStartOffs <= ReadSize)
            return ir_eof;

        Assert(ReadSize == MaxReadSize);
        ChunkStartOffs += ReadSize - Margin;
    }
    CurrentPos = 0;

    Assert(ChunkStartOffs <= FileSize);
    long TailFileSize = FileSize - ChunkStartOffs;
    ReadSize = MaxReadSize <= TailFileSize ? MaxReadSize : (short)TailFileSize;
    if (!ReadSize)
    {   ChunkStartOffs = 0;
        return ir_eof;
    }

    Source.Seek(ChunkStartOffs);
    Source.Read((char *)Chunk, ReadSize);
    *(((char *)Chunk) + ReadSize) = 0;
    // do not read binary files
    if (strlen((char *)Chunk) < ReadSize)
    {   // set things up so that EndValue returns at_eof 
        //ValidSize = 0;
        ReadSize = 0;
        ChunkStartOffs = FileSize;
        return ir_eof;
    }

    // ? might be optimized by skipping Iterate when 
    // CurrentPos >= Change.GetValidSize() ?
    Change.SetValidSize(Watch->GetValidSize((char *)Chunk, 
        ReadSize,
        ChunkStartOffs == DocStartOffs));
    Watch->Iterate((char *)Chunk, Change);
    if (CurrentPos >= Change.GetValidSize())
        return ir_docend;
    return ir_ok;
}

void CReader::CheckedStrCat(DString &Target, const DString &Tail)
{
    if (Target.GetLength() + Tail.GetLength() > DString::MaxAllowedLength)
        throw CPassageOverflow();
	Target += Tail;
}

void CReader::CheckedCharCat(DString &Target, char Tail)
{
    if (Target.GetLength() + 1 > DString::MaxAllowedLength)
        throw CPassageOverflow();
	Target += Tail;
}

/*int CReader::IsSeparator(char C)
{
	return C == ' ' || C == '\n' || C == '\r' || C == '\t';
}*/

ushort CReader::GetWords(ushort Count, DString &Words)
{	DString One;

	Words = "";
	for (ushort i = 0; i < Count; i++)
	{	if (!GetWordWithWhite(One))
			CheckedStrCat(Words, One);
		else
			return i;
	}
	return Count;
}

ushort CReader::GetAllWords(DString &Words)
{	DString One;

	Words = "";
	ushort i = 0;
	while (!GetWordWithWhite(One))
	{	++i;
		CheckedStrCat(Words, One);
	}
	return i;
}

int CReader::GetWordWithWhite(DString &WordStr)
{	char c;
	EInternRes docend;
	int dummy;

	DString WhiteStr = "";
	while ((docend = GetChar()) == ir_ok)
    {   char c = *(((char *)Chunk) + CurrentPos);
        if (!IsSeparator(c))
            break;
	     CheckedCharCat(WhiteStr, c);
    }
	if (docend != ir_ok)
		return 1;

	WordStr = "";
	CheckedCharCat(WordStr, *(((char *)Chunk) + CurrentPos));
	while ((docend = GetChar()) == ir_ok)
    {   char c = *(((char *)Chunk) + CurrentPos);
        if (IsSeparator(c))
            break;
	    CheckedCharCat(WordStr, c);
    }

    if (WordStr.GetLength() + WhiteStr.GetLength() > DString::MaxAllowedLength)
	     throw CPassageOverflow();
	WordStr = WhiteStr + WordStr;
	if (docend == ir_ok)
		CheckedCharCat(WordStr, *(((char *)Chunk) + CurrentPos));
	return 0;
}

long CReader::GetOffsAfter(long ChunkStartOffs)
{
    Assert(Change.GetValidSize() >= 0);
    return ChunkStartOffs + Change.GetValidSize();
}
