#if !defined(__preader_h)
#define __preader_h

#include "rxw.h"
#include "arrays.h"
#include "change.h"

class CWatch;

class CReader
{
public:
    enum EShiftRes { ok, over_newline, at_sign, at_docend, at_eof };

public:
    CReader(const DString &aBreaks, CWatch *InWatch);
    ~CReader();

    void SetIn(const DString &FName, long Offs);
    void Seek(long Offs);
    short GetPerc();

    EShiftRes SkipWhite(long &Offs); 
    EShiftRes GetWord(DString &WordStr); // never returns at_sign
    short GetCurrentSubdoc() { return CurrentSubdoc; }

    ushort GetWords(ushort Count, DString &Words);
	// Words is out Param
	// returns number of words read (<= Count)
    ushort GetAllWords(DString &Words);
	// Words is out Param
	// returns number of words read

private:
    enum EInternRes { ir_ok, ir_docend, ir_eof };

    CFileObj Source;
    long DocStartOffs;
    long SeekOffs;
    long FileSize;

    long ChunkStartOffs;
    CMemPtr<char> Chunk;
    short CurrentPos;
    static const short MaxReadSize;
    short ReadSize;
    short ValidSize;
    const short Margin;
    short SafeSize;

    CWatch *Watch;
    CChange Change;
    short CurrentSubdoc;
    DSArray<char> Breaks;

    void InitPos();
    EInternRes GetChar();
    EInternRes ReadChunk();
    static short GetValidSize(const char *Chunk, short ReadSize, 
	    const CRegex &DocBegin, int FromBegin);

    EShiftRes EndDoc(EInternRes Reason);

    int GetWordWithWhite(DString &WordStr);
	// Word is out param
       	// returns 0 OK (Word valid), 1 eof

    int IsSeparator(char C)
    {	DSArray<char>::iterator iter = lower_bound(Breaks.begin(),
                Breaks.end(), C);
	if (iter == Breaks.end() || !(*iter == C))
	    return 0;
	return 1;
    }

    void CheckedStrCat(DString &Target, const DString &Tail);
    void CheckedCharCat(DString &Target, char Tail);

    long GetOffsAfter(long ChunkStartOffs);
};

#endif

