#include "myfndt.h"
#include "recbase.h"
#include "tl.h"

//#include <stdlib.h>

const uchar CRecBase::EmptyFlag = 0x01;

CRecBase::CRecBase(const DString &BaseName, ushort Opt, ushort aRecSize):
	Stream(NULL), 
    RecSize(aRecSize),
    FName(BaseName)
{	DString Dummy;

	Assert(!(Opt & IT_TEMP)); // not implemented, because it's not needed
    if (OpenData(BaseName, Opt, Dummy, Stream))
	{   SBaseHeader Header;
        Header.HoleCount = 0;
		WriteHeader(Header);
	}
}

CRecBase::~CRecBase()
{
}

ulong CRecBase::Add()
{
	Stream.Seek(0, SEEK_END);
	ulong Offs = Stream.Tell();
	Stream.WriteZero(GetInternalRecSize());
	return Offs;
}

void CRecBase::Del1(ulong Offs)
{	SBaseHeader Header;

	uchar Flags = EmptyFlag;
	Stream.Seek(Offs);
	Stream.Write(&Flags, 1);
	ReadHeader(Header);
	Header.HoleCount++;
	WriteHeader(Header);
}

void *CRecBase::Alloc(ulong Offs)
{	uchar Flags;

	CMemPtr<void> Ptr(RecSize);
	Stream.Seek(Offs);
	Stream.Read(&Flags, 1);
	Assert(!(Flags & EmptyFlag));
	Stream.Read((void *)Ptr, RecSize);
	return Ptr.Drop();
}

void CRecBase::Put(ulong Offs, const void *Stuff)
{
	Stream.Seek(Offs + 1);
	Stream.Write(Stuff, RecSize);
}

short CRecBase::GetFirst(ulong &Offs)
{	uchar Flags;

	CMemPtr<void> Ptr(RecSize);

	Stream.Seek(0, SEEK_END);
	long LastOffs = Stream.Tell();
	if (LastOffs < GetInternalRecSize())
		return 1;
	LastOffs -= GetInternalRecSize();

	Offs = sizeof(SBaseHeader);
	Stream.Seek(Offs);
	while (LastOffs >= (long)Offs)
	{	Stream.Read(&Flags, 1);
		if (!(Flags & EmptyFlag))
			return 0;
		Stream.Read((void *)Ptr, RecSize);
		Offs += GetInternalRecSize();
	}
	return 1;
}

short CRecBase::GetLast(ulong &Offs)
{	uchar Flags;

	Stream.Seek(0, SEEK_END);
	Offs = Stream.Tell();
	if (Offs < GetInternalRecSize())
		return 1;
	Offs -= GetInternalRecSize();

	while (Offs >= sizeof(SBaseHeader))
	{  Stream.Seek(Offs);
		Stream.Read(&Flags, 1);
		if (!(Flags & EmptyFlag))
			return 0;

		Offs -= GetInternalRecSize();
	}
	return 1;
}

void CRecBase::ReadHeader(SBaseHeader &Header)
{
	Stream.Seek(0);
	Stream.Read(&Header, sizeof(SBaseHeader));
}

void CRecBase::WriteHeader(SBaseHeader &Header)
{
	Stream.Seek(0);
	Stream.Write(&Header, sizeof(SBaseHeader));
}

/*void CRecBase::PackStep()
{	uchar Flags;

	Stream.Seek(GetInternalRecSize(), SEEK_END);
	ulong Offs = Stream.Tell();
	while (Offs >= sizeof(SBaseHeader))
	{  Stream.Seek(Offs);
		Stream.Read(&Flags, 1);
		if (Flags & EmptyFlag)
		{	PackFromOffs(Offs);
			return;
		}
	}
}

void CRecBase::PackFromOffs(ulong Offs)
{	SBaseHeader Header;
	uchar Flags;

	ulong PackNum = 1;
	ulong Src = Offs + GetInternalRecSize();

	Offs -= GetInternalRecSize();
	while (Offs >= sizeof(SBaseHeader))
	{  Stream.Seek(Offs);
		Stream.Read(&Flags, 1);
		if (!(Flags & EmptyFlag))
			break;
		Offs -= GetInternalRecSize();
		PackNum++;
	}
	Offs += GetInternalRecSize();

	MoveTail(Offs, Src);
	ReadHeader(Header);
	Assert(Header.HoleCount >= PackNum);
	Header.HoleCount -= PackNum;
	WriteHeader(Header);
}

void CRecBase::MoveTail(ulong Dest, ulong Src)
{	size_t CurSize;

	Assert(Dest < Src);
	ulong Delta = Src - Dest;
	Stream.Seek(0, SEEK_END);
	ulong Size = Stream.Tell();
	Assert(Size > Src);
	Size -= Src;
	if (Delta > 512) // just a picked number
		CurSize = 512;
	else
		CurSize = static_cast<size_t>(Delta);

	CMemPtr<char> Buf(CurSize);
	while (Size > 0)
	{	if (Size < CurSize)
			CurSize = static_cast<size_t>(Size);

		Stream.Seek(Src);
		Stream.Read((char *)Buf, CurSize);
		Stream.Seek(Dest);
		Stream.Write((char *)Buf, CurSize);

		Src += CurSize;
		Dest += CurSize;
		Size -= CurSize;
	}
}*/

