#include "myfndt.h"
#include "subdatum.h"
#include "subdesc.h"

//#define DEBUG_CSubDesc

CSubDatum::CSubDatum(CStorage *aStorage, long aTicket):
	CDatum(aStorage, aTicket),
    Dirty(0),
    SubDesc(NULL)
{
}

void CSubDatum::InitSubDesc()
{
#if defined(DEBUG_CSubDesc)
    printf("begin CSubDatum::InitSubDesc\n");
#endif
    long Ticket = GetTicket();
    Assert(Ticket != 0);
    CConductor &Conductor = GetConductor();

    short Count;
    Conductor.Get(Ticket, 0, &Count, sizeof(short));
#if defined(DEBUG_CSubDesc)
    printf("doc on %d has %d parts\n", Ticket, Count);
#endif
    CMemPtr<long> Stairs(Count * sizeof(long));
    Conductor.Get(Ticket,
        sizeof(short),
		(long *)Stairs,
      	Count * sizeof(long));
    CMemPtr<short> Ids(Count * sizeof(short));
    Conductor.Get(Ticket,  
        sizeof(short) + Count * sizeof(long),
		(short *)Ids,
        Count * sizeof(short));

    Assert(!SubDesc);
    SubDesc = new CSubDesc;
    for (short i = 0; i < Count; ++i)
    {
#if defined(DEBUG_CSubDesc)
        printf("Stair: %d\t\tId: %d\n", 
            *(((long *)Stairs) + i), 
            *(((short *)Ids) + i));
#endif
        SubDesc->Add(*(((long *)Stairs) + i), *(((short *)Ids) + i));
    }
#if defined(DEBUG_CSubDesc)
    printf("end CSubDatum::InitSubDesc\n");
#endif
}

CSubDatum::~CSubDatum()
{
    delete SubDesc;
}

long CSubDatum::Finish()
{
    return !Dirty ? 0 : GetTicket();
}

void CSubDatum::SetDesc(CSubDesc *InSubDesc)
{
    Assert(!SubDesc && !GetTicket());
    SubDesc = InSubDesc;
    StoreSubDesc();
    Dirty = 1;
}

void CSubDatum::StoreSubDesc()
{
    Assert(SubDesc != NULL);
    short Count = SubDesc->GetCount();
    CConductor &Conductor = GetConductor();
	long CurTicket = Conductor.Add(sizeof(short) + 
        Count * sizeof(long) +
        Count * sizeof(short));
#if defined(DEBUG_CSubDesc)
    if (CurTicket == 324)
    {   printf("storing %d parts on 324\n", Count);
        for (int i = 0; i < Count; ++i)
            printf("Stair: %d\t\tId: %d\n", 
                *((SubDesc->GetStairs()) + i), 
                *((SubDesc->GetIds()) + i));
    }
#endif
	Conductor.Put(CurTicket, 0, &Count, sizeof(short));
	Conductor.Put(CurTicket,
		sizeof(short),
		SubDesc->GetStairs(),
		Count * sizeof(long));
	Conductor.Put(CurTicket,
		sizeof(short) + Count * sizeof(long),
		SubDesc->GetIds(),
		Count * sizeof(short));

	SetTicket(CurTicket);
}

CSubDesc *CSubDatum::NewSubDesc()
{
    if (!SubDesc)
        InitSubDesc();
    Assert(SubDesc != NULL);
    CSubDesc *Out = SubDesc;
    SubDesc = NULL;
    return Out;
}

