#include "myfndt.h"
#include "watch.h"
#include "locusopt.h"
#include "change.h"

#include "osubdoc.h"
#include "oreader.h"

#include <stdio.h>

//#define DEBUG_newline

#if defined(DEBUG_newline)
void PrintMatch(const char *id, const char *chunk, int from, int to);
void PrintNoMatch(const char *id, const char *chunk);
#endif

CWatch::CRegexArray::CRegexArray()
{
}

void CWatch::CRegexArray::Add(const DString &Regex, int IgnoreCase)
{
    int CompFlags = CRegex::DefaultCompFlags;
    if (IgnoreCase)
        CompFlags |= CRegex::Icase;
    push_back(CRegex(Regex, CompFlags));
}

CWatch::CWatch(const CRegex &aDocBegin, short aMargin):
    Margin(0),
	DocBegin(aDocBegin)
{
    if (!DocBegin.IsEmpty())
 		Margin = aMargin;
}	

CWatch::CWatch(CLocusOptions &Options):
    Margin(0),
	DocBegin(Options.Get(O_DIVIDER))
{
    if (!DocBegin.IsEmpty())
 		Margin = Options.COptions<int>::Get(O_READERMARGIN);

    CStringListValue List = Options.COptions<CStringListValue>::Get(O_DOCPARTS);
    CStringListValue::DIterator Itr = List.InitIterator();
    while (List.IsDereferenceable(Itr))
    {   DString Name = *Itr;
        PushMatch.Add(Options.COptions<DString>::GetUno(Name, O_PUSHPART),
            !Options.COptions<int>::GetUno(Name, O_PUSHCASESENSITIVE));
        PopMatch.Add(Options.COptions<DString>::GetUno(Name, O_POPPART),
            !Options.COptions<int>::GetUno(Name, O_POPCASESENSITIVE));
        Itr++;
    }
}

void CWatch::SetSubdocClass(CUnivOptions &Options,
	const CStringListValue &DocParts)
{   static const char *StringAttr[] = {O_PUSHPART,
        O_POPPART,
        O_CRUFT,
        O_SUBTEMPL_1};
    static const char *IntAttr[] = {O_PUSHCASESENSITIVE,
        O_POPCASESENSITIVE,
        O_CRUFTCASE};

    Options.COptions<DString>::SetUnoClass(DocParts, StringAttr,
            sizeof(StringAttr) / sizeof(StringAttr[0]));
    Options.COptions<int>::SetUnoClass(DocParts, IntAttr,
            sizeof(IntAttr) / sizeof(IntAttr[0]));
}

/* profile 3. 2. 1997: VIRUS-L, optimized build
with only document divider: 5 hr 43 min
with 3 doc parts (from, subject, quote): 6 hr 10 min */
void CWatch::Iterate(const char *Chunk, CChange &Change)
{   static const int PushIndex = 0;
    static const int PopIndex = 1;
    static const int PushMatch = 1;
    static const int PopMatch = 2;
    int From[2], To[2];
    short SubdocId;

    int HeadSize = Change.GetPosAfterSign();
    Change.ClearSign();

    const char *Tail = Chunk + HeadSize;
    int t = 0;
    if (GetPushMatch(Tail, From[PushIndex], To[PushIndex], SubdocId))
        t += PushMatch;
    short OldSubdoc = Change.GetChangedSubdoc();
    if (OldSubdoc && GetPopMatch(Tail, OldSubdoc, From[PopIndex], To[PopIndex]))
        t += PopMatch;

    if (!t)
    {
#if defined(DEBUG_newline)
        printf("no match at all\n\n");
#endif
        return;
    }

    if (t == PushMatch + PopMatch)
    {   if (From[PushIndex] < From[PopIndex])
            t = PushMatch;
        else
        {   if (From[PushIndex] > To[PopIndex])
                t = PopMatch;
            else
            {   
#if defined(DEBUG_newline)
                printf("matching both pop and push\n\n");
#endif
                Change.Pop();
                Change.SetSign(From[PushIndex] + HeadSize, 
                    To[PushIndex] + HeadSize);
                Change.Push(SubdocId);
                return;
            }
        }
    }

    if (t == PushMatch)
    {   
#if defined(DEBUG_newline)
        printf("matching push\n\n");
#endif
        Change.SetSign(From[PushIndex] + HeadSize, To[PushIndex] + HeadSize);
        Change.Push(SubdocId);
    }
    else
    {   
#if defined(DEBUG_newline)
        printf("matching pop\n\n");
#endif
        Change.SetSign(From[PopIndex] + HeadSize, To[PopIndex] + HeadSize);
        Change.Pop();
    }
}

short CWatch::GetValidSize(const char *Chunk, short ReadSize, int FromBegin)
{   int From, To, From2, To2;
    
    if (!DocBegin.Matches(Chunk, From, To))
        return ReadSize;

    if (!FromBegin || From > 0)
    {   Assert(From < ReadSize);
        return From;
    }

    if (To >= ReadSize)
        return ReadSize;

    if (!DocBegin.Matches(Chunk + To, From2, To2))
        return ReadSize;
    else
    {   Assert(To + From2 < ReadSize);
        return To + From2;
    }
}

int CWatch::GetPushMatch(const char *Chunk,
    int &From,
	int &To,
	short &SubdocId)
{   int CurrentFrom, CurrentTo;

    int ChunkLen = strlen(Chunk);
    int Count = PushMatch.size();
    From = ChunkLen;
#if defined(DEBUG_newline)
    printf("checking push rxs...\n");
#endif
    for (CRegexArray::iterator i = PushMatch.begin();
         i != PushMatch.end();
         ++i)
    {   
        if ((*i).Matches(Chunk, CurrentFrom, CurrentTo))
        {   Assert(CurrentFrom < ChunkLen);
            if (CurrentFrom < From)
            {   From = CurrentFrom;
                To = CurrentTo;
                SubdocId = (i - PushMatch.begin()) + 1;
            }
        }
    }
#if defined(DEBUG_newline)
    if (From < ChunkLen)
        printf("...%d won\n", SubdocId);
    else
        printf("...no match\n");
#endif
    return From < ChunkLen;
}

int CWatch::GetPopMatch(const char *Chunk,
	short Subdoc,
    int &From,
	int &To)
{
    Assert(Subdoc > 0);
    int r = PopMatch[Subdoc - 1].Matches(Chunk, From, To);
#if defined(DEBUG_newline)
    if (r)
        PrintMatch("pop rx", Chunk, From, To);
    else
        PrintNoMatch("pop rx", Chunk);
#endif
    return r;
}

short CWatch::GetMargin() const
{
    return Margin;
}

#if defined(DEBUG_newline)
char *AllocDigest(const char *str)
{
    int size = strlen(str) + 1;
    CMemPtr<char> m(size);
    memcpy((char *)m, str, size);
    if (size > 15)
        *(((char *)m) + 15) = 0;
    return m.Drop();
}

void PrintMatch(const char *id, const char *chunk, int from, int to)
{
    CMemPtr<char> mp = AllocDigest(chunk + from);
    CMemPtr<char> mw = AllocDigest(chunk);
    printf("%s matches %s (%d - %d) in %s\n",
        id,
        (char *)mp,
        from,
        to,
        (char *)mw);
}

void PrintNoMatch(const char *id, const char *chunk)
{
    CMemPtr<char> m = AllocDigest(chunk);
    printf("%s does not match anywhere in %s\n", id, (char *)m);
}
#endif
