#include "fileobj.h"
#include "myfndt.h"

#include <stdarg.h>
#include <string.h>

//#define DEBUG_of

CFileObj::CFileObj(const char *FName, const char *Flags):
	Handle(NULL)
{
	Init(FName, Flags);
}

CFileObj::CFileObj(FILE *aHandle):
	Handle(aHandle)
{
}

CFileObj::~CFileObj()
{
	Done();
}

void CFileObj::Open(const char *FName, const char *Flags)
{
	Done();
	Init(FName, Flags);
}

void CFileObj::Assign(FILE *aHandle)
{
	Done();
	Handle = aHandle;
}

void CFileObj::Seek(long Offs, int Whence)
{
	if (fseek(Handle, Offs, Whence))
		throw CReadException();
}

void CFileObj::Write(const void *Ptr, size_t Size, size_t Num)
{
	if (fwrite(Ptr, Size, Num, Handle) != Num)
	{
#if defined(DEBUG_of)
		printf("DEBUG_of: write into %x failed\n", Handle);
#endif
		throw CWriteException();
	}
}

void CFileObj::WriteZero(long Size)
{
	int CurSize = 200;
	CMemPtr<char> Buf(CurSize);
	memset((char *)Buf, 0, CurSize);
	while (Size > 0)
	{	if (Size < CurSize)
			CurSize = static_cast<int>(Size);
		Write((char *)Buf, CurSize);
		Size -= CurSize;
	}
}

void CFileObj::Read(void *Ptr, size_t Size, size_t Num)
{
	if (fread(Ptr, Size, Num, Handle) != Num)
		throw CReadException();
}

long CFileObj::Tell()
{
	long Offs = ftell(Handle);
	if (Offs < 0)
		throw CReadException();
	return Offs;
}

void CFileObj::Flush()
{
/*	if (!fflush(Handle))
		throw CWriteException();*/
   fflush(Handle);
}

void CFileObj::Init(const char *FName, const char *Flags)
{
	Handle = fopen(FName, Flags);
#if defined(DEBUG_of)
	printf("DEBUG_of: %s (opened with %s) has handle %x\n",
		FName,
		Flags,
		Handle);
#endif
	if (!Handle)
		throw COpenException(DString(FName), DString(Flags));
}

void CFileObj::Done()
{
	if (Handle)
	{	fclose(Handle);
		Handle = NULL;
	}
}

void CFileObj::Print(const char *Format, ...)
{	va_list List;

	va_start(List, Format);
	int Count = vfprintf(Handle, Format, List);
	va_end(List);
	if (Count == EOF)
	{
#if defined(DEBUG_of)
		printf("DEBUG_of: can't print %s into %x\n", Format, Handle);
#endif
		throw CWriteException();
	}
}

