#include "myfndt.h"

#if defined(PRINT_BEFORE_THROW)
#include <stdio.h>
#include <syslog.h>
#endif

#if defined(__GNUG__)
#include <unistd.h>
#endif

#if defined(__GNUG__)
xmsg::xmsg(const DString &msg):
    str(msg)
{
#if defined(PRINT_BEFORE_THROW)
    const char *format = "exception: %s\n";
	fprintf(stderr, format, msg.c_str());
    syslog(LOG_ERR, format, msg.c_str());
#endif
}
#endif   

CGeneralException::CGeneralException(const char *Text):
	DException(Text)
{
}

CGeneralException::CGeneralException(const CGeneralException &Source):
	DException(Source.
#if defined(__BORLANDC__) || defined(__GNUG__)
		why
#else
		what
#endif
			())
{
}

CGeneralException::~CGeneralException()
{
}

CGeneralException &CGeneralException::operator = (
		CGeneralException &Source)
{
	DException::operator = (Source);
	return *this;
}

DMemoryException::DMemoryException(size_t size):
#if !defined(_MSC_VER)
	xalloc(DString("memory"), size)
#else
	DException("memory"),
	Size(size)
#endif
{
}

DMemoryException::DMemoryException(const DMemoryException &Source):
#if !defined(_MSC_VER)
	xalloc(DString("memory"), Source.requested())
#else
	DException("memory"),
	Size(Source.Size)
#endif
{
}

DMemoryException &DMemoryException::operator = (
		DMemoryException &Source)
{
#if !defined(_MSC_VER)
	xalloc::operator = (Source);
#else
	DException::operator = (Source);
	Size = Source.Size;
#endif
	return *this;
}

COpenException::COpenException(const DString &aFile, const DString &aMode):
	DException("open"),
	File(aFile),
	Mode(aMode)
{
#if defined(PRINT_BEFORE_THROW)
    static const char *format = "exception: could not open %s with mode %s\n";
    fprintf(stderr, format, aFile.c_str(), aMode.c_str());
    syslog(LOG_ERR, format, aFile.c_str(), aMode.c_str());
#endif
}

COpenException::COpenException(const COpenException &Other):
	DException("open"),
	File(Other.File),
	Mode(Other.Mode)
{
}

COpenException &COpenException::operator = (
    COpenException &Other)
{
	File = Other.File;
	Mode = Other.Mode;
	return *this;
}

const char *COpenException::GetFile()
{
	return File.IsEmpty() ? "*unknown*" : File.c_str();
}

const char *COpenException::GetMode()
{
	return Mode.IsEmpty() ? "*unknown*" : Mode.c_str();
}

CReadException::CReadException():
	DException("read")
{
}

CReadException::CReadException(const CReadException &):
	DException(DString("read"))
{
}

CReadException &CReadException::operator = (
	CReadException &)
{
	return *this;
}

CWriteException::CWriteException():
	DException(DString("write"))
{
}

CWriteException::CWriteException(const CWriteException &):
	DException("write")
{
}

CWriteException &CWriteException::operator = (
	CWriteException &)
{
	return *this;
}

DAssertException::DAssertException(const char *aFile,
	int aLine):
	DException("assertion failed"), File(aFile), Line(aLine)
{
}

DAssertException::DAssertException(const DAssertException &Source):
	DException(Source)
{
	File = Source.File;
	Line = Source.Line;
}

DAssertException &DAssertException::operator = (DAssertException &Source)
{
	File = Source.File;
	Line = Source.Line;
	DException::operator = (Source);
	return *this;
}


