#include "myfndt.h"
#include "univopt.h"

extern "C"
{
#include "paction.h"
#include "rferr.h"
};

static int SafeAdd(CStringListValue &List, 
	const DString &ValueStr);

CUnivOptions *PAGlobOptions = NULL;
	// set by override.cpp

char *AddInt(char *Section, char *Key, int Value)
{
	if (!PAGlobOptions->HasSection(Section))
		return NULL;

	if (PAGlobOptions->Override(Section, Key, Value))
	{	// it's not int; it can be double
		double DoubleVal = Value;
		if (PAGlobOptions->Override(Section, Key, DoubleVal))
			return PAGlobOptions->UnknownKeyError(Key);
	}
    else
    {   // if it's int, it can't be double
	PAGlobOptions->COptions<double>::AssertCanAdd(Key);
    }

	return NULL;
}

char *AddStr(char *Section, char *Key, char *Value)
{
	DString ValueStr = Value;
	free(Value);

	if (!PAGlobOptions->HasSection(Section))
		return NULL;

	if (PAGlobOptions->Override(Section, Key, ValueStr))
	{	// it's not a string; it can be a string list
		CStringListValue List(1, 0);
		if (SafeAdd(List, ValueStr))
			return OE_OUTOFMEM;
		if (PAGlobOptions->Override(Section, Key, List))
			return PAGlobOptions->UnknownKeyError(Key);
	}
    else
    {   // if it's string, it can't be string list
        PAGlobOptions->COptions<CStringListValue>::AssertCanAdd(Key);
    }
	return NULL;
}

char *AddList(char *Section, char *Key, SStrList *Value)
{
	if (!PAGlobOptions->HasSection(Section))
		return NULL;

	CStringListValue List(Value->Count, 0);
	for (int i = 0; i < Value->Count; i++)
		List.Add(Value->List[i]); // throws!

	if (PAGlobOptions->Override(Section, Key, List))
		return PAGlobOptions->UnknownKeyError(Key);

	return NULL;
}

char *AddDouble(char *Section, char *Key, double Value)
{
	if (PAGlobOptions->HasSection(Section))
	{	if (PAGlobOptions->Override(Section, Key, Value))
			return PAGlobOptions->UnknownKeyError(Key);
	}
   return NULL;
}

char *AllocMungedSectionName(char *InSection)
{
    return PAGlobOptions->AllocMungedSectionName(InSection);
}

static int SafeAdd(CStringListValue &List, 
	const DString &ValueStr)
{
	try
	{	List.Add(ValueStr); 
		return 0;
	}
	catch (...)
	{	return -2;
	}
}
