#if !defined(__univopt_h)
#define __univopt_h

#include "opt1.h"
#include "opt2.h"
#include "opt3.h"
#include "opt4.h"

#include "arrays.h"

// templates are split into multiple modules to keep segment
// under 64K
class CUnivOptions : public COptions1,
	public COptions2,
	public COptions3,
	public COptions4
{
// consts
private:
    static const int SectionLimit;
    static const int SectionDelta;

// members
private:
    DISArray<DString> SectList;
#if defined(LIGHT_COMPILE)
    DISArray<DString> Unsupported;
#endif
protected:
    int ComposedKeys;

public:
    CUnivOptions();
    virtual ~CUnivOptions();

    // returns empty string OK, otherwise key which has
    // initial invalid value
    virtual DString FirstInvalidKey();

    void WatchSection(const char *SectName);
    void ChangeWatchedSections(const CStringListValue &ToList);

    // bug: can throw and caller does not handle it
    virtual short HasSection(const char *SectName);

    void Add(const char *Key, const char *Value);
    // Borland C++ cannot distinguish overloaded functions from
    // sibling classes
    void Add(const char *Key, const DString &Value)
        { COptions<DString>::Add(Key, Value); }
    void Add(const char *Key, COptions<DString>::FCompute Compute)
      	{ COptions<DString>::Add(Key, Compute); }
    void Add(const char *Key, const CStringListValue &Value)
       	{ COptions<CStringListValue>::Add(Key, Value); }
    void Add(const char *Key,
       	COptions<CStringListValue>::FCompute Compute)
	{ COptions<CStringListValue>::Add(Key, Compute); }
    void Add(const char *Key, int Value) // no need to pass int by reference
        { COptions<int>::Add(Key, Value); }
    void Add(const char *Key, COptions<int>::FCompute Compute)
        { COptions<int>::Add(Key, Compute); }
    void Add(const char *Key, double &Value)
       	{ COptions<double>::Add(Key, Value); }
    void Add(const char *Key, COptions<double>::FCompute Compute)
        { COptions<double>::Add(Key, Compute); }

#if defined(LIGHT_COMPILE)
    void AddUnsupported(const char *Key)
	{ Unsupported.Add(new DString(Key)); }
#endif

        // dtto
        inline short Override(const char *Section, 
		const char *Key, 
		const DString &Value);
	inline short Override(const char *Section,
		const char *Key, 
		const CStringListValue &Value);
	inline short Override(const char *Section,
		const char *Key, 
		int Value);
	inline short Override(const char *Section,
		const char *Key, 
		const double &Value);

    // do not distinguish overloaded functions by return type
    // (rules are too subtle) - call them explicitly
    // COptions<DString>::Get is called most often, so it's the
    // only one with a short name
    DString Get(const char *Key) const
     	{ return COptions<DString>::Get(Key); }

    virtual DString GetKey(const char *Section, const char *Key) const;

    char *UnknownKeyError(const char *Key);
    int GetComposedMode() { return ComposedKeys; }
    void SetComposedMode(int ComposedMode) { ComposedKeys = ComposedMode; }

    virtual char *AllocMungedSectionName(char *InSection);
};

inline short CUnivOptions::Override(const char *Section,
	const char *Key, 
	const DString &Value)
{	
	return COptions<DString>::Override(Section, Key, Value); 
}

inline short CUnivOptions::Override(const char *Section,
	const char *Key, 
	const CStringListValue &Value)
{	
	return COptions<CStringListValue>::Override(Section, Key,
			Value); 
}

inline short CUnivOptions::Override(const char *Section,
	const char *Key, 
	int Value)
{ 
	return COptions<int>::Override(Section, Key, Value); 
}

inline short CUnivOptions::Override(const char *Section,
	const char *Key, 
	const double &Value)
{ 
	return COptions<double>::Override(Section, Key, Value); 
}

#endif

