#include "myfndt.h"
#include "catcher.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

CCatcher::CCatcher(const DString &aMemExcFormat,
    const DString &aOpenExcFormat,
    const DString &aUnknownExc):
    MemExcFormat(aMemExcFormat),
    OpenExcFormat(aOpenExcFormat),
    UnknownExc(aUnknownExc)
{   int l;

    Assert(!MemExcFormat.IsEmpty() && 
	   !OpenExcFormat.IsEmpty() && 
	   !UnknownExc.IsEmpty());

    ErrBufSize = 1000;
    l = 4 * MemExcFormat.GetLength();
    if (ErrBufSize < l)
	   ErrBufSize = l;

    l = 4 * OpenExcFormat.GetLength() + 256;
    if (ErrBufSize < l)
	   ErrBufSize = l;

    l = UnknownExc.GetLength() + 1;
    if (ErrBufSize < l)
	   ErrBufSize = l;

    if (ErrBufSize > 1024)
	   throw CGeneralException("too long error messages");
}

CCatcher::~CCatcher()
{
}

char *CCatcher::SafeCall(FThrowing ThrowingFnc, void *Par)
{
    char *ErrBuf = (char *)malloc(ErrBufSize);
    if (!ErrBuf)
	   abort();

    try
    {   char *Ret = ThrowingFnc(this, Par);
	   free(ErrBuf);
	   return Ret;
    }
    catch (DAssertException &x)
    {   DString File = x.GetFile();
	sprintf(ErrBuf, "assertion failed in file %s on line %d",
            File.c_str(),
            x.GetLine());
	return ErrBuf;
    }
    catch (DMemoryException &x)
    {   sprintf(ErrBuf, MemExcFormat.c_str(), x.requested());
	   return ErrBuf;
    }
    catch (COpenException &x)
    {   sprintf(ErrBuf, OpenExcFormat.c_str(), x.GetFile(), x.GetMode());
	   return ErrBuf;
    }
    catch (DException &x) 
    {   strncpy(ErrBuf, 
#if !defined(_MSC_VER)
		  x.why().c_str(),
#else
		  x.what(),
#endif
		  ErrBufSize - 1);
	   return ErrBuf;
    }
    catch (...)
    {   strncpy(ErrBuf, UnknownExc.c_str(), ErrBufSize - 1);
	   return ErrBuf;
    }
}
