#include "myfndt.h"
#include "qebb.h"
#include "oqbexc.h"
#include "formbase.h"
#include "binpar.h"
#include "sdlpar.h"
#include "filter.h"

extern "C"
{
#include "pqerr.h"

extern int yqparse(void);
extern FILE *yqin;
};

#include <ctype.h>

CQueryBuilderBase *CQueryBuilderBase::Global = NULL;

CQueryBuilderBase::CQueryBuilderBase(CUnivOptions &Options):
    CCatcher(Options.Get(T_QBMEMEXC_D),
	   Options.Get(T_QBOPENEXC_SS),
	   Options.Get(T_QBUNKNEXC)),
    OutErrFormat(Options.Get(T_PQWRAP_SD)),
    FinalFormula(NULL)
{
    Assert(!Global);
    Global = this;
}

CQueryBuilderBase::~CQueryBuilderBase()
{
    Assert(Global == this);
    Global = NULL;
}

char *CQueryBuilderBase::Build(const DString &QueryFile,
    CFormulaBase **OutFormula)
{
    *OutFormula = NULL;

    FILE *h = fopen(QueryFile.c_str(), "r");
    if (!h)
	   throw COpenException(QueryFile, DString("r"));

    yqin = h;
    query_parser_line = 0;
    query_parser_error[0] = 0;
    yqparse();
    fclose(h);

    CMemPtr<char> EndErr = NULL;
    if (query_parser_error[0] != 0)
    {    CMemPtr<char> OutErr(2 * (strlen(query_parser_error) + 
			  OutErrFormat.GetLength() + 10));
	    sprintf((char *)OutErr, 
		   OutErrFormat.c_str(), 
		   query_parser_error, 
		   query_parser_line);
	    return OutErr.Drop();
    }

    *OutFormula = FinalFormula;
    Assert((*OutFormula) != NULL);
    return NULL;
}

char *CQueryBuilderBase::StaticSetSoftOpName(CCatcher *Self, void *InName)
{
    CMemPtr<char> Name = reinterpret_cast<char *>(InName);

    Assert(Self == Global && Global != NULL);
    if ((char *)Name == NULL)
	   throw DMemoryException();

    return Global->SetSoftOpName((char *)Name);
}

char *CQueryBuilderBase::StaticAddSoftOpArgs(CCatcher *Self, void *InString)
{
    CMemPtr<char> Str = reinterpret_cast<char *>(InString);

    Assert(Self == Global && Global != NULL);
    if ((char *)Str == NULL)
	   throw DMemoryException();

    int l = strlen((char *)Str);
    char *p = (char *)Str;
    for (int i = 0; i < l; i++)
    {   char c = *(((char *)Str) + i);
	if (isspace(c))
	{   *(((char *)Str) + i) = 0;
	    if (*p)
	    {   char *Err = Global->AddSoftOpArg(p);
   	        if (Err)
		    return Err;

		p = (char *)Str + i + 1;
	   }
       }
    }

    if (*p)
	   return Global->AddSoftOpArg(p);
    else
        return NULL;
}

char *CQueryBuilderBase::StaticCreateTerm(CCatcher *Self, void *OutFormula)
{
    Assert(Self == Global && Global != NULL);
    return Global->CreateTerm(reinterpret_cast<CFormulaBase **>(OutFormula));
}

char *CQueryBuilderBase::StaticCreateBinary(CCatcher *Self, void *InBinPar)
{
    CMemPtr<SBinaryPar> BinPar = reinterpret_cast<SBinaryPar *>(InBinPar);

    Assert(Self == Global && Global != NULL);
    if (!InBinPar)
        throw DMemoryException();

    return Global->CreateBinary(reinterpret_cast<SBinaryPar *>(InBinPar)->Op, 
        reinterpret_cast<SBinaryPar *>(InBinPar)->Left, 
        reinterpret_cast<SBinaryPar *>(InBinPar)->Right,
        reinterpret_cast<SBinaryPar *>(InBinPar)->Out);
}

char *CQueryBuilderBase::StaticSetFinalFormula(CCatcher *Self, void *Formula)
{
    Assert(Self == Global && Global != NULL);
    Assert(Formula != NULL && Global->FinalFormula == NULL);
    Global->FinalFormula = reinterpret_cast<CFormulaBase *>(Formula);
    return NULL;
}

char *CQueryBuilderBase::StaticResetSoftOpName(CCatcher *Self, void *Dummy)
{
    Assert(Self == Global && Global != NULL);
    Assert(Dummy == NULL);
    return Global->ResetSoftOpName();
}

char *CQueryBuilderBase::StaticCreateSubdocList(CCatcher *Self, void *InCreateSDPar)
{
    CMemPtr<SCreateSDPar> CreateSDPar = reinterpret_cast<SCreateSDPar *>(
        InCreateSDPar);
    if (!InCreateSDPar)
        throw DMemoryException();

    CMemPtr<char> Name = reinterpret_cast<SCreateSDPar *>(InCreateSDPar)->InName;
    if (Name == NULL)
        throw DMemoryException();

    // in case of error, leaks CSubdocFilter
    CSubdocFilter *Alias = new CSubdocFilter;
    *(reinterpret_cast<SCreateSDPar *>(InCreateSDPar)->Out) = reinterpret_cast<
        SRestrictedDomain *>(Alias);
    return Global->AddSubdocName(Alias, 
        reinterpret_cast<SCreateSDPar *>(InCreateSDPar)->InName);
}

char *CQueryBuilderBase::StaticAddSubdocName(CCatcher *Self, void *InAddSDPar)
{
    CMemPtr<SAddSDPar> AddSDPar = reinterpret_cast<SAddSDPar *>(InAddSDPar);
    if (!InAddSDPar)
        throw DMemoryException();

    CMemPtr<char> Name = reinterpret_cast<SAddSDPar *>(InAddSDPar)->InName;
    if (Name == NULL)
        throw DMemoryException();

    // in case of error, leaks CSubdocFilter
    return Global->AddSubdocName(reinterpret_cast<CSubdocFilter *>(
            reinterpret_cast<SAddSDPar *>(InAddSDPar)->List),
        reinterpret_cast<SAddSDPar *>(InAddSDPar)->InName);
}

char *CQueryBuilderBase::StaticCreateRestrictedTerm(CCatcher *Self, void *InPar)
{
    if (!InPar)
        throw DMemoryException();
    CMemPtr<SRestrictedTermPar> RestrictedTermPar = reinterpret_cast<
            SRestrictedTermPar *>(InPar);
    SRestrictedTermPar *Alias = reinterpret_cast<SRestrictedTermPar *>(InPar);
    return Global->CreateRestrictedTerm(
        reinterpret_cast<CSubdocFilter *>(Alias->In),
        Alias->Out);
}
