#if !defined(__qebb_h)
#define __qebb_h

#include "catcher.h"
#include "options.h"

class CFormulaBase;
class CSubdocFilter;

class CQueryBuilderBase : public CCatcher
{
public:
    static CQueryBuilderBase *Global;

    CQueryBuilderBase(CUnivOptions &Options);
    virtual ~CQueryBuilderBase();

    // returns allocated string on error
    char *Build(const DString &QueryFile,
        CFormulaBase **OutFormula); 
            // owned by ancestor of this, valid only in its scope!

    // see buildqe.h for pointer memory management
    static char *StaticSetSoftOpName(CCatcher *Self, void *InName);
    static char *StaticAddSoftOpArgs(CCatcher *Self, void *InString);
        // string may contain more than 1 word (white-character delimited)
    static char *StaticCreateTerm(CCatcher *Self, void *OutFormula);
    static char *StaticCreateBinary(CCatcher *Self, void *InBinPar);
    static char *StaticSetFinalFormula(CCatcher *Self, void *Formula);
    static char *StaticResetSoftOpName(CCatcher *Self, void *Dummy);
    static char *StaticCreateSubdocList(CCatcher *Self, void *InCreateSDPar);
    static char *StaticAddSubdocName(CCatcher *Self, void *InAddSDPar);
    static char *StaticCreateRestrictedTerm(CCatcher *Self, void *InPar);

protected:
    // see buildqe.h for pointer memory management
    virtual char *SetSoftOpName(char *Name) = 0;
    virtual char *AddSoftOpArg(char *String) = 0;
    virtual char *CreateTerm(CFormulaBase **Out) = 0;
    virtual char *CreateBinary(int Op,
        CFormulaBase *Left,
        CFormulaBase *Right,
        CFormulaBase **Out) = 0;
    virtual char *ResetSoftOpName() = 0;
    virtual char *AddSubdocName(CSubdocFilter *Filter, const char *Name) = 0;
    virtual char *CreateRestrictedTerm(CSubdocFilter *InFilter,
        CFormulaBase **Out) = 0;

private:
    const DString OutErrFormat;
    CFormulaBase *FinalFormula;
};

#endif
