%{

#include "buildqe.h"
#include "pqerr.h"
#include "yaccmac.h"

%}

%union
{  int ival;
   char *sval;
   void *formula;
   SRestrictedDomain *sdlist;
}

%token <sval> NAME
%token <sval> STRING
%type <formula> expr
%type <formula> softop
%type <formula> softop_arg_list
%type <formula> softop_attr
%type <ival> unquoted_arg
%type <sdlist> subdoc_list

%left '|'
%left '&'

%%
formula: expr { CALL_AND_TEST_1(SetFinalFormula, $1) }
   ;

expr: softop_arg_list { $$ = $1; }
   | softop { $$ = $1; }
   | '(' expr ')' { $$ = $2; }
   | expr '&' expr { CALL_AND_TEST_4(CreateBinary, '&', $1, $3, (&($$))) }
   | expr '|' expr { CALL_AND_TEST_4(CreateBinary, '|', $1, $3, (&($$))) }
   ;

softop: softop_name softop_attr { CALL_AND_TEST_0(ResetSoftOpName); $$ = $2; }
   ;

softop_attr: '(' softop_arg_list ')' { $$ = $2; }
   | '[' subdoc_list ']' '(' arg_list ')' { 
       CALL_AND_TEST_2(CreateRestrictedTerm, $2, (&($$))); }
   ;

softop_name: NAME { CALL_AND_TEST_1(SetSoftOpName, $1) }
   ;

softop_arg_list: arg_list { CALL_AND_TEST_1(CreateTerm, (&($$))) }
   ;

arg_list: softop_arg
   | arg_list softop_arg
   ;

subdoc_list: NAME { CALL_AND_TEST_2(CreateSubdocList, (&($$)), $1); }
   | subdoc_list NAME { CALL_AND_TEST_2(AddSubdocName, $1, $2); $$ = $1; }
   ;

softop_arg: string_arg 
   | name_arg
   | unquoted_arg { ErrorEx(AllocUnquotedError($1)); YYERROR; }
   ;

unquoted_arg: '-' { $$ = '-'; }
   | '\'' { $$ = '\''; }
   ;

string_arg: STRING { CALL_AND_TEST_1(AddSoftOpArgs, $1) }
   ;

name_arg: NAME { CALL_AND_TEST_1(AddSoftOpArgs, $1) }
   ;

%%

int yqwrap()
{
    return 1;
}







