#include "myfndt.h"
#include "block.h"
#include "tltree.h"
#include "lockptr.h"
#include "handler.h"

/*#if !defined(_MSC_VER)
#include <alloc.h>
#include <mem.h>
#endif*/
#include <string.h>
#include <stdlib.h>

#if defined(CUSTOM_ITEM_ALLOCATOR)
CChainAllocator CItem::Allocator(sizeof(CItem), 1024);
#endif

CItem::CItem(CKeyHandler *Handler, long aRef, LONG aData, KEY aKey):
	Ref(aRef), Data(aData), Key(NULL)
{
	Key = Handler->NewCopy(aKey);
}

CItem::CItem(CKeyHandler *Handler,
	CFileObj *Stream,
	CItem *PrevItem):
	Ref(NOOFFS), Data(0), Key(NULL)
{
	Stream->Read(&Ref, sizeof(long), 1);
	Stream->Read(&Data, sizeof(LONG), 1);
	Key = Handler->Load(PrevItem == 0 ? 0 : PrevItem->Key, Stream);
         // GCC does not like NULL
}

CItem::~CItem()
{
	CKeyHandler::DelKey(Key);
}

CItem *CItem::Duplicate(CKeyHandler *Handler)
{
	return new CItem(Handler, GetRef(), GetData(), GetKey());
}

long CItem::GetDetachedRef()
{	long r;

	r = Ref;
	Ref = NOOFFS;
	return r;
}

void CItem::AttachRef(long aRef, short Force)
{
	Assert(Force || Ref == NOOFFS);
	Ref = aRef;
}

ushort CItem::GetDiskLen(CKeyHandler *Handler, CItem *PrevItem)
{
	// Borland 5.01 warns without a cast
	return (ushort)(GetDiskHeaderLen() + Handler->GetDiskLen(
			PrevItem == 0 ? 0 : PrevItem->Key, Key));
}

ushort CItem::GetAveDiskLen(CKeyHandler *Handler)
{
	return (ushort)(sizeof(long) + sizeof(LONG) + Handler->GetAveDiskLen());
}

void CItem::Save(CKeyHandler *Handler,
	CFileObj *Stream,
	CItem *PrevItem)
{
	Stream->Write(&Ref, sizeof(long), 1);
	Stream->Write(&Data, sizeof(LONG), 1);
	Handler->Save(PrevItem == 0 ? 0 : PrevItem->Key,
		Key,
		Stream);
}

void CItem::UpdateRef(long Parent, CTree *Tree)
{
	Tree->UpdateNode(Ref, Parent);
}

void CItem::ResetContent()
{
	Key = NULL;
	Data = 0;
}

void CItem::SwapContentWith(CItem *Other)
{
	KEY OtherKey = Other->GetKey();
	LONG OtherData = Other->GetData();
	Other->SetContent(this);
	Key = OtherKey;
	Data = OtherData;
}

CBlockList::CBlockList(long aOffs):
	Offs(aOffs), Next(NULL)
{
}

CBlockList::~CBlockList()
{
	if (Next)
		delete Next;
}

CBlockList *CBlockList::AddBlock(long NewOffs)
{
	Assert(Next == NULL);
	Next = new CBlockList(NewOffs);
	return Next;
}

ushort CBlockList::SaveBlock(CKeyHandler *Handler, ushort Len,
		CItem **Info, ushort Count, CFileObj *Stream)
{	ushort BlockCount, BlockLen;
	long NextOffs, CountOffs;
	CBlockList *NextBlock;

	Assert(Info[0]->GetDiskLen(Handler, NULL) + sizeof(ushort) +
			sizeof(long) <= Len);
	BlockCount = 0;
	if ((NextBlock = GetNext()) == NULL)
		NextOffs = NOOFFS;
	else
		NextOffs = NextBlock->GetOffs();
	CountOffs = Stream->Tell();
	Stream->Write(&BlockCount, sizeof(ushort), 1);
		// just to reserve a place
	Stream->Write(&NextOffs, sizeof(long), 1);
	BlockLen = GetBlockHdrLen();
	CItem *PrevItem = NULL;
	while (BlockCount < Count &&
		(BlockLen += Info[BlockCount]->GetDiskLen(Handler, PrevItem)) <= Len)
	{	Info[BlockCount]->Save(Handler, Stream, PrevItem);
		PrevItem = Info[BlockCount];
		BlockCount++;
	}
	Stream->Seek(CountOffs);
	Stream->Write(&BlockCount, sizeof(ushort), 1);
	return BlockCount;
}

ushort CBlockList::LoadBlock(CKeyHandler *Handler, CItem **Info,
		CFileObj *Stream, long &NextOffs)
{	ushort BlockCount, i;

	Stream->Read(&BlockCount, sizeof(ushort));
	Stream->Read(&NextOffs, sizeof(long));
	CItem *PrevItem = NULL;
	for (i = 0; i < BlockCount; i++)
	{	Info[i] = new CItem(Handler, Stream, PrevItem);
		PrevItem = Info[i];
	}
	return BlockCount;
}

ushort CBlockList::GetBlockCount(CKeyHandler *Handler, ushort Len,
		CItem **Info, ushort Count)
{  ushort BlockCount, BlockLen;

	Assert(Count > 0);
	BlockCount = 0;
	BlockLen = GetBlockHdrLen();
	CItem *PrevItem = NULL;
	while (BlockCount < Count &&
		(BlockLen += Info[BlockCount]->GetDiskLen(Handler, PrevItem)) <= Len)
	{	PrevItem = Info[BlockCount];
		BlockCount++;
	}
	if (BlockCount == 0)
		throw CGeneralException("key too long");
	return BlockCount;
}

void CBlockList::DeleteList()
{
	if (Next)
	{	delete Next;
		Next = NULL;
	}
}

CBlock::CBlock(long aOffs, CTree *aTree, CItem **aInfo, ushort aCount):
	CBlockList(aOffs),
	Tree(aTree), Info(aInfo), Count(aCount)
{
	if (!Info)
		Info = AllocInfo();
}

CBlock::~CBlock()
{
	FreeInfo(Info);
}

void CBlock::SaveItems(ushort Len, ushort ListLen, CFileObj *Stream)
{  ushort BlockCount, BlockSum;
	CBlockList *ListItem;
	long BlockPos;
	CKeyHandler *Handler;

	Handler = Tree->GetKeyHandler();

	// save fixed-len members
	Stream->Write(&Count, sizeof(ushort));
	if (Count == 0) // root can be empty
		return;

	// save current position - DelListOffs (i. e. CTree::DelOffs)
	// and CTree:AddOffs change it
	BlockPos = Stream->Tell();

	// free old linked list
	DelListOffs();
	DeleteList();

	// construct linked list
	BlockCount = GetBlockCount(Handler,
		(ushort)(Len - sizeof(ushort)),
		Info,
		Count);
	BlockSum = BlockCount;
	ListItem = this;
	while (BlockSum < Count)
	{	ListItem = ListItem->AddBlock(Tree->AddOffs());
		BlockCount = ListItem->GetBlockCount(Handler,
			ListLen,
			Info + BlockSum,
			(ushort)(Count - BlockSum));
        		// Borland 5.01 warns without a cast
		BlockSum += BlockCount;
	}

	// save
	Stream->Seek(BlockPos);
	BlockCount = SaveBlock(Tree->GetKeyHandler(),
		(ushort)(Len - sizeof(ushort)),
		Info,
		Count,
		Stream);
	BlockSum = BlockCount;
	ListItem = this;
	while (BlockSum < Count)
	{	ListItem = ListItem->GetNext();
		Stream->Seek(ListItem->GetOffs());
		BlockCount = ListItem->SaveBlock(Tree->GetKeyHandler(),
			ListLen,
			Info + BlockSum,
			(ushort)(Count - BlockSum),
			Stream);
		BlockSum += BlockCount;
	}
}

void CBlock::LoadItems(CFileObj *Stream)
{  ushort BlockCount, BlockSum;
	long NextOffs;
	CBlockList *ListItem;

	// load fixed-len members
	Stream->Read(&Count, sizeof(ushort));
	if (Count == 0) // root can be empty
		return;

	// load
	BlockCount = LoadBlock(Tree->GetKeyHandler(),
		Info,
		Stream,
		NextOffs);
	BlockSum = BlockCount;
	ListItem = this;
	while (BlockSum < Count)
	{	ListItem = ListItem->AddBlock(NextOffs);
		Stream->Seek(NextOffs);
		BlockCount = ListItem->LoadBlock(Tree->GetKeyHandler(),
			Info + BlockSum,
			Stream,
			NextOffs);
		BlockSum += BlockCount;
	}
	Assert(NextOffs == NOOFFS);
}

void CBlock::DelListOffs()
{
	CBlockList *CurItem = GetNext();
		// first block is never freed
	while (CurItem != NULL)
	{	Tree->DelOffs(CurItem->GetOffs());
		CurItem = CurItem->GetNext();
	}
}


