#if !defined(__block_h)
#define __block_h

#include "tltype.h"

#define CUSTOM_ITEM_ALLOCATOR

#if defined(CUSTOM_ITEM_ALLOCATOR)
#include "chaina.h"
#endif

#define NOOFFS (-1l)
#define NOIDX	0xffff

class CItem
{
private:
#if defined(CUSTOM_ITEM_ALLOCATOR)
	static CChainAllocator Allocator;
#endif
	long Ref;
	LONG Data;
	KEY Key;

public:
#if defined(CUSTOM_ITEM_ALLOCATOR)
     void *operator new(size_t Size)
	     { return Allocator.Alloc(Size); }
     void operator delete(void *self)
	     { Allocator.Free(self); }
#endif

	// copies aKey
	CItem(CKeyHandler *Handler, long aRef, LONG aData, KEY aKey);
	CItem(CKeyHandler *Handler, CFileObj *Stream, CItem *PrevItem);
	~CItem();

	CItem *Duplicate(CKeyHandler *Handler);

	inline long GetRef() { return Ref; }
	long GetDetachedRef();
	void AttachRef(long aRef, short Force = 0);
	inline LONG GetData() { return Data; }
	void SetData(LONG aData) { Data = aData; }
	inline const KEY GetKey() { return Key; }
	ushort GetDiskLen(CKeyHandler *Handler, CItem *PrevItem);
	static ushort GetAveDiskLen(CKeyHandler *Handler);
	void Save(CKeyHandler *Handler, CFileObj *Stream, CItem *PrevItem);

	// if the reference is valid, gets corresponding node and
	// changes its parent
	void UpdateRef(long Parent, CTree *Tree);
	void SwapContentWith(CItem *Other);
	void SwapRefWith(CItem *Other)
	     { long TmpRef = Ref; Ref = Other->Ref; Other->Ref = TmpRef; }
	static ushort GetDiskHeaderLen() { return sizeof(long) + sizeof(LONG); }

private: // for SwapContentWith
	inline void SetContent(CItem *OldItem)
		{ Key = OldItem->GetKey(); // N. B.: pointer aliasing!
		  Data = OldItem->GetData(); }
	void ResetContent();
};

class CBlockList
{
private:
	long Offs;
	CBlockList *Next;
public:
	CBlockList(long aOffs);
	virtual ~CBlockList();

	long GetOffs() const { return Offs; }
	static inline ushort GetBlockHdrLen()
		{ return sizeof(long) + sizeof(ushort); }

	// can be called for last item of list only!
	CBlockList *AddBlock(long NewOffs);

	// returns same as GetBlockCount
	ushort SaveBlock(CKeyHandler *Handler, ushort Len, CItem **Info,
			ushort Count, CFileObj *Stream);

	// does not seek! NextOffs is out par
	ushort LoadBlock(CKeyHandler *Handler, CItem **Info, CFileObj *Stream,
			long &NextOffs);

	ushort GetBlockCount(CKeyHandler *Handler, ushort Len, CItem **Info,
			ushort Count);
	inline CBlockList *GetNext() { return Next; }
	void DeleteList();
};

class CBlock : public CBlockList
{
protected:
	CTree *Tree;
	CItem **Info;
	ushort Count;
	CBlock(long aOffs, CTree *aTree, CItem **aInfo, ushort aCount);
public:
	virtual ~CBlock();
	void SaveItems(ushort Len, ushort ListLen, CFileObj *Stream);
protected:
	// for constructors
	void LoadItems(CFileObj *Stream);

private:
	void DelListOffs();
};

#endif
