#if !defined(__book_h)
#define __book_h

#include <stdio.h>

#include "tltype.h"
#include "tltree.h"

class CBook : private CTree
{
private:
	ushort LockKeyCount;

public:
	CBook(CKeyHandler *aKeyHandler,
		const DString &FName, // only path if Opt & IT_TEMP
		ushort Opt);
	virtual ~CBook() { }

	// ret. NULL => Key not in tree
	KEY LockKey(KEY Key)
   	{ return UnivLock(Root->LockKey(Key)); }

	// ret. 0 => tree empty
	KEY LockFirstKey()
		{ return UnivLock(Root->LockFirstKey()); }

	// ret. 0 => tree empty
	KEY LockLastKey()
		{ return UnivLock(Root->LockLastKey()); }

	// does not unlock Key
	// ret. 0 => Key is first (or not in tree)
	KEY LockPrevKey(KEY Key)
		{ return UnivLock(Root->LockPrevKey(Key)); }

	// does not unlock Key
	// ret. 0 => Key is last (or not in tree)
	KEY LockNextKey(KEY Key)
		{ return UnivLock(Root->LockNextKey(Key)); }

	// if Key is in tree, returns it (locked);
	// otherwise returns the biggest key in tree smaller than
	// Key;
	// if there is no such key, returns 0
	KEY LockKeyBefore(KEY Key)
		{ return UnivLock(Root->LockKeyBefore(Key)); }

	// if Key is in tree, returns it (locked);
	// otherwise returns the smallest key in tree bigger than
	// Key;
	// if there is no such key, returns 0
	KEY LockKeyAfter(KEY Key)
		{ return UnivLock(Root->LockKeyAfter(Key)); }

	// 0 OK, 1 key not found, 2 key found but already unlocked
	short UnlockKey(KEY Key);

	// 0 OK, 1 key not found
	short GetData(KEY Key, LONG &Data);

	// 0 OK, 1 key already exists
	// no key may be locked when this function is called!
	short Add(KEY Key, LONG Data);

	// 0 OK, 1 key not found
	// no key may be locked when this function is called!
	short Del(KEY Key);

	// 0 OK, 1 key not found
	short ChangeData(KEY Key, LONG Data)
		{ return Root->ChangeData(Key, Data); }

	// must be called before destructor to update disk image
	void StoreAll() { CTree::StoreAll(); }

	short IsEmpty() { return Root->IsEmpty(); }

#if defined(DEBUG_CNode)
	void Check();
#endif

	unsigned long GetTotal() { return Root->GetTotal(); }

private:
	KEY UnivLock(KEY Key);
};

#endif

