#if !defined(__cache_h)
#define __cache_h

#include "arrays.h"

#include "tltype.h"
#include "remnodes.h"

class CCacheItem
{
public: // used only by CCache, which needs total control
	CSortedNode *Node;
	long PrevOffs;
	long NextOffs;

private:
	CCacheItem(CCacheItem &Other); // no copy constructor
public:
	CCacheItem(CSortedNode *aNode, long aPrevOffs, long aNextOffs);
	virtual ~CCacheItem();
		// deletes node!

	long GetOffs() const { return Node->GetOffs(); }
#if defined(NOCPCONSTRUCT)
	operator == (const CCacheItem *Other);
	operator < (const CCacheItem *Other);
	operator > (const CCacheItem *Other);
#else
	operator == (const CCacheItem &Other) const
		{ return GetOffs() == Other.GetOffs(); }
	operator < (const CCacheItem &Other) const
		{ return GetOffs() < Other.GetOffs(); }
#endif
};

class CCache
{
private:
	static const int Limit;

// members
private:
	DISArray<CCacheItem> CacheArr;
	long FirstOffs;
	long LastOffs;
	CTree *Tree;
public:
	CCache(CTree *aTree, int aLimit = Limit);
	virtual ~CCache() { }

	// can store and discard old item
	// node must not be already cached
	void Add(CSortedNode *Node);

	// returns 0 when no offset exists
	CSortedNode *Yank(long Offs);

	void Store();

	short Has(CSortedNode *Node);
private:
	// returns NOIDX when there is no item
	ushort FindIndex(long Offs);

	// stores node before deleting
	void Del(long Offs);

	// cannot return 0
	CCacheItem *Detach(long Offs, ushort Idx);

	CCacheItem *GetPrev(ushort Idx);
	CCacheItem *GetNext(ushort Idx);
	void SkipPrev(CCacheItem *Item);
	void SkipNext(CCacheItem *Item);
#if defined(DEBUG_CCache)
	void Check();
#endif
};

#endif
