#include "myfndt.h"
#include "hcompstr.h"
#include "numconst.h"

//#define DEBUG_CCompressedStrHandler

const ushort CCompressedStrHandler::PLFlag[3] = { 0x4000,
	0x2000,
	0x1000};

CCompressedStrHandler::CCompressedStrHandler(ushort aAveKeyLen):
	CStringHandler(aAveKeyLen)
{
}

KEY CCompressedStrHandler::Load(KEY PrevKey, CFileObj *Stream)
{	ushort Head;
	ushort Length;
	uchar PatternLength;
#if defined(DEBUG_CCompressedStrHandler)
	ushort CheckLength = 0;
#endif

	char FirstDiff = 0;
	Stream->Read(&Head, sizeof(short));
#if defined(DEBUG_CCompressedStrHandler)
	CheckLength += sizeof(short);
#endif
	if (Head & CONST_SHORT_HIGHEST)
	{	Length = Head & ~CONST_SHORT_HIGHEST;
		Stream->Read(&PatternLength, 1);
#if defined(DEBUG_CCompressedStrHandler)
		CheckLength++;
#endif
	}
	else
	{	Length = (Head & HIGH_BYTE_IN_SHORT & ~GetPLFlags()) >> 8;
		PatternLength = 0;
		char PLDelta = 1;
		for (int i = 0; i < 3; i++)
		{	if (Head & PLFlag[i])
				PatternLength |= PLDelta;
			PLDelta *= 2;
		}
		if (PatternLength == 7)
			PatternLength = Head & LOW_BYTE_IN_SHORT;
		else
		{	FirstDiff = Head & LOW_BYTE_IN_SHORT;
			Assert(FirstDiff || PatternLength == Length);
		}
	}

	KEY Key = CKeyHandler::NewKey(Length + 1);
	Assert(PrevKey != NULL || !PatternLength);
	Assert(PatternLength <= Length);
	char *Ptr = reinterpret_cast<char *>(Key);
	if (PatternLength)
	{	memcpy(Key, PrevKey, PatternLength);
		Ptr += PatternLength;
	}
	ushort ReadLength = Length - PatternLength;
	if (FirstDiff)
	{	Assert(ReadLength > 0);
		ReadLength--;
		*Ptr = FirstDiff;
		Ptr++;
	}
	if (ReadLength)
		Stream->Read(Ptr, ReadLength);
#if defined(DEBUG_CCompressedStrHandler)
	CheckLength += ReadLength;
#endif
	*(reinterpret_cast<char *>(Key) + Length) = 0;
#if defined(DEBUG_CCompressedStrHandler)
	Assert(CheckLength == GetDiskLen(PrevKey, Key));
#endif
	return Key;
}

void CCompressedStrHandler::Save(KEY PrevKey, KEY Key, CFileObj *Stream)
{	ushort Head;
#if defined(DEBUG_CCompressedStrHandler)
	ushort CheckLength = 0;
#endif

	ushort Length = strlen(reinterpret_cast<char *>(Key));
	uchar PatternLength = GetPatternLength(PrevKey, Key);
	Assert(PatternLength <= Length);
	uchar *Ptr = reinterpret_cast<uchar *>(Key) + PatternLength;
		// Ptr cannot be signed char!
	ushort WriteLength = Length - PatternLength;
	if (Length >= 16)
	{  Assert(!(Length & CONST_SHORT_HIGHEST));
		Head = Length | CONST_SHORT_HIGHEST;
		Stream->Write(&Head, sizeof(short));
		Stream->Write(&PatternLength, 1);
#if defined(DEBUG_CCompressedStrHandler)
		CheckLength += sizeof(short) + 1;
#endif
	}
	else
	{	Head = Length << 8;
		if (PatternLength >= 7)
		{	SetPatternFlags(Head, 7);
			Head |= PatternLength;
		}
		else
		{	SetPatternFlags(Head, PatternLength);
			if (WriteLength)
			{
#if defined(DEBUG_CCompressedStrHandler)
				ushort OldHead = Head;
#endif
				Head |= *Ptr;
#if defined(DEBUG_CCompressedStrHandler)
				Assert(Head > OldHead);
#endif
				Ptr++;
				WriteLength--;
			}
		}
		Stream->Write(&Head, sizeof(short));
#if defined(DEBUG_CCompressedStrHandler)
		CheckLength += sizeof(short);
#endif
	}

	if (WriteLength)
		Stream->Write(Ptr, WriteLength);
#if defined(DEBUG_CCompressedStrHandler)
	CheckLength += WriteLength;
	Assert(CheckLength == GetDiskLen(PrevKey, Key));
#endif
}

ushort CCompressedStrHandler::GetDiskLen(KEY PrevKey, KEY Key)
{	ushort DiskLength = 0;

	ushort Length = strlen(reinterpret_cast<char *>(Key));
	uchar PatternLength = GetPatternLength(PrevKey, Key);
	Assert(PatternLength <= Length);
	ushort VarLength = Length - PatternLength;
	DiskLength += sizeof(short);
	if (Length >= 16)
		DiskLength++;
	else
	{	if (PatternLength < 7 && VarLength)
			VarLength--;
	}
	DiskLength += VarLength;
	return DiskLength;
}

uchar CCompressedStrHandler::GetPatternLength(KEY PrevKey, KEY Key)
{
	if (!PrevKey)
		return 0;

	uchar *PrevPtr = reinterpret_cast<uchar *>(PrevKey);
	uchar *Ptr = reinterpret_cast<uchar *>(Key);
	uchar PatternLength = 0;
	while (PatternLength < CONST_UCHAR_MAX)
	{	if (*PrevPtr != *Ptr ||
			!(*PrevPtr) ||
			!(*Ptr))
			return PatternLength;
		PatternLength++;
		PrevPtr++;
		Ptr++;
	}
	Assert(0);
	return 0;
}

void CCompressedStrHandler::SetPatternFlags(ushort &Head,
	uchar PatternLength)
{
	uchar PLDelta = 1;
	for (int i = 0; i < 3; i++)
	{	if (PatternLength & PLDelta)
			Head |= PLFlag[i];
		PLDelta *= 2;
	}
}

