#include "myfndt.h"
#include "node.h"
#include "tltree.h"
#include "lockptr.h"
#include "handler.h"

#include <stdlib.h>
#include <string.h>

// returns 1 exact match; Index can be Count
static short ArrGetIndex(CKeyHandler *Handler, CItem **Info,
		ushort Count, KEY Key, ushort &Index);

// asserts InItem not already in Info, Count < MAXCOUNT
// InItem is inserted or destroyed
// increases Count
static void ArrInsert(CKeyHandler *Handler, CItem **Info, ushort &Count,
		CItem *InItem);

// decreases Count
static CItem *ArrGetFirst(CItem **Info, ushort &Count);

// decreases Count
static CItem *ArrGetLast(CItem **Info, ushort &Count);

// decreases Count
static CItem *ArrGetNth(CItem **Info, ushort &Count, ushort Index);

static void ArrUpdateRefs(CItem **Info, ushort Count, long Offs,
		CTree *Tree);

// inserts first or last item;
// asserts Count < CNode::MaxInfo;
// InItem is inserted or destroyed;
// increases Count
static void ArrInsertExtreme(CItem **Info, ushort &Count, CItem *InItem,
		short From0);

// decreases Count
static void ArrPullFirst(CItem **Info, ushort &Count, CItem **TargetInfo,
		ushort MoveCount);

// decreases Count
static void ArrPullLast(CItem **Info, ushort &Count, CItem **TargetInfo,
		ushort MoveCount);

// increases Count
static void ArrPushFirst(CItem **Info, ushort &Count, CItem **SourceInfo,
		ushort MoveCount);

// increases Count
static void ArrPushLast(CItem **Info, ushort &Count, CItem **SourceInfo,
		ushort MoveCount);

static CItem *ShiftItems(CItem **Info, ushort Count, CItem *Extreme,
		short LeftToRight);

static CItem *ShiftForward(CItem **Info, ushort Count, CItem *Extreme);

static CItem *ShiftBackward(CItem **Info, ushort Count, CItem *Extreme);

static long ShiftRefsToLeft(CItem **Info, ushort Count, long NewExtremeRef);

static long ShiftRefsToRight(CItem **Info, ushort Count, long NewExtremeRef);

const int CNode::MaxInfo = 256; //64; //16; //256; even!

CNode::CNode(long aOffs, long aParent, CTree *aTree, ushort aAveLen):
	CBlock(aOffs, aTree, NULL, 0),
	LastChild(NOOFFS), Parent(aParent),	AveLen(aAveLen), Dirty(1)
{
}

CNode::CNode(long aOffs, long aParent, CTree *aTree, CItem **aInfo,
		ushort aCount, long aLastChild, ushort aAveLen):
	CBlock(aOffs, aTree, aInfo, aCount),
	LastChild(aLastChild), Parent(aParent), AveLen(aAveLen),	Dirty(1)
{
	ArrUpdateRefs(Info, Count, GetOffs(), Tree);
	Tree->UpdateNode(LastChild, GetOffs());
}

CNode::CNode(long aOffs, CTree *aTree, CFileObj *Stream):
	CBlock(aOffs, aTree, NULL, 0),
	LastChild(NOOFFS), Parent(NOOFFS), AveLen(0), Dirty(0)
{
	Stream->Seek(GetOffs());
	Stream->Read(&LastChild, sizeof(long));
	Stream->Read(&Parent, sizeof(long));
	Stream->Read(&AveLen, sizeof(ushort));
	LoadItems(Stream);
}

CNode::~CNode()
{
}

void CNode::SetParent(long ParentOffs)
{
	Dirty = 1;
	Parent = ParentOffs;
}

KEY CNode::LockKey(KEY Key)
{	ushort index;
	short match;

    Assert(Key != NULL);
	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (!match)
		return NULL;

	KEY Out = Node->GetKey(index);
	Node.Drop();
	return Out;
}

KEY CNode::LockFirstKey()
{
	CLockedPtr Node1(Tree->GetLockedNode(GetOffs()), Tree);
		// locks this node
	while (1)
	{	CLockedPtr Node2(Node1->GetFirstChild(), Tree);
		if (Node2 == NULL)
		{	KEY Key = Node1->GetFirstKeyLoc();
			if (Key)
				Node1.Drop(); // Node1 stays locked after return
			return Key;
		}
		else
			Node1 = Node2.Drop();
	}
}

KEY CNode::LockLastKey()
{
	CLockedPtr Node1(Tree->GetLockedNode(GetOffs()), Tree);
		// locks this node
	while (1)
	{	CLockedPtr Node2(Node1->GetLastChild(), Tree);
		if (Node2 == NULL)
		{	KEY Key = Node1->GetLastKeyLoc();
			if (Key)
				Node1.Drop(); // Node1 stays locked after return
			return Key;
		}
		else
			Node1 = Node2.Drop();
	}
}

KEY CNode::LockPrevKey(KEY Key)
{	ushort index;
	short match;

	Assert(Key != NULL);
	CLockedPtr Node1(GetSubNode(Key, index, match), Tree);
	if (!match)
		return NULL;

	CLockedPtr Child(Node1->GetChild(index), Tree);
	if (Child)
		return Child->LockLastKey();

	while (Node1->IsLeftmost(index))
	{	CLockedPtr Node2(Node1->GetParent(), Tree);
		if (Node2 == NULL)
			 return NULL;
		index = Node2->GetChildIndex(Node1);
		Node1 = Node2.Drop();
	}

	index--;
	KEY NewKey = Node1->GetKey(index);
	Node1.Drop();
	return NewKey;
}

KEY CNode::LockNextKey(KEY Key)
{	ushort index;
	short match;

	Assert(Key != NULL);
	CLockedPtr Node1(GetSubNode(Key, index, match), Tree);
	if (!match)
		return NULL;

	Assert(!Node1->IsRightmost(index));
	index++;
	CLockedPtr Child(Node1->GetChild(index), Tree);
	if (Child)
		return Child->LockFirstKey();

	while (Node1->IsRightmost(index))
	{	CLockedPtr Node2(Node1->GetParent(), Tree);
		if (!Node2)
			return NULL;
		index = Node2->GetChildIndex(Node1);
		Node1 = Node2.Drop();
	}

	KEY NewKey = Node1->GetKey(index);
	Node1.Drop();
	return NewKey;
}

KEY CNode::LockKeyBefore(KEY Key)
{	ushort index;
	short match;

	Assert(Key != NULL);
	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (match)
	{	KEY NewKey = Node->GetKey(index);
		Node.Drop();
		return NewKey;
	}
	if (Node->IsEmpty())
		return NULL;

	if (!Node->IsLeftmost(index))
	{	KEY NewKey = Node->GetKey((ushort)(index - 1));
		Node.Drop();
		return NewKey;
	}

	return LockPrevKey(Node->GetKey(0));
}

KEY CNode::LockKeyAfter(KEY Key)
{	ushort index;
	short match;

	Assert(Key != NULL);
	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (match)
	{	KEY NewKey = Node->GetKey(index);
		Node.Drop();
		return NewKey;
	}
	if (Node->IsEmpty())
		return NULL;

	if (!Node->IsRightmost(index))
	{	KEY NewKey = Node->GetKey(index);
		Node.Drop();
		return NewKey;
	}

	KEY LastKey = Node->GetLastKeyLoc();
	Assert(LastKey != NULL);
	return LockNextKey(LastKey);
}

short CNode::UnlockKey(KEY Key)
{	ushort index;
	short match;

	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
		// now Node should be locked at least two times
	if (!match)
		return 1;
	if (Tree->UnlockNode(Node))
		/* UnlockNode() will be called again in
			CLockedPtr::~CLockedPtr, but this is harmless since
			DecLockCount will return 0 */
		return 2;
	return 0;
}

short CNode::GetData(KEY Key, LONG &Data)
{	ushort index;
	short match, rv;

#if defined(DEBUG_cut_exception)
    printf("CNode::GetData\n");
#endif
	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (!match)
		rv = 1;
	else
	{  rv = 0;
		Data = Node->GetDataLoc(index);
	}
	return rv;
}

short CNode::Add(KEY Key, LONG Data)
{	CItem *FreeItem;
	ushort index;
	short match;

	CLockedPtr Node1(GetSubNode(Key, index, match), Tree);
	if (match)
		return 1;
	FreeItem = new CItem(Tree->GetKeyHandler(), NOOFFS, Data, Key);
	while (1)
	{	CObjPtr<CItem> Item = Node1->AddStep(FreeItem);
		if (Item == NULL)
			return 0;
		CLockedPtr Node2(Node1->GetParent(), Tree);
		FreeItem = Item.Drop();
		if (Node2 == NULL)
		{	Tree->AddLevel(FreeItem); // consumes FreeItem
			return 0;
		}
		else
			Node1 = Node2.Drop(); // FreeItem is consumed in the next iteration
	}
}

short CNode::Del(KEY Key)
{	ushort index;
	short match;
	CNode *FreeNode;

	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (!match)
		return 1;

	FreeNode = Node->DelDown(index);
	if (FreeNode)
		// index is updated
		Node = FreeNode;

	CLockedPtr NewRoot(NULL, Tree);
	while (!Node->IsRoot())
	{	Node->DelLoc(index);
		if (!Node->IsThin())
			return 0;
		CLockedPtr Parent(Node->GetParent(), Tree);
		Assert(Parent != NULL);
		if ((index = Parent->BeefUp(Node)) == NOIDX)
			return 0;
		NewRoot = Node.Drop();
			// NewRoot is used only when Parent->IsRoot()
		Node = Parent.Drop();
	}

	Node->DelLoc(index);

	// unlock everything before trying to destroy old root
	Tree->UnlockNode(Node.Drop());
	if (NewRoot)
		Tree->CheckDelLevel(NewRoot.Drop());
	return 0;
}

short CNode::ChangeData(KEY Key, LONG Data)
{	ushort index;
	short match;

	CLockedPtr Node(GetSubNode(Key, index, match), Tree);
	if (!match)
		return 1;

	Node->ChangeDataLoc(index, Data);
	return 0;
}

void CNode::Save()
{	ushort blen;
	CFileObj *Stream;

	blen = GetBlockLen();
	Assert(blen > GetNodeHdrLen());
	Stream = Tree->GetStream();
	Stream->Seek(GetOffs());
	Stream->Write(&LastChild, sizeof(long));
	Stream->Write(&Parent, sizeof(long));
	Stream->Write(&AveLen, sizeof(ushort));
	SaveItems((ushort)(blen - GetNodeHdrLen()), blen, Stream);
	Dirty = 0;
}

CItem *CNode::AddStep(CItem *InItem)
{	CItem *RawItem;
	long offs, ref;

	Dirty = 1;
	CObjPtr<CItem> Item = InItem;
	CKeyHandler *Handler = Tree->GetKeyHandler();

	if (Count < MaxInfo)
	{	Item->UpdateRef(GetOffs(), Tree);
			// this node is parent of Item->Ref
		ArrInsert(Handler, Info, Count, Item.Drop());
		return NULL;
	}

	Assert(Count % 2 == 0);
	CInfoPtr NewInfo(Split());
	KEY FirstKey = GetFirstKeyLoc();
	Assert(FirstKey != NULL);
	if (Handler->Compare(Item->GetKey(), FirstKey) < 0)
		// smaller than first item in array
	{	// it's useless to update Item's reference; whole
		// NewInfo will be updated
		ArrInsert(Handler, NewInfo, Count, Item.Drop());
		RawItem = ArrGetLast(NewInfo, Count);
	}
	else
	{	// Possible optimization: Item's reference needs to be
		// updated now, but only if Item stays in this node, i. e.
		// Item != NewItem.
		Item->UpdateRef(GetOffs(), Tree);

		ArrInsert(Handler, Info, Count, Item.Drop());
		RawItem = ArrGetFirst(Info, Count);
	}
	CObjPtr<CItem> NewItem = RawItem;

	offs = Tree->AddOffs();
	ref = NewItem->GetDetachedRef();
	CObjPtr<CNode> NewNode = CNodeFactory::Create(offs,
		NOOFFS, // parent is unknown yet
		Tree,
		NewInfo, // all references will be updated
		Count,
		ref, // will be updated
		AveLen);
	NewInfo.Drop();
	NewItem->AttachRef(offs);
	Tree->Remember(NewNode.Drop());
	return NewItem.Drop();
}

CItem **CNode::Split()
{	CItem **IPtr, **NewInfo, **SwapInfo;

	Dirty = 1;
	Assert(Count % 2 == 0 && Info != NULL);
	Count /= (ushort)2;
	NewInfo = AllocInfo();
	IPtr = Info + Count;
	memcpy(NewInfo, IPtr, Count * sizeof(CItem *));
	memset(IPtr, 0, Count * sizeof(CItem *));
	SwapInfo = Info; Info = NewInfo; NewInfo = SwapInfo;
	return NewInfo;
}

CNode *CNode::DelDown(ushort &Index)
{	ushort NewIndex = 0;

	CLockedPtr Node1(GetChild(Index), Tree);
	if (!Node1)
		return NULL;

	while (1)
	{	CLockedPtr Node2(Node1->GetLastChild(), Tree);
		if (Node2 == NULL)
			break;
		else
			Node1 = Node2.Drop();
	}

	CItem *ReplItem = Node1->GetLastPtr(NewIndex);
#if !defined(FAST_SWAP)
	SwapItemContent(Info[Index], ReplItem, Node1);
#else
	Info[Index]->SwapRefWith(ReplItem);
	Node1->SetLastPtr(Info[Index]);
	Info[Index] = ReplItem; 
	Dirty = 1;
#endif
	Index = NewIndex;
	return Node1.Drop();
}

ushort CNode::BeefUp(CNode *ThinNode)
{	ushort DoomedIndex;

	ushort ThinIndex = GetChildIndex(ThinNode);
	Assert(ThinIndex != NOIDX);
	ushort NeighborIndex = ThinIndex;
	CLockedPtr Neighbor(GetNeighbor(NeighborIndex), Tree);
	Assert(Neighbor &&
		NeighborIndex != NOIDX &&
		NeighborIndex != ThinIndex);

	if (NeighborIndex < ThinIndex) // LeftToRight
		DoomedIndex = BeefUpForward(NeighborIndex, ThinNode, Neighbor);
	else
		DoomedIndex = BeefUpBackward(ThinIndex, ThinNode, Neighbor,
				NeighborIndex);

	if (DoomedIndex != NOIDX)
	{	Dirty = 1; // ???
		Tree->DestroyNode(Neighbor.Drop());
	}
	return DoomedIndex;
}

ushort CNode::BeefUpForward(ushort PivotIndex, CNode *ThinNode,
		CNode *Neighbor)
{
	SlantForward();
	ThinNode->SlantForward();
	Neighbor->SlantForward();

	short DelEndangered = BeefUpSlanted(1,
		PivotIndex, // NeighborIndex
		ThinNode,
		Neighbor);

	Neighbor->SlantBackward();
	ThinNode->SlantBackward();
	SlantBackward();

	return (short)(DelEndangered ? PivotIndex : NOIDX);
   	// NOIDX shoul not be defined as int
}

ushort CNode::BeefUpBackward(ushort PivotIndex, CNode *ThinNode,
		CNode *Neighbor, ushort NeighborIndex)
{
	short DelEndangered = BeefUpSlanted(0,
		PivotIndex, // ThinIndex
		ThinNode,
		Neighbor);
	if (!DelEndangered)
		return NOIDX;

	Dirty = 1;
	long Orphan = Info[PivotIndex]->GetDetachedRef();
		// Item which refers to this subtree will be deleted
	SetChildRef(NeighborIndex, Orphan);
		// referred to neighbor
	return PivotIndex;
}

short CNode::BeefUpSlanted(short LeftToRight, ushort PivotIndex,
		CNode *ToNode, CNode *FromNode)
{
	Assert(ToNode->Count < /* MaxInfo / 2 <= */ FromNode->Count);
	if (ToNode->Count + FromNode->Count >= MaxInfo)
	{	MoveItems(LeftToRight,
			PivotIndex,
			ToNode,
			FromNode,
			(ushort)((FromNode->Count - ToNode->Count) / 2));
				// minimally (((MaxInfo / 2) + 1) - ((MaxInfo / 2) - 1) / 2 ==
				// 2 / 2 == 1
		return 0;
	}

	MoveItems(LeftToRight,
		PivotIndex,
		ToNode,
		FromNode,
		FromNode->Count);
	CObjPtr<CItem> Item = DuplNthItem(PivotIndex);
	Item->AttachRef(FromNode->GetExtremeChild(), 1);
		// old ref is FromNode
	ToNode->InsertExtreme(Item.Drop(), LeftToRight);
		// updates ref
	return 1;
}

void CNode::MoveItems(short LeftToRight, ushort PivotIndex,
		CNode *ToNode, CNode *FromNode, ushort MoveCount)
{
	Assert(0 < MoveCount && MoveCount <= FromNode->Count);
	Dirty = 1; // pivot changes
	CInfoPtr MoveInfo = AllocInfo();
	FromNode->PullItems(MoveInfo, MoveCount, !LeftToRight);

	Info[PivotIndex]->AttachRef(ToNode->GetExtremeChild(), 1);
		// ref is ToNode
	Info[PivotIndex] = ShiftItems(MoveInfo,
		MoveCount,
		Info[PivotIndex],
		!LeftToRight);
	long NewExtreme = Info[PivotIndex]->GetDetachedRef();
	ToNode->SetExtremeChild(NewExtreme);

	Info[PivotIndex]->AttachRef(ToNode->GetOffs());
	Info[PivotIndex]->UpdateRef(GetOffs(), Tree);

	ArrUpdateRefs(MoveInfo, MoveCount, ToNode->GetOffs(), Tree);
	ToNode->PushItems(MoveInfo, MoveCount, LeftToRight);
}

void CNode::DelLoc(ushort Index)
{
	Dirty = 1;
	CItem *Item = ArrGetNth(Info, Count, Index);
	delete Item;
}

CNode *CNode::GetSubNode(KEY Key, ushort &Index, short &Match)
{	CNode *RawNode;

	CLockedPtr Node(Tree->GetLockedNode(GetOffs()), Tree);
		// locks this node
	while (1)
	{	if ((Match = Node->GetIndex(Key, Index)) != 0)
			return Node.Drop();
		if ((RawNode = Node->GetChild(Index)) == NULL)
			return Node.Drop();
		Assert(RawNode->Parent == Node->GetOffs());
		Node = RawNode;
	}
}

CNode *CNode::GetParent()
{
	if (Parent == NOOFFS)
		return NULL;
	else
		return Tree->GetLockedNode(Parent);
}

CNode *CNode::GetChild(ushort Index)
{	long ref;

	Assert(Index <= MaxInfo);
	ref = GetChildRef(Index);
	if (ref == NOOFFS)
		return NULL;
	else
		return Tree->GetLockedNode(ref);
}

short CNode::GetIndex(KEY Key, ushort &Index)
{
	return ArrGetIndex(Tree->GetKeyHandler(),
		Info,
		Count,
		Key,
		Index);
}

LONG CNode::GetDataLoc(ushort Index)
{
	Assert(Index < MaxInfo);
	return (Info[Index])->GetData();
}

void CNode::ChangeDataLoc(ushort Index, LONG Data)
{
	Dirty = 1;
	Assert(Index < MaxInfo);
	Info[Index]->SetData(Data);
}

ushort CNode::GetBlockLen()
{
	return (ushort)(GetBlockHdrLen() +
		GetNodeHdrLen() +
		MaxInfo * CItem::GetAveDiskLen(Tree->GetKeyHandler()));
}

CNode *CNode::GetFirstChild()
{
	if (Count == 0)
		return NULL;
	return GetChild(0);
}

KEY CNode::GetFirstKeyLoc()
{
	if (!Count)
		return NULL;
	return Info[0]->GetKey();
}

CNode *CNode::GetLastChild()
{
	if (Count == 0)
		return NULL;
	return GetChild(Count);
}

KEY CNode::GetLastKeyLoc()
{
	if (!Count)
		return NULL;
	return Info[Count - 1]->GetKey();
}

short CNode::IsThin()
{
	Assert(!IsRoot());
	return Count < MaxInfo / 2;
		// this is invalid status which can arise only temporarily
		// during key deletion
}

CItem *CNode::DuplNthItem(ushort Index)
{
	Assert(Index < Count);
	return Info[Index]->Duplicate(Tree->GetKeyHandler());
}

CItem *CNode::DuplLastItem(ushort &Index)
{
	Index = (ushort)(Count - 1);
	return Info[Index]->Duplicate(Tree->GetKeyHandler());
}

/*void CNode::ReplaceLoc(ushort Index, CItem *InItem)
{
	Assert(Index < Count);
	Dirty = 1;
	delete Info[Index];
	Info[Index] = InItem;
}*/

ushort CNode::GetChildIndex(CNode *Child)
{
	long ChildOffs = Child->GetOffs();
	for (ushort i = 0; i <= Count; i++)
	{	if (ChildOffs == GetChildRef(i))
			return i;
	}
	return NOIDX;
}

CNode *CNode::GetNeighbor(ushort &Index)
{	CNode *LooseNode;

	Assert(Count >= 1 && Index <= Count);
	if (Index > 0)
	{	LooseNode = GetChild((ushort)(Index - 1));
		Assert(LooseNode != NULL);
	}
	else
		LooseNode = NULL;
	CLockedPtr LeftNeighbor(LooseNode, Tree);

	if (Index < Count)
	{	LooseNode = GetChild((ushort)(Index + 1));
		Assert(LooseNode != NULL);
	}
	else
		LooseNode = NULL;
	CLockedPtr RightNeighbor(LooseNode, Tree);

	if (!LeftNeighbor)
	{	Assert(RightNeighbor != NULL);
		Index++;
		return RightNeighbor.Drop();
	}
	else if (!RightNeighbor)
	{	Index--;
		return LeftNeighbor.Drop();
	}

	if (LeftNeighbor->Count >= RightNeighbor->Count)
	{	Index--;
		return LeftNeighbor.Drop();
	}
	else
	{	Index++;
		return RightNeighbor.Drop();
	}
}

void CNode::PullItems(CItem **TargetInfo, ushort MoveCount,
		short From0)
{
	Assert(MoveCount <= Count);
	Dirty = 1;
	if (From0)
		ArrPullFirst(Info, Count, TargetInfo, MoveCount);
	else
		ArrPullLast(Info, Count, TargetInfo, MoveCount);
}

void CNode::PushItems(CItem **SourceInfo, ushort MoveCount,
		short From0)
{
	Assert(MoveCount + Count <= MaxInfo);
	Dirty = 1;
	if (From0)
		ArrPushFirst(Info, Count, SourceInfo, MoveCount);
	else
		ArrPushLast(Info, Count, SourceInfo, MoveCount);
}

void CNode::InsertExtreme(CItem *InItem, short From0)
{
	Dirty = 1;
	CObjPtr<CItem> Item = InItem;

	long Ref = Item->GetDetachedRef();
	Item->AttachRef(GetExtremeChild());
	Tree->UpdateNode(Ref, GetOffs());
	SetExtremeChild(Ref);

	ArrInsertExtreme(Info, Count, Item.Drop(), From0);
}

CItem *CNode::GetLastPtr(ushort &Index)
{
	Assert(Count > 0);
	Index = (ushort)(Count - 1);
	return Info[Index];
}

#if !defined(FAST_SWAP)
void CNode::SwapItemContent(CItem *Item1, CItem *Item2,
		CNode *OtherNode)
{
	Item1->SwapContentWith(Item2);
	Dirty = OtherNode->Dirty = 1;
}
#else
void CNode::SetLastPtr(CItem *InItem)
{
	Assert(Count > 0);
	Info[Count - 1] = InItem;
	Dirty = 1;
}
#endif

long CNode::GetChildRef(ushort Index)
{
	Assert(Index <= Count);
	return Index == Count ? LastChild : (Info[Index])->GetRef();
}

void CNode::SetChildRef(ushort Index, long Ref)
{
	Assert(Index <= Count);
	Dirty = 1;
	if (Index < Count)
		Info[Index]->AttachRef(Ref, 1);
	else
		LastChild = Ref;
}

void CNode::SlantForward()
{
	if (LastChild == NOOFFS) //  => all refs are the same
		return;

	// does not set dirty; slanted node is inconsistent and
	// cannot be saved anyway
	LastChild = ShiftRefsToLeft(Info, Count, LastChild);
}

void CNode::SlantBackward()
{
	if (LastChild == NOOFFS || //  => all refs are the same
		!Count) // BeefUpSlanted moved all items out
		return;

	LastChild = ShiftRefsToRight(Info, Count, LastChild);
}

void CNode::SetExtremeChild(long ExtremeChild)
{
	Dirty = 1;
	Tree->UpdateNode(ExtremeChild, GetOffs());
	LastChild = ExtremeChild;
}

#if defined (DEBUG_CNode)
void CNode::Check(ushort Depth, ushort *Path)
{
	for (ushort i = 0; i <= Count; i++)
	{  Path[Depth] = i;
		CLockedPtr Child(GetChild(i), Tree);
		if (Child)
		{	Assert(Child->Parent == GetOffs());
			Assert(Child->Count);
			if (i < Count)
				CheckOrder(Child, i);
			else
			{	if (Depth)
					CheckLastChildOrder(Child);
			}
			Child->Check(Depth + 1, Path);
		}
	}
}

void CNode::CheckOrder(CNode *Child, ushort ChildIndex)
{
	CKeyHandler *Handler = Tree->GetKeyHandler();
	KEY HighKey = GetKey(ChildIndex);
	KEY LowKey = Child->GetLastKeyLoc();
	Assert(LowKey != NULL);
	int c = Handler->Compare(HighKey, LowKey);
	Assert(c > 0);
}

void CNode::CheckLastChildOrder(CNode *Child)
{	ushort Index;

	CKeyHandler *Handler = Tree->GetKeyHandler();

	CLockedPtr BelowAncestor(Tree->GetLockedNode(GetOffs()), Tree);
	Assert(BelowAncestor != NULL);
	CLockedPtr Ancestor(BelowAncestor->GetParent(), Tree);
	Assert(Ancestor != NULL);
	Index = Ancestor->GetChildIndex(BelowAncestor);
	while (Index == Ancestor->Count)
	{	BelowAncestor = Ancestor.Drop();
		Ancestor = BelowAncestor->GetParent();
		if (!Ancestor)
			return;
		Index = Ancestor->GetChildIndex(BelowAncestor);
	}
	KEY HighKey = Ancestor->GetKey(Index);

	KEY LowKey = Child->GetLastKeyLoc();
	Assert(LowKey != NULL);

	int c = Handler->Compare(HighKey, LowKey);
	Assert(c > 0);
}
#endif

unsigned long CNode::GetTotal()
{
	unsigned long Total = Count;
	for (ushort i = 0; i <= Count; i++)
	{	CLockedPtr Child(GetChild(i), Tree);
		if (Child)
			Total += Child->GetTotal();
	}
	return Total;
}

static short ArrGetIndex(CKeyHandler *Handler, CItem **Info,
		ushort Count, KEY Key, ushort &Index)
{	ushort min, max, ave;
	short t;

	if (Count == 0)
	{	Index = 0;
		return 0;
	}
	min = 0;
	max = Count;
	while (min + 1 < max)
	{	ave = (ushort)((min + max) / 2);
		if (Handler->Compare(Key, Info[ave]->GetKey()) < 0)
			max = ave;
		else
			min = ave;
	}
	if ((t = Handler->Compare(Key, Info[min]->GetKey())) <= 0)
	{	Index = min;
		return (short)(t == 0 ? 1 : 0);
	}
	else
	{	Index = max;
		return 0;
	}
}

static void ArrInsert(CKeyHandler *Handler, CItem **Info, ushort &Count,
		CItem *InItem)
{  ushort Index;
	CItem **IPtr;

	Assert(Count < CNode::MaxInfo);
	CObjPtr<CItem> Item = InItem;
#if !defined(NDEBUG)
	short match =
#endif
		ArrGetIndex(Handler, Info, Count, Item->GetKey(), Index);
#if !defined(NDEBUG)
	Assert(!match);
#endif
	IPtr = Info + Index;
	if (Index < Count)
		memmove(IPtr + 1, IPtr, (Count - Index) * sizeof(CItem *));
	*IPtr = Item.Drop();
	Count++;
}

static CItem *ArrGetFirst(CItem **Info, ushort &Count)
{	CItem *RetItem;

	Assert(Count > 0);
	RetItem = Info[0];
	memmove(Info, Info + 1, (Count - 1) * sizeof(CItem *));
	Info[Count - 1] = NULL; // necessary for FreeInfo!
	Count--;
	return RetItem;
}

static CItem *ArrGetLast(CItem **Info, ushort &Count)
{	CItem *RetItem;

	Assert(Count > 0);
	RetItem = Info[Count - 1];
	Info[Count - 1] = NULL; // necessary for FreeInfo!
	Count--;
	return RetItem;
}

static CItem *ArrGetNth(CItem **Info, ushort &Count, ushort Index)
{
	Assert(Count > 0 && Index < Count);
	CItem *RetItem = Info[Index];

	CItem **ShiftedInfo = Info + Index;
	ushort ShiftedCount = (ushort)(Count - Index);
	memmove(ShiftedInfo, ShiftedInfo + 1, (ShiftedCount - 1) *
			sizeof(CItem *));

	Info[Count - 1] = NULL; // necessary for FreeInfo!
	Count--;
	return RetItem;
}

static void ArrUpdateRefs(CItem **Info, ushort Count, long Offs,
		CTree *Tree)
{	ushort i;

	for (i = 0; i < Count; i++)
		Info[i]->UpdateRef(Offs, Tree);
}

/*static CItem *ArrShift(CItem **Info, ushort Count, CItem *NewExtreme,
		short LeftToRight)
{	CItem *LooseItem;

	if (LeftToRight)
		LooseItem = ArrGetLast(Info, Count);
	else
		LooseItem = ArrGetFirst(Info, Count);
	ArrInsertExtreme(Info, Count, NewExtreme, LeftToRight);
	return LooseItem;
}*/

static void ArrInsertExtreme(CItem **Info, ushort &Count, CItem *InItem,
		short From0)
{
	Assert(Count < CNode::MaxInfo);
	if (From0)
	{	memmove(Info + 1, Info, Count * sizeof(CItem *));
		Info[0] = InItem;
	}
	else
		Info[Count] = InItem;
	Count++;
}

static void ArrPullFirst(CItem **Info, ushort &Count, CItem **TargetInfo,
		ushort MoveCount)
{
    Assert(TargetInfo != NULL && Info != NULL);
	memcpy(TargetInfo, Info, MoveCount * sizeof(CItem *));
	memmove(Info, Info + MoveCount, (Count - MoveCount) * sizeof(CItem *));
	memset(Info + Count - MoveCount, 0, MoveCount * sizeof(CItem *));
	Count -= MoveCount;
}

static void ArrPullLast(CItem **Info, ushort &Count, CItem **TargetInfo,
		ushort MoveCount)
{
    Assert(TargetInfo != NULL && Info != NULL);
	memcpy(TargetInfo, Info + Count - MoveCount, MoveCount * sizeof(CItem *));
	memset(Info + Count - MoveCount, 0, MoveCount * sizeof(CItem *));
	Count -= MoveCount;
}

static void ArrPushFirst(CItem **Info, ushort &Count, CItem **SourceInfo,
		ushort MoveCount)
{
    Assert(Info != NULL && SourceInfo != NULL);
	memmove(Info + MoveCount, Info, Count * sizeof(CItem *));
	memcpy(Info, SourceInfo, MoveCount * sizeof(CItem *));
	Count += MoveCount;
	memset(SourceInfo, 0, MoveCount * sizeof(CItem *));
}

static void ArrPushLast(CItem **Info, ushort &Count, CItem **SourceInfo,
		ushort MoveCount)
{
    Assert(Info != NULL && SourceInfo != NULL);
	memcpy(Info + Count, SourceInfo, MoveCount * sizeof(CItem *));
	Count += MoveCount;
	memset(SourceInfo, 0, MoveCount * sizeof(CItem *));
}

static CItem *ShiftItems(CItem **Info, ushort Count, CItem *Extreme,
		short LeftToRight)
{
	Assert(Count > 0);
	if (LeftToRight)
		return ShiftForward(Info, Count, Extreme);
	else
		return ShiftBackward(Info, Count, Extreme);
}

static CItem *ShiftForward(CItem **Info, ushort Count, CItem *Extreme)
{
	CItem *NewExtreme = Info[Count - 1];
	memmove(Info + 1, Info, (Count - 1) * sizeof(CItem *));
	Info[0] = Extreme;
	return NewExtreme;
}

static CItem *ShiftBackward(CItem **Info, ushort Count, CItem *Extreme)
{
	CItem *NewExtreme = Info[0];
	memmove(Info, Info + 1, (Count - 1) * sizeof(CItem *));
	Info[Count - 1] = Extreme;
	return NewExtreme;
}

static long ShiftRefsToLeft(CItem **Info, ushort Count, long NewExtremeRef)
{	long Ref;

	Assert(Count > 0);
	ushort StartIndex = 0;
	ushort StopIndex = (ushort)(Count - 1);
	long OldExtremeRef = Info[StartIndex]->GetDetachedRef();
	for (ushort i = StartIndex; i < StopIndex; i++)
	{  Ref = Info[i + 1]->GetDetachedRef();
		Info[i]->AttachRef(Ref);
	}
	Info[StopIndex]->AttachRef(NewExtremeRef);
	return OldExtremeRef;
}

static long ShiftRefsToRight(CItem **Info, ushort Count, long NewExtremeRef)
{	long Ref;

	Assert(Count > 0);
	ushort StartIndex = (ushort)(Count - 1);
	ushort StopIndex = 0;
	long OldExtremeRef = Info[StartIndex]->GetDetachedRef();
	for (ushort i = StartIndex; i > StopIndex; i--)
	{  Ref = Info[i - 1]->GetDetachedRef();
		Info[i]->AttachRef(Ref);
	}
	Info[StopIndex]->AttachRef(NewExtremeRef);
	return OldExtremeRef;
}


