#if !defined(__node_h)
#define __node_h

#define FAST_SWAP
//#define DEBUG_CNode

#include "tltype.h"
#include "block.h"

class CNode : public CBlock
{
public:
	static const int MaxInfo;

private:
	long LastChild;
	long Parent;
	ushort AveLen;
	short Dirty;

protected:
	// empty node
	CNode(long aOffs, long aParent, CTree *aTree, ushort aAveLen);

	// filled node; automatically updates references!
	CNode(long aOffs, long aParent, CTree *aTree, CItem **aInfo,
			ushort aCount, long aLastChild, ushort aAveLen);

	// node saved in file
	CNode(long aOffs, CTree *aTree, CFileObj *Stream);
public:
	virtual ~CNode();

	inline short IsDirty() { return Dirty; }
	inline ushort GetAveLen() { return AveLen; }
	inline short IsRoot() { return Parent == NOOFFS; }
		// this is *not* the only function used for testing!
	inline short IsEmpty() { return Count == 0; }

	void SetParent(long ParentOffs);

	// ret. NULL => Key not under node
	KEY LockKey(KEY Key);

	// ret. 0 => node empty
	KEY LockFirstKey();

	// ret. 0 => node empty
	KEY LockLastKey();

	// does not unlock Key
	// ret. 0 => Key is first (or not in tree)
	KEY LockPrevKey(KEY Key);

	// does not unlock Key
	// ret. 0 => Key is last (or not in tree)
	KEY LockNextKey(KEY Key);

	// if Key is in tree, returns it (locked);
	// otherwise returns the biggest key in tree smaller than
	// Key;
	// if there is no such key, returns 0
	KEY LockKeyBefore(KEY Key);

	// if Key is in tree, returns it (locked);
	// otherwise returns the smallest key in tree bigger than
	// Key;
	// if there is no such key, returns 0
	KEY LockKeyAfter(KEY Key);

	// 0 OK, 1 key not found, 2 key found but already unlocked
	short UnlockKey(KEY Key);

	// 0 OK, 1 key not found
	short GetData(KEY Key, LONG &Data);

	// 0 OK, 1 key already exists
	short Add(KEY Key, LONG Data);

	// 0 OK, 1 key not found
	short Del(KEY Key);

	// 0 OK, 1 key not found
	short ChangeData(KEY Key, LONG Data);

	void Save();
	ushort GetBlockLen();

	// fixed length fields at the beginning of block
	inline static ushort GetNodeHdrLen()
		{ return sizeof(long) + sizeof(long) + sizeof(ushort); }

#if defined (DEBUG_CNode)
	void Check(ushort Depth, ushort *Path);
private:
	void CheckOrder(CNode *Child, ushort ChildIndex);
	void CheckLastChildOrder(CNode *Child);
#endif

public:
	unsigned long GetTotal();

private:
	// InItem's key must not exist;
	// InItem is inserted or destroyed
	// returns item => that item must be added to the parent
	CItem *AddStep(CItem *InItem);

	// asserts Count even
	// returns first half of Info
	CItem **Split();

	// if index'th key does not have subkeys, returns NULL
	// (does not change Index, does not delete key);
	// otherwise locks the leftmost leaf, swaps content of
	// its last key with index'th key and returns the leaf
	// (then, Index is out par. specifying the key to be deleted)
	CNode *DelDown(ushort &Index);

	// copies some keys to ThinNode from its neigbor
	// updates this node
	// returns index of item to be deleted or NOIDX
	ushort BeefUp(CNode *ThinNode);

	ushort BeefUpForward(ushort PivotIndex, CNode *ThinNode,
			CNode *Neighbor);
	ushort BeefUpBackward(ushort PivotIndex, CNode *ThinNode,
			CNode *Neighbor, ushort NeighborIndex);

	// deletes index'th key
	// subtree of deleted key is ignored (can be called even if
	// the reference is invalid)
	void DelLoc(ushort Index);

	// cannot return 0; returned node is locked!
	CNode *GetSubNode(KEY Key, ushort &Index, short &Match);

	// can return 0; returned node is locked!
	CNode *GetParent();

	// can return 0; returned node is locked!
	CNode *GetChild(ushort Index);

	// returns 1 exact match; Index can be Count
	short GetIndex(KEY Key, ushort &Index);

	LONG GetDataLoc(ushort Index);

	void ChangeDataLoc(ushort Index, LONG Data);

	// can return 0; returned node is locked!
	CNode *GetFirstChild();

	// can return 0
	KEY GetFirstKeyLoc();

	// can return 0; returned node is locked!
	CNode *GetLastChild();

	// can return 0
	KEY GetLastKeyLoc();

	KEY GetKey(ushort Index) { return Info[Index]->GetKey(); }

	short IsThin();

	// cannot return 0;
	CItem *DuplNthItem(ushort Index);

	// cannot return 0;
	// index is out par
	CItem *DuplLastItem(ushort &Index);

	// Child is not child => returns NOIDX
	ushort GetChildIndex(CNode *Child);

	// tree must have at least 3 nodes, this must have children;
	// cannot return NULL; returned node is locked!
	// Index is in&out par
	CNode *GetNeighbor(ushort &Index);

	void PullItems(CItem **TargetInfo, ushort MoveCount,
			short From0);
	void PushItems(CItem **SourceInfo, ushort MoveCount,
			short From0);

	// inserts first or last item;
	// asserts Count < MAXINFO;
	// InItem is inserted or destroyed;
	void InsertExtreme(CItem *InItem, short From0);

	// does not yank the item from Info
	// Index is out param (Count - 1)
	CItem *GetLastPtr(ushort &Index);

#if !defined(FAST_SWAP)
	void SwapItemContent(CItem *Item1, CItem *Item2, CNode *OtherNode);
#else
     void SetLastPtr(CItem *InItem);
#endif
	long GetChildRef(ushort Index);
	void SetChildRef(ushort Index, long Ref);

	// returns 1 - delete FromNode and update Info, 0 - OK
	short BeefUpSlanted(short LeftToRight, ushort PivotIndex,
			CNode *ToNode, CNode *FromNode);

	// ToNode and FromNode must be neighbors;
	// this adds MoveCount keys to ToNode, removes MoveCount
	// keys from FromNode and updates this node
	void MoveItems(short LeftToRight, ushort PivotIndex,
		CNode *ToNode, CNode *FromNode, ushort MoveCount);

	void SlantForward();
	void SlantBackward();

	long GetExtremeChild() { return LastChild; }
	void SetExtremeChild(long ExtremeChild);
		// updates parent of ExtremeChild

	short IsLeftmost(int Index) { return Index == 0; }
	short IsRightmost(int Index) { return Index == Count; }
};

#endif
