#if !defined(__tltree_h)
#define __tltree_h

#include "tltype.h"
#include "remnodes.h"
#include "cache.h"
#include "handler.h"
	 // GCC insists on class declaration for template

class CTree
{
private:
	static const int RNLimit;
	static const int RNDelta;

private:
	CFileObj Stream;
	DString StreamName;
	DISArray<CSortedNode> RemNodes;
	CCache Cache;
	ushort CreateOptions;
protected:
	CObjPtr<CKeyHandler> KeyHandler;
	CSortedNode *Root;

public:
	CTree(CKeyHandler *aKeyHandler,
		const DString &FName, // only path if Opt & IT_TEMP
		ushort Opt);
	virtual ~CTree();

	long AddOffs();
	void DelOffs(long Offs);
	// cannot return 0; increments lock count
	CNode *GetLockedNode(long Offs);

	short UnlockNode(CNode *Node);

	void AddLevel(CItem *Item);

	// unlocks NewRoot!
	void CheckDelLevel(CNode *NewRoot);
	void DestroyNode(CNode *Doomed);

	inline CFileObj *GetStream() { return &Stream; }
	void Store(CNode *Node);
	void Remember(CNode *Node);
	void StoreAll();

	// cannot return 0
	CKeyHandler *GetKeyHandler() { return KeyHandler; }

	void UpdateNode(long Offs, long Parent);

private:
	void DestroyUnlocked(CNode *Doomed);
	void InitNew(); // called from constructor

#if defined(DEBUG_duplicate)
public:
	void CheckDupl();
	int IsLocked(CNode *Node);
	void CheckRemNodes();
#endif
};

#endif

